%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  File: 	pxc/lapopt.sl
%
%  Description: Code Optimization for Mips
%               Reformation of jump sequences
%
%  Author:      H. Melenk, ZIB Berlin
%  Date:        22 Oct 1989
%
% Copyright (c) 1989-1992  Konrad-Zuse-Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% 17-Aug-1992 Winfried Neun (ZIB Berlin)
% Fixed lapopt problem (SW was moved in front of corresponding LW)
          
(compiletime (flag '(bef lapoptpatternmatch lapoptpatternmatch1 lapopt0
		     lapopt2 findlabelfw findlabelbw lapoptremoveauxlabel
		     lapoptcompatiblep lapoptfindusedlabels lapoptsearchbj
		     lapoptsearchfl lapoptsearchbl operands lapoptdispatch)
      'internalfunction))

(fluid '(lapoptpatternvars* lapoptpatterns* lapoptnonmovables* source* target*
          loadinstructions* jumpinstructions* cjumpinstructions* 
          lapopttestfunctions!* lapoptslotinstructions* lispregisters*
          &x &y &z &p &q &r &load &jump *lapopt))
 
(setq lapoptpatternvars* '(&x &y &z &p &q &r))
 
(setq loadinstructions* '(lw lh lb move-relocatable-object-and-load))
(setq jumpinstructions* '(j jr jal jalr))
(setq cjumpinstructions* '(beq bne bgtz bltz blez bgez bc1f bc1t))
(setq jumpinstructions*(nconc jumpinstructions* cjumpinstructions*))
(setq lapoptnonmovables* (append jumpinstructions* 
     '(*ENTRY sw sb sh lw lwl lwr lhu lh lbu lb lwc1 ctc1 cfc1
       mtc1 mfc1 c!.ngt!.d  cvt!.d!.w c!.lt!.d 
       move-relocatable-object move-relocatable-object-and-load
       move-relocatable-object-and-store)))
(setq lapoptslotinstructions* (append loadinstructions* jumpinstructions*))
(setq lapoptslotinstructions* (append lapoptslotinstructions* '(C!.NGT!.D
								C!.LT!.D)))
(setq lispregisters* '((reg 1)(reg 2)(reg 3)(reg 4)(reg 5)))

(setq lapopttestfunctions!* '(**rjump **condjump **ujump atom **exitjump 
      **load))
 
(de **rjump (u) (when (memq u '(jr jalr)) (setq &jump u)))
(de **exitjump (u) (eq u 'jr))
(de **condjump (u) (memq u cjumpinstructions*))
(de **ujump (u) (memq u '(j jr )))
(de **load (u) (when (memq u loadinstructions*) (setq &load u)))


(de lapoptpatternmatch(m pat) 
	(setq &x nil &y nil &z nil &p nil &q nil &r nil)
	(lapoptpatternmatch1 m pat))

(de lapoptpatternmatch1(m pat)
    (cond ((equal m pat) t)
          ((null pat) nil)
          ((memq pat lapoptpatternvars*) 
		(if (null (eval pat)) 
		    (progn (set pat m) t)
		    (equal (eval pat) m)))
	  ((memq pat lapopttestfunctions!*)
	        (apply pat (list m)))
          ((or (atom m)(atom pat)) nil)
          ((lapoptpatternmatch1 (car m) (car pat))
           (lapoptpatternmatch1 (cdr m) (cdr pat)))))


(de Lapopt0(u)
    (when (not *lapopt) (return u))
    (prog (f rcode targ src lab x y oldu done n)
     (setq done t n 0)
     (while done
       (setq done nil n (add1 n))
          %(prin2t "")(prin2"============= durchlauf ") (prin2t n)
          %(mapc u 'print)

       (while u  % main loop
 
            %  (print (list ":::::::" (car u)))
        (setq oldu u)    % save value

	(when
	  (lapoptpatternmatch u
		'((SW &x &y)		% (SW (reg 1) (frame 1))
		  (LW &r &y)		% (LW (reg 2) (frame 1))
		  (NOP) 
                   .&z))
	  (if (equal &x &r)
	      (setq u `(,(car u) ., &z))
	      (setq u `( 
		 ,(car u)
	          (OR ,&r (reg zero) ,&x)
                       . , &z))))

        (when
	 (and
          (lapoptpatternmatch u
                '((**load &x &y)        % (LW (reg 1)(frame 1))
                  (NOP)                 % (NOP)
		  &p                    % (SW (reg 2) (frame 2))
                  . &z))
          (not (atom &p))
	  (progn (lapoptdispatch &p) t)
	  (not (member &x source*)))
             (setq u(cons (car u) (cddr u))))       % kill the NOP

        (when
          (lapoptpatternmatch u  
		'((**load &x &y)	% (LW (reg 1)(frame 1))
		  (NOP)			% (NOP)
                   . &z))
	   (setq targ (list &x) src (operands &y))
           (cond((setq f (lapoptsearchbl rcode targ src))
		 (setq rcode (delete f rcode))
		 (setq u `( ,(car u) ,f . ,&z)))
		((setq f (lapoptsearchfl &z targ src))
		 (setq &z (delete f &z))
		 (setq u `( ,(car u) ,f . ,&z)))))

        (when
         (and
          (lapoptpatternmatch u
                '(&p              % (LW (REG 1) (DISPLACEMENT (REG ST) 8))
                  (LW &x &y)      % (LW (REG RETURNADDR) (DISPLACEMENT (REG ST) 0))
                  (NOP)		  % (NOP)
                  (**rjump &x)    % (JR (REG RETURNADDR))
                   . &z))
          rcode
	  (not (atom (car rcode)))
          (not (memq (caar rcode) lapoptslotinstructions*))
          (not (atom &p))
	  (not (memq (car &p) lapoptnonmovables*))
	  (lapoptdispatch &p)
	  (not (member &x source*))
	  (null (intersection (operands &y) target*))
	  (not (member &x target*)))
            (setq u `( ,(cadr u) ,&p . , (cdddr u))))

        (when 
         (and
          (lapoptpatternmatch u   
                '((LW &x &y)      % (LW (REG RETURNADDR) (DISPLACEMENT (REG ST) 0)) 
                  (NOP)		  % (NOP) 
                  (**rjump &x)    % (JR (REG RETURNADDR))
                   . &z)) 
          (setq f (lapoptsearchbj rcode (list &x) (list &x &y) )))
         (setq rcode (delete f rcode)) 
         (setq u `( ,(car u) ,f ,(caddr u) . ,&z))) 

        (when
          (lapoptpatternmatch u
                '((**load  &x &y)
                  (LW (reg t4) (entryaddress &q))  % load jump register
                  (NOP)           % another instruction
                  (**rjump (reg returnaddr) (reg t4)) % jump with register
                   . &z))
           (setq u `((LW (reg t4) (entryaddress ,&q))
                     (,&load ,&x ,&y)
                     (,&jump (reg returnaddr) (reg t4))
                     . ,&z)))

        (when
          (lapoptpatternmatch u
                '((**load  &x &y)
                  (LW (reg returnaddr) &q)  % load jump register
                  (NOP)           % another instruction
                  (JR (reg returnaddr))   % jump with register
                   . &z))
           (setq u `((LW (reg returnaddr) ,&q)
                     (,&load ,&x ,&y)
                     (jr (reg returnaddr))
                     . ,&z)))

        (when
         (and
          (lapoptpatternmatch u 
                '((LW &x &y)      % load jump register
		   &q             % another instruction
		  (**rjump &x)    % jump with register
                  (NOP)           % NOP
                   . &z))
          (setq f (lapoptsearchbj rcode (list &x) (list &x &y) )))
         (setq rcode (delete f rcode))
         (setq u `( ,(car u) ,f ,(caddr u) ,&q . ,&z)))

        (when                                               
         (and
          (lapoptpatternmatch u 
                '((&q &p . &r) 		% (and (reg 2)(reg 2)(reg infmask))
		  &y 			% (or (reg 1)(reg zero)(reg nil))
		  (**rjump &x)		% (jr (reg returnaddr))
		  (NOP)			% (NOP)
		  . &z))
	  (pairp &y)
	  (not (memq (car &y) lapoptslotinstructions*))
          (not (memq (car &y) lapoptnonmovables*))
          (lapoptdispatch &y) 
          (null (intersection (list &x) target*))
	  (or (not (eq &q 'lw))         % should not be the load for jump target
	      (not (equal &p &x))))
            (setq u `(,(car u) ,(caddr u) ,&y  . ,&z)))

        (when 
          (lapoptpatternmatch u   
                '((J &y)          % uncond absolute jump
                  (NOP)           % NOP 
                   . &z)) 
          (setq x (gensym) lab &y)
	  (cond 
		((setq f (lapoptsearchbj rcode nil nil ))
         	 (setq rcode (delete f rcode)) 
         	 (setq u `( ,(car u) ,f  . ,&z)))
		((or (setq f (findlabelbw rcode lab x))
		     (setq f (findlabelfw &z lab x)))
		 (setq u `((J ,x) ,f . ,&z))))) 

        (when
          (lapoptpatternmatch u 
               '((**condjump . &x)  % conditional jump
		 (NOP)
		  . &z))
		(lapoptdispatch (car u))
		(setq lab (car target*) y source*)
                (setq f (lapoptsearchbj rcode nil y))
                (if f
		    (progn
			(setq rcode (delete f rcode))
			(setq u `( ,(car u) ,f . ,&z)))
		    (progn
			(setq x (gensym)) % new label
			(when (or   (setq f (findlabelbw rcode lab x))
		      		    (setq f (findlabelfw &z lab x)))
                                (lapoptdispatch f)
                                (setq y target*)
				(if   (lapoptcompatiblep &z y)
				      (setq u
					`(,(subst x lab (car u))  % new jump
					  ,f
					  . ,&z))
                                      (lapoptremoveauxlabel) )))))

		
        (when
          (and (equal (car u) '(NOP))     % NOP superfluous?
              % NIL                        % blocked for now
               rcode 
               (or (atom (car rcode))
                   (not (memq (caar rcode) lapoptslotinstructions*))))
              (pop u))
          
	%%%%%%% (when (not (eq u oldu)) (setq done t))
 
%       (when (and (pairp (car u)) (memq (caar u) lapoptslotinstructions*))

	(push (pop u) rcode))
      (setq u (reversip rcode) rcode nil))
      (when (greaterp n 3) (prin2t (bldmsg "#### %w durchlaeufe sinnvoll" n)))

    (return (lapopt2 u)))) 
 
(de lapopt2(u)
  % remove dead code parts
  (prog (rcode usedlabels)
	(setq usedlabels (lapoptfindusedlabels u))
 
     (while u
       
	(when 
	 (and
	  (lapoptpatternmatch u
	     '((**ujump . &x)
		&y
		&p
		&q
		&r
		. &z))
	 (idp &p)
	 (idp &r)
	 (not (memq &p usedlabels)))
 
             (setq u `(,(car u) ,&y ,&r . ,&z)))
 
	(push (pop u) rcode))
 
    (return (reversip rcode))))
%
% looking for labels
%

(fluid '(auxlabel*))

(de findlabelfw (code lab newlab)
      % find label lab in code , return next instruction and
      % insert new label behind
    (cond ((null code) nil)
          ((not (atom lab)) nil)
	  ((and (setq code (member lab code)) 
		(setq code (cdr code))
		(cdr code))
           (let (( nextins (car code)))
                (when (not (or (atom nextins) (member (car nextins) lapoptnonmovables*)))
	           	(setq auxlabel* code)
	   		(rplacd code (cons newlab (cdr code)))
	   		(car code))))
))

(de findlabelbw (code lab newlab)
      % here code is in reverse order
      % find label lab in code , return next instruction and 
      % insert new label behind 
   (cond ((or (null code) (null (cdr code)) (null (cddr code))) nil)
	 ((eq (caddr code) lab)
	  (setq lab (cadr code))
          (when (not (or (atom lab) (member (car lab) lapoptnonmovables*)))
	     (setq auxlabel* code)
             (rplacd code (cons newlab (cdr code)))
	     lab))
	 (t (findlabelbw (cdr code) lab newlab))))

(de lapoptremoveauxlabel ()
	(when (pairp auxlabel*) (rplacd auxlabel* (cddr auxlabel*)))
	(setq auxlabel* nil))

(de lapoptcompatiblep (code targ)
  % look, if the instructions in code overwrite the members in targ completely
  % or if there is an exit without using them again
   (cond((null code) t)
	((atom (car code)) nil)
	((and (**exitjump  (caar code))
	      (null (intersection targ (cdar code)))
	      (null (intersection targ lispregisters*))) t) 
	((or (atom (car code))
	     (memq (caar code) jumpinstructions*)) nil)
        (t (lapoptdispatch (car code)) 
	   (if   (intersection source* targ) 
		 nil
                 (progn 
			(foreach x in target* do (setq targ (delete x targ)))
		 	(lapoptcompatiblep (cdr code) targ ))))))
 
(de lapoptfindusedlabels (u)
  (cond ((null u) nil)
	((or (atom (car u)) (not (memq (caar u) jumpinstructions*)))
	 (lapoptfindusedlabels (cdr u)))
	(t (cons (lastcar (car u)) (lapoptfindusedlabels (cdr u))))))

%
% looking for movable instructions
%

(de lapoptsearchbj(u targs srcs)
      % backwards, for a jump slot
  (prog (inst)
 loop
    (when (or (null u) 
              (atom (car u)) 
	      (memq (caar u) lapoptnonmovables*)
              (memq (caar u) lapoptslotinstructions*)
              (and 
                   (cdr u) 
                   (pairp (cadr u))
                   (memq (caadr u) lapoptslotinstructions*)))
          (return nil)) 
    (setq inst (pop u))
    (lapoptdispatch inst)
    (when (and (null (intersection target* srcs))
	       (null (intersection source* targs))
               (not (equal inst '(NOP))))
	  (return inst))
    (setq targs (union targs target*))
    (setq srcs  (union srcs source*))
      (go loop))) 

(de lapoptsearchfl(u targs srcs)
      % forwarddirection, for load slot
  (prog (inst)
 loop
    (when (or (null u)
              (atom (car u))
              (memq (caar u) lapoptnonmovables*))
          (return nil))
    (setq inst (pop u))
    (lapoptdispatch inst)
    (when (and (null (intersection target* srcs))
               (null (intersection source* targs))
               (not (equal inst '(NOP))))
          (return inst))
    (setq targs (union targs target*))
    (setq srcs  (union srcs source*))
   (go loop)))
 
(de lapoptsearchbl(u targs srcs)
      % backward direction, for load slot
      % here we have to be more conservative, if we
      % get near to a delayslot
  (prog (inst)    
 loop
    (when (or (null u)
              (atom (car u))
              (memq (caar u) lapoptnonmovables*)
              (memq (caar u) lapoptslotinstructions*)
              (and (cdr u)
                   (pairp (cadr u))
                   (memq (caadr u) lapoptslotinstructions*)))
          (return nil))
    (setq inst (pop u))
    (lapoptdispatch inst)
    (when (and (null (intersection target* srcs))
               (null (intersection source* targs))
               (not (equal inst '(NOP)))
               (not (memq (car inst)loadinstructions*)))
          (return inst))
    (setq targs (union targs target*))
    (setq srcs  (union srcs source*))
      (go loop))) 


(de operands (x)
   % extract the operands from x
   (cond ((atom x) nil)
	 ((member x '((reg zero))) nil)
         ((or (memq (car x) '(reg $fluid fluid $global global))
	      (and (eqcar x 'displacement) (equal (cadr x) '(reg st))))
	  (list x))
         ((memq (car x) '(displacement label wconst indexed indirect))
	  (operands (cadr x)))
	 (t (nconc (operands (car x)) (operands (cdr x))))))

 
(de lapoptdispatch(x)
  % select source and target from an instruction
  (when (pairp x)
   (let ((op (car x)))
     (cond((memq op '(sw sh sb mtc1))
	    (setq source* (cons (cadr x) (operands (cddr x))))
	    (setq target* (operands (cddr x))))
          ((eq op 'nop) (setq source* nil target* nil))
	  ((eq op 'jal) (setq target* (lastcar x) source* nil))
	  ((eq op 'jalr) (setq target* nil source* (lastpair x)))
	  ((eq op 'jr)(setq target* (cdr x) source* (cdr x)))
          ((memq op '(mult div divu multu))
		      (setq target* '(hi&lo))
                      (setq source* (cdr x)))
          ((memq op '(mfhi mflo)) (setq target* (cdr x))
				  (setq source* '(hi&lo)))
	  ((memq op cjumpinstructions*)
	   (setq op (reverse (cdr x)))
	   (setq target* (list (car op)))
	   (setq source* (cdr op))) 
          (t(setq source* (operands (cddr x)))
	    (setq target* (list (cadr x))))))))

 
(fluid '(*lapopttrace))
 
(de lapopt(code1)
  (when (null *lapopt) (return code1))
  (let*((code2 (lapopt0 code1))
	(c1 code1)
	(c2 code2))
     (when *lapopttrace
       (linelength 120)
       (while (or c1 c2)
	(cond ((null c1) (bef '| (pop c2)))
		((null c2) (bef (pop c1) '|))
		((equal (car c1)(car c2)) (bef (pop c1)(pop c2)))
		((and (atom (car c1)) (member (car c1) c2))
			(bef '| (pop c2)))
		((and (atom (car c2)) (member (car c2) c1))
			(bef (pop c1) '|))
		(t (bef (pop c1) (pop c2))))))
     code2))
 
(de bef (b1 b2)
	(prin2 b1)
	(tab 50)
	(prin2t b2))
	
