%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           spectrum-asm.sl
% Title:          HP RISC specific information for LAP-TO-ASM
% Author:         Winfried Neun, ZIB Berlin
% Created:        25 August 1991
% Package:        Compiler
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime
 (progn (load if!-system common)
        (off usermode)))

(fluid '(codefilenameformat!* datafilenameformat!* inputsymfile!* 
         outputsymfile!* commentformat!* labelformat!* 
         externaldeclarationformat!* exporteddeclarationformat!* 
         fullwordformat!* doublefloatformat!* reservezeroblockformat!* 
         reservedatablockformat!* definedfunctioncellformat!* 
         undefinedfunctioncellinstructions!* mainentrypointname!* 
         codeout!* dataout!* !*lower asmopenparen!* asmcloseparen!* 
         numericregisternames!* pathin!*))
%%
%%       assembler must declare global name before use.
%%
(fluid '(*declarebeforeuse))

(setq *declarebeforeuse t)

(setq codefilenameformat!* "%w.s")

(setq datafilenameformat!* "d%w.s")

(setq inputsymfile!* "spectrum.sym")

(setq outputsymfile!* "spectrum.sym")

(setq mainentrypointname!* (intern "psl_main"))

(setq commentformat!* ";  %p %n")

(setq labelformat!* "%w%n")

(setq exporteddeclarationformat!* "	.EXPORT 	%w%n")
(setq externaldeclarationformat!* "	.IMPORT 	%w%n")

(setq fullwordformat!* "	.WORD	%e%n")

% FullWord expects %e for parameter                                        

(setq doublefloatformat!* "	.DOUBLE	0d%w%n")

(setq reservezeroblockformat!* "%w	.BLOCKZ	 %e%n")

(setq reservedatablockformat!* "	.BLOCK	%e%n")

(setq definedfunctioncellformat!* "	.WORD  %w%n")

(setq undefinedfunctioncellinstructions!*
      '((!.WORD undefinedfunction)))

%% From lap-to-asm: 

(de initsymfnc1 (x)
  (prog (ep)
        (setq ep (get x 'entrypoint))
        (if (null ep)
          (dataprintundefinedfunctioncell (id2int x))
          (dataprintdefinedfunctioncell ep x))))

(compiletime (fluid '(**low-case)))

(de dataprintundefinedfunctioncell (idn)
  (prog (oldout)
        (setq oldout (wrs dataout*))
	(if (and (boundp '**low-case) **low-case)
	    (dataprintf "   .word   undefinedfunction%n")
            (dataprintf "	.word	UNDEFINEDFUNCTION%n"))
        (wrs oldout)))

(de dataprintdefinedfunctioncell (x idn)
   (datadeclareexternal x)
     % dont disturb undefined function and interpreter
   (dataprintf definedfunctioncellformat* x (id2int idn)))

(fluid '(&links-present& locallabels*))

(de asmoutlap (u)
  (prog (locallabels* oldout)
        (setf &links-present& (or (look-for-links u)
                                  (assoc '*Id-apply u)
				  (assoc '*call u)
                                  (assoc '*link u)))

        (setq u (pass1lap (&fillframeholes u)))
        (setq u (lapoptreformjumps u)) % lapopt pass
        % Expand cmacros, quoted expressions
        (codeblockheader)
        (setq oldout (wrs codeout*))
        (foreach x in u do (asmoutlap1 x))
        (wrs oldout)
        (codeblocktrailer)))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(setq asmopenparen!* '!()

(setq asmcloseparen!* '!))

(deflist '((remainder !%) (land !&) (lor !|) (lxor !^) (lsh !<!<)) 
         'binaryasmop)

(put 'lnot 'unaryasmop '!!)

%(put 'mkitem 'asmexpressionformat "(%e<<27)+%e")
(fluid '(*mkitemformat))
(setq *mkitemformat "%e+(%e<<27)") % ZIB

(put 'mkitem 'asmexpressionfunction 'asm-mkitem)

(de asm-mkitem (mkitem-tag-inf)
  (printf
    *mkitemformat
    `(printexpression ,(caddr mkitem-tag-inf))
    `(printexpression ,(cadr mkitem-tag-inf))))

(global '(*writingasmfile))

(de codefileheader ()

  (setq *writingasmfile t)      %% Somebody must set this flag. but who?

  (codeprintf "     .SPACE   $PRIVATE$ %n")
  (codeprintf "     .SUBSPA  $DATA$ %n")
  (codeprintf "; #include ""pcc_prefix.s""%n"))

(de datafileheader ()
  (dataprintf "     .SPACE   $PRIVATE$ %n")
  (dataprintf "     .SUBSPA  $DATA$ %n"))

(de codefiletrailer ()
  (codeprintf "     .END %n")  nil)

(de datafiletrailer ()
  (dataprintf "     .END %n")  nil)

(de codeblockheader ()
  (codeprintf "	.ALIGN	4%n"))

(de codeblocktrailer ()
  nil)

(de dataalignfullword ()
  (dataprintf "	.ALIGN4%n"))

(de printstring (s)
  (prog (n)
        (setq n (size s))
        (for (from i 0 n 1) (do (printbyte (indx s i))))
        (printbyte 0)
        (printf "	.ALIGN	4%n")))

(de printbytelist (l)
  (foreach x in l do (printbyte x)))

(de printbyte (x)
  (printf "	.byte ")
  (printexpression x)
  (terpri))

(de printhalfwordlist (l)
  (foreach x in l do (printhalfword x)))

(de printhalfword (x)
  (printf "     .HALF ")
  (printexpression x)
  (terpri))

(de printopcode (x)
        (prin2 "  ")(prin2 x))

(de printopcodewithcompleter (x compl)
       (prin2 "  ") (prin2 x) (prin2 ",") (prin2 (cadr compl)))

(de specialactionformainentrypoint ()
  (codeprintf "	.EXPORT	psl_main%n  .EXPORT psl_main2%n psl_main2%n  psl_main"))

(de asmsymbolp (x)
  (unixsymbolp (if (idp x)
                 (id2string x)
                 x)))

(de unixsymbolp (x)
  (prog (n c i)
        (setq n (size x))
        (setq c (indx x 0))
        (unless (or (and (geq c (char a)) (leq c (char z))) 
                    (equal c (char _)) 
                    (and (geq c (char !a)) (leq c (char !z))))
          (return nil))
        (setq i 1)
   loop
        (when (greaterp i n)
          (return t))
        (setq c (indx x i))
        (unless (or (and (geq c (char a)) (leq c (char z))) 
                    (equal c (char _)) 
                    (and (geq c (char !0)) (leq c (char !9))) 
                    (and (geq c (char !a)) (leq c (char !z))))
          (return nil))
        (setq i (plus i 1))
        (go loop)))

(de printnumericoperand (x)
  (prin2 x))

(de operandprintdeferred (x)
  (setq x (cadr x))
  (if (spectrum-register-p x)
    (progn (printoperand x))
    (error 0 (list " On HP RISC,deferred operand can't be used like this:" x))))

(put 'deferred 'operandprintfunction 'operandprintdeferred)

(de operandprintdisplacement (x)
  (setq x (cdr x))
  (printoperand (caddr x))
  (prin2 '!( )
  (printoperand  (car x))
  (prin2 '!, )
  (printoperand  (cadr x))
  (prin2 '!) ) )

(put 'displacement 'operandprintfunction 'operandprintdisplacement)

(de operandprintdisplacement2 (x)
  (setq x (cdr x))
  (printoperand (cadr x))
  (prin2 '!( )
  (printoperand  (car x))
  (prin2 '!) ) )

(put 'displacement2 'operandprintfunction 'operandprintdisplacement2)

(de operandprintindexed (x)
  (setq x (cdr x))
  (printoperand (car x))
  (prin2 '!( )
  (printoperand (cadr x))
  (prin2 '!, )
  (printoperand (caddr x))
  (prin2 '!) ))

(put 'indexed 'operandprintfunction 'operandprintindexed)

(de operandprintscaled (x)
  (setq x (cdr x))
  (printoperand (car x))
  (prin2 '![ )
  (printoperand (cadr x))
  (prin2 '!] ))

(put 'scaled 'operandprintfunction 'operandprintscaled)

(de operandprintimmediate (x)
  (printexpression (cadr x)))

(put 'immediate 'operandprintfunction 'operandprintimmediate)
%%
%%      high and low is special keyword for 88k-aviion assembler.
%%
(de operandprinthigh (x)
  (prin2 "l%")
  (printoperand (cadr x)))

(put 'high 'operandprintfunction 'operandprinthigh)
(put 'hi16 'operandprintfunction 'operandprinthigh)

(de operandprintlow (x)
  (prin2 "r%")
  (printoperand (cadr x)))

(put 'low 'operandprintfunction 'operandprintlow)
(put 'lo16 'operandprintfunction 'operandprintlow)

(de operandprintwarray (x)
  (prin2 (cadr x)))

(put 'WARRAY 'operandprintfunction 'operandprintwarray)

(de operandprintforeignentry (x)
  (prog (!*lower symnamx)
        (setq symnamx (symnam (id2int (cadr x))))
        (setq !*lower t)
        (prin2 (cadr x))))

(put 'foreignentry 'operandprintfunction 'operandprintforeignentry)

(de mcprint (x)
  % Echo of MC's                                                           
  (codeprintf "; %p %n" x)
  )

(de instructionprint (x)
  (codeprintf ";  %p %n" x)
  )

(de !*cerror (x)
  (prog (i)
        (setq i (wrs nil))
        (printf "%n *** CERROR: %r %n " x)
        (wrs i)
        (return (list (list 'cerror x)))))

(put 'cerror 'asmpseudoop 'printcomment)

(defcmacro !*cerror)
%%
%%      jmpl has something different syntax in sun4 `as'.
%%
(put 'jmpl 'asmpseudoop 'printjmpl)

(de printjmpl (x)
  (let ((src (cadr x))
        (dst (caddr x)))
       (prin2 " jmpl ")
       (cond ((eq (car src) 'deferred)
              (printoperand (cadr src)))
             ((eq (car src) 'displacement)
              (printoperand (cadr src))
              (print-secondary-source (caddr src)))
             (t
              (printoperand (car src))
              (cond ((cdr src)
                     (print-secondary-source (cadr src))))))
       (prin2 ",")
       (printoperand dst)
       (terpri)))
%%
%%      (word label)
%%
(put 'word 'asmpseudoop 'print-word)

(de print-word (x)
  (prin2 "	.WORD ")
  (printoperand (cadr x))
  (terpri))

(put '*entry 'asmpseudoop 'asmprintentry)
(fluid '(*mainfound))
(de asmprintentry (x)
  (prog (y)
        (printcomment x)
        (setq x (cadr x))
        (setq y (findentrypoint x))
        (unless (flagp x 'internalfunction)
          (findidnumber x))
        (if (eq x mainentrypointname*)
          (progn (setq *mainfound t)
                 (specialactionformainentrypoint))
          (codedeclareexporteduse y))))

(de codedeclareexporteduse (y)
    (codedeclareexported y)
    (printlabel y)
    (terpri))

(de findentrypoint (x)
  (cond ((get x 'entrypoint) (get x 'entrypoint))
        ((and (asmsymbolp x)
              (not (get x 'symbol))
              (not (flagp x 'foreignfunction)))
         (put x 'entrypoint x) 
         x)
        (t 
         (let ((name (stringgensym)))
           (put x 'entrypoint name)
           name))
        ))

%%
%%      define pseude operation
%%
(de printdefine (x)
  (prin2 (cadr x))
  (prin2 " = "))

%%
%%      define pseude operation
%%
(de printdefine (x)
  (prin2 (cadr x))
  (prin2 " = ")
  (prin2 (caddr x))
  (terpri))

(put 'define 'asmpseudoop 'printdefine)

(de print-register-or-immediate(name)
  (cond ((spectrum-register-p name) (printoperand name))
        (t (printoperand name))))

(de print-immediate (val)
  (cond ((atom val)
         (printoperand val))
        ((eq (car val) 'WARRAY)
         (prin2 (cadr val)))
        (t
         (error 0 (list "Unknown constant expression:" val)))))

(de print-register-name(name)
  (cond ((numberp name) (setq *lower t) (prin2 (to-standard-name name))
					(setq *lower nil))
        ((get name '*register-code) (setq *lower t)
                                    (prin2 (to-standard-name name ))
                                    (setq *lower nil))
        ((get name '*f-register-code)
         (prin2 "%f")
         (prin2 (get name '*f-register-code)))
        ((get name '*c-register-code)
         (prin2 "%c")
         (prin2 (get name '*c-register-code)))
        (t
         (printoperand name))))         %% if reg-name is not a register, it
                                        %% must be a constant.

%%
%%
%%      This function is originally defined in lap-to-asm.sl.
%%
(de printoperand (x)
  (setq x (to-standard-name x))
  (cond ((stringp x) (prin2 x))
        ((numberp x) (printnumericoperand x))
        ((spectrum-register-p x) (print-register-name x))
        ((idp x) (prin2 (findlabel x)))
        ((spectrum-register-p (car x)) (print-src-operand x))
        (t (prog (hd fn)
                 (setq hd (car x))
                 (cond ((setq fn (get hd 'operandprintfunction)) 
                        (apply fn (list x)))
                       ((and (setq fn (getd hd)) (equal (car fn) 'macro)) 
                        (printoperand (apply (cdr fn) (list x))))
                       ((setq fn (wconstevaluable x)) (printoperand fn))
                       (t (printexpression x)))))))
%
%
(de print-src-operand (src)
  (printoperand (car src))
  (and (cdr src)
       (print-secondary-source (cadr src)))
  )

(de print-secondary-source (src)
  (cond ((and (pairp src)
              (eq (car src) '-))
         (cond ((numberp (cadr src))
                (cond ((minusp (cadr src))
                       (prin2 ",")
                       (prin2 (minus (cadr src))))
                      (t
                       (prin2 "-")
                       (prin2 (cadr src)))))
               (t
                (prin2 "-")
                (printoperand (cadr src)))))

        ((numberp src)
         (cond ((minusp src)
                (prin2 "-")
                (prin2 (minus src)))
               (t
                (prin2 ",")
                (printoperand src))))
        (t
         (prin2 ",")
         (printoperand src))))

