%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           sun4-instrs.sl
% Title:          SPARC instructions                                         
% Author:         Tsuyoshi Yamamoto
% Created:        19 May 1988
% Status:         Experimental
% Mode:           Lisp
% Package:        Compiler
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
% 2 Jan 1990  Winfried Neun, ZIB Berlin
%   Included immedatequote and trap instructions
% 20 Nov 1988 Winfried Neun, ZIB Berlin
%     separated this file out of sparc-lap.sl.
%     added handling of  ,a  instructions for conditional jumps
%     added function-cell relocation for Branch operations to entries.
% 30 Jan 1989 Winfried Neun, ZIB Berlin
%     added direct load and store versions for move-relocatable-object  
%     which saves one instruction 
     
(compiletime (load fasl-decls common))

(global '(*writingasmfile *asmpass* *outputmode* *object*))

(fluid '(labeloffsets!* currentoffset* codesize* codebase* entries*
	 forwardinternalreferences* lapreturnvalue*  *writingfaslfile
	 initoffset* *pgwd *pwrds label-alist*))

%%|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%k
%%      Assembler code generator
%%
%%|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

(de define-machine-code (name length op genfun)
  (put name '*op op)
  (put name '*code-length length)
  (put name '*code-gen genfun))
%%
%%      Type-2 instructions
%%
(de type-2 (name rs1 rs2 rd)
  (cond ((sparc-register-p rs2)
	   (make-type-2 (get-register-code rd)
	                (get name '*op)
	                (get-register-code rs1)
	                0
	                (get-register-code rs2)))
	(t (make-type-2 (get-register-code rd)
	                (get name '*op)
	                (get-register-code rs1)
	                1
	                (get-register-code rs2)))))
%%
%%      Type-3 instructions include load and store.
%%
(de type-3 (name rs1 rs2 rd)
  (cond ((null rs2)
	   (make-type-3 (get-any-register-code rd)
	                (get-register-type rd)
	                (get name '*op)
	                (get-register-code rs1)
	                0
	                (get-register-code 'r0)))         %% r0 -> 0 register
	((sparc-register-p rs2)
	   (make-type-3 (get-any-register-code rd)
	                (get-register-type rd)
	                (get name '*op)
	                (get-register-code rs1)
	                0
	                (get-register-code rs2)))
	(t (make-type-3 (get-any-register-code rd)
	                (get-register-type rd)
	                (get name '*op)
	                (get-register-code rs1)
	                1
	                (get-register-code rs2)))))

%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%       Type-0 instructions
%
%       |00|  rd | op2 |      imm22        |
%           29 25 24 22 21                0
%
%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
(de make-type-0 (a cond op disp22)
  (let ((code (wplus2 (wshift a 29)
	              (wplus2 (wshift cond 25)
	                      (wplus2 (wshift op 22)
	                              disp22)))))
       (store-word code)))
%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%       Type-1 instructions  (CALL)
%
%       |01|    30bit displacement        |
%           29                           0
%
%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
(de make-type-1 (offset)
  (let ((op-type 1))
       (store-word (wplus2 (wshift op-type 30) offset))))
%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%       Type-2 instructions
%
%       |10|  rd | rt | op3  | rs1 |0|   0     |  rs2 |
%           29 25 24 23    19 18 14   12      5 4    0
% or,
%       |10|  rd | rt | op3  | rs1 |1|   simm13       |
%           29 25 24 23    19 18 14   12             0
%
%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
(de make-type-2 (rd op3 rs1 i rs2-or-simm13)
  (let ((code (wplus2 (wshift rd 25)
	              (wplus2 (wshift op3 19)
	                      (wplus2 (wshift rs1 14)
	                              (wplus2 (wshift i 13)
	                                      rs2-or-simm13)))))
	(op-type 2))
       (store-word (wplus2 (wshift op-type 30)
	                   code))))
%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%       Type-3 instructions
%
%       |11|  rd |   op3   | rs1 |0|   asi    | rs2 |
%           29 25 24     19 18 14   12       5 4   0
% or,
%       |11|  rd |   op3   | rs1 |1|   simm13       |
%           29 25 24     19 18 14   12             0
%
%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
(de make-type-3 (rd rd-type op3 rs1 i rs2-or-simm13)
  (let ((code (wplus2 (wshift rd-type 23)
	              (wplus2 (wshift rd 25)
	                      (wplus2 (wshift op3 19)
	                              (wplus2 (wshift rs1 14)
	                                      (wplus2 (wshift i 13)
	                                              rs2-or-simm13))))))
	(op-type 3))
       (store-word (wplus2 (wshift op-type 30)
	                   code))))
%%
%%      get-register-code returns register code or 13bit immediate values
%%      assocoated from argument r. r mustbe an integer register name or
%%      13bit constant.
%%
(de get-register-code (r)
  (cond ((and (numberp r)
	      (lessp r 4096)
	      (greaterp r -4097))               %% must be 13 bit or less
	 (wand r 16#01fff))
	((idp r)
	 (cond ((get r '*register-code) (get r '*register-code))
	       ((get r '*defined)  (get-register-code (get r '*defined)))
	       (t  (print-error-status)
	           (error 0 (list "Unknown register name:" r)))))
	((eqcar r 'REG)
	 (cond ((numberp (cadr r))    (cadr r))
	       ((idp (cadr r))        (get-register-code (cadr r)))
	       (t              (print-error-status)
	                       (error 0 (list "Unknown register name:" r)))))
	((and (pairp r) (member (car r) '(low lo)))
	           (wand  1023 (get-value (cadr r))))
	                  % low operator takes 10 bit.
	((eqcar r '-)   (minus (get-register-code (cadr r))))
	(t              (print-error-status)
	                (error 0 (list "Unknown register name:" r)))))
%%
%%      get-any-register-code returns register number associated with r.
%%      r may be integer, florting or co-processor registers.
%%
(de get-any-register-code (r)
  (or (get r '*register-code)
      (get r '*f-register-code)
      (get r '*c-register-code)
      (progn (print-error-status)
	     (error 0 (list "Unknown SPARC register:" r)))))
%%
%%      get-register-type returs:
%%              0: if r is integer regiser.
%%              2: if r is floating register.
%%              3: if r is co-processor register.
%%
(de get-register-type (r)
  (cond ((get r '*register-code)         0)
	((get r '*f-register-code)       2)
	((get r '*c-register-code)       3)
	(t                   (print-error-status)
	                     (error 0 (list "Unknown SPARC register:" r)))))


(de get-value (label)
  (cond ((numberp label) label)
        ((equal *asmpass* 1)         0)
        ((and (idp label) (get label '*defined))
                        (get-value (get label '*defined)))
        ((assoc label label-alist*)
         (wplus2 codebase*
                (cdr (assoc label label-alist*))))
        ((and (pairp label)
              (member (car label) '(FLUID !$FLUID GLOBAL !$GLOBAL)))
              (wplus2 SYMVAL
                  (wtimes2 (id2int (cadr label))
                           addressingunitsperitem)))
        ((and (pairp label) (member (car label) '(entry foreignentry)))
              (wplus2 SYMFNC
                     (wtimes2 (id2int (cadr label))
                             addressingunitsperfunctioncell)))
        ((eqcar label 'immediate) (get-value (cadr label)))
        ((and (eqcar label 'mkitem) 
              (eqcar (caddr label) 'idloc))
                 (mkitem (cadr label) (idinf (cadr (caddr label))))) 
        ((and (eqcar label 'mkitem) 
              (stringp (caddr label)))
                 (mkitem (cadr label) (get-value (caddr label))))
        ((and (eqcar label 'mkitem) (codep (caddr label)))
              (mkitem (cadr label)  (inf (caddr label))))
	((and (eqcar label 'idloc) (idp (cadr label)))
			(wshift (idinf (cadr label)) 10))
        (t       (print-error-status)
                 (error 0 (list "Undefined label:" label)))))

%
%       label must be an atom or (INTERNALENTRY id)
%
(de make-offset (label)
  (prog (there)
	(cond ((null (equal *asmpass* 2))    (return 0))
	      ((numberp label)               (return label))
	      ((atom label)
	       (setq there (assoc label label-alist*))
	       (cond (there
	              (return (wdifference (cdr there) currentoffset!*)))
	             (t
	              (print-error-status)
	              (error 0 (list "Label is undefined at pass 2. " label)))))
	      ((eq (car label) 'internalentry)
	       (cond ((assoc (cadr label) label-alist!*)
	              (return (make-offset (cadr label))))
	             (t
	              (setq there (get (cadr label) 'internalentryoffset))
	              (cond (there
	                     (return (wdifference there currentoffset!*)))
	                    (t
	                     (print-error-status)
	                     (error 0 (list "Undefined internal function:"
	                                    label)))))))
	      (t
	       (return 0)))))

(de forward-internal-reference-p (label)
  (not (or (assoc label label-alist!*)
	   (get label 'internalentryoffset))))

(de make-30bit (val)
   (wshift (wshift val 2) -2))

(de make-22bit (val)    % val may (high xxxx) form
  (cond ((numberp val)   (wshift (wshift val 10) -10))
	(t               (wshift (get-value (cadr val)) -10)))) % x/1024

%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      Pseude-Instructions
%%
(define-machine-code  '*ENTRY 0 NIL 'entry-gen)
%
%       (*ENTRY name type argn)
%
(de entry-gen (instruction name type argn)
  (force-word-aligned)  %% SPARC instruction must be word aligned. 
  (cond ((equal 1 *asmpass*)
	 (setq label-alist* (cons (cons name currentoffset!*) label-alist*))
	 (cond ((flagp name 'internalfunction)
	        (put name 'internalentryoffset currentoffset!*)))))
  (if (equal 2 *asmpass*)
      (saveentry (list instruction name type argn)))
  (if (null *outputmode*)
      (setq *object* (cons (list instruction name type argn)
	                   *object*))))
%
%       (DEFINE label value)
%
(define-machine-code 'DEFINE 0 nil 'define-gen)

(de define-gen (instruction label value)
  (put label '*defined value)
  nil)          %% it generates nothing.
%
%       (WORD value ......)
%
(deflist '((WORD expand-word)
	   (FULLWORD expand-word)
	   (HALFWORD expand-half)
	   (BYTE expand-byte)
	   (STRING expand-string)
	   (FLOAT expand-float))
  'asmexpandop)

(de expand-word (x)
  (mapcar (cdr x)
	  (function (lambda(u)`(DEF-LONG ,u)))))

(de expand-half (x)
  (mapcar (cdr x)
	  (function (lambda(u)`(DEF-HALF ,u)))))

(de expand-byte (x)
  (mapcar (cdr x)
	  (function (lambda(u)`(DEF-CHAR ,u)))))

(de expand-string (x)
  (prog (res)
	(for (from i 0 (size (cadr x)))
	     (do (push `(DEF-char ,(indx (cadr x) i))
	               res)))
	(push `(DEF-char 0) res)
	(return (reverse res))))

(de expand-float (x)
  (mapcar (cdr x)
	  (function (lambda(u)`(DEF-DOUBLE ,u)))))
(define-machine-code 'DEF-double 8 nil 'define-double)
(define-machine-code 'DEF-long 4 nil 'define-long)
(define-machine-code 'DEF-half 2 nil 'define-half)
(define-machine-code 'DEF-char 1 nil 'define-char)

(de define-long (instruction value)
  (store-word-expression value))

(de define-half (instruction value)
  (store-halfword-expression value))

(de define-char (instr value)
  (store-byte value))

(de define-double (instruction value)
  (store-double value))
%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%
%%    1. Load Instructions.
%%
(de ld-gen (name src rd)
  (cond ((cdr src)  (type-3 name (car src) (cadr src) rd))
	(t          (type-3 name (car src) 'r0 rd))))

(define-machine-code 'LDSB 4 9  'ld-gen)
(define-machine-code 'LDSH 4 10 'ld-gen)
(define-machine-code 'LDUB 4 1  'ld-gen)
(define-machine-code 'LDUH 4 2  'ld-gen)
(define-machine-code 'LD   4 0  'ld-gen)
(define-machine-code 'LDD  4 3  'ld-gen)

%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      2. Store instructions.
%%
(de st-gen (name rd src)
  (cond ((cdr src)  (type-3 name (car src) (cadr src) rd))
	(t          (type-3 name (car src) 'r0 rd))))

(define-machine-code 'STB 4 5 'st-gen)
(define-machine-code 'STH 4 6 'st-gen)
(define-machine-code 'ST  4 4 'st-gen)
(define-machine-code 'STD 4 7 'st-gen)
%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      3. Swap instrutions.
%%
(define-machine-code 'SWAP 4 15 'ld-gen)
%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      4. Add instructions.
%%
(define-machine-code 'ADD    4  0 'type-2)
(define-machine-code 'ADDCC  4 16 'type-2)
(define-machine-code 'ADDX   4  8 'type-2)
(define-machine-code 'ADDXCC 4 24 'type-2)
%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      5. Subtract instructions.
%%
(define-machine-code 'SUB    4  4 'type-2)
(define-machine-code 'SUBCC  4 20 'type-2)
(define-machine-code 'SUBX   4 12 'type-2)
(define-machine-code 'SUBXCC 4 28 'type-2)
%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      ??. Multiply and Divide ("SuperSPARC")
%%
(define-machine-code 'UMUL   4 2#001010 'type-2)
(define-machine-code 'SMUL   4 2#001011 'type-2)
(define-machine-code 'UMULcc 4 2#011010 'type-2)
(define-machine-code 'SMULcc 4 2#011011 'type-2)

(define-machine-code 'UDIV   4 2#001110 'type-2)
(define-machine-code 'SDIV   4 2#001111 'type-2)
(define-machine-code 'UDIVcc 4 2#011110 'type-2)
(define-machine-code 'SDIVcc 4 2#011111 'type-2)

%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      6. Logical instructions.
%%
(define-machine-code 'AND    4  1 'type-2 )
(define-machine-code 'ANDCC  4 17 'type-2 )
(define-machine-code 'ANDN   4  5 'type-2 )
(define-machine-code 'ANDNCC 4 21 'type-2 )
(define-machine-code 'OR     4  2 'type-2 )
(define-machine-code 'ORCC   4 18 'type-2 )
(define-machine-code 'ORN    4  6 'type-2 )
(define-machine-code 'ORNCC  4 22 'type-2 )
(define-machine-code 'XOR    4  3 'type-2 )
(define-machine-code 'XORCC  4 19 'type-2 )
(define-machine-code 'XNOR   4  7 'type-2 )
(define-machine-code 'XNORCC 4 13 'type-2 )
%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      7. Shift instructions
%%
(define-machine-code 'SLL 4 37 'type-2 )
(define-machine-code 'SRL 4 38 'type-2 )
(define-machine-code 'SRA 4 39 'type-2 )
%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      7.1 Save and restore instructions
%%
(define-machine-code 'SAVE    4 60 'type-2 )
(define-machine-code 'RESTORE 4 61 'type-2 )
%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      8. Sethi instruction
%%
(define-machine-code 'SETHI 4 4 'sethi-gen)

(de sethi-gen (name value rd)
  (make-type-0 0
	       (get-register-code rd)
	       (get name '*op)
	       (make-22bit value)))
%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      9. Save and restore instructions.
%%
(define-machine-code 'SAVE    4 60 'type-2)
(define-machine-code 'RESTORE 4 61 'type-2)
%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      10. Branch instructions.
%%
(fluid '(the-a))

(prog (*lower)
(setq the-a (car (explode2 'aa))))  % an a in the actual character case.

(de define-bcc-code (name code)
  (prog (comma-a *lower)
    (setq comma-a (intern (bldmsg "%w,%c" name the-a)))
    (put comma-a '*bcc code)
    (put comma-a 'annul 1)
    (define-machine-code comma-a 4 2 'bcc-gen)
    (put name '*bcc code)
    (put name 'annul 0)
    (define-machine-code name 4 2 'bcc-gen )))

(de define-fbcc-code (name code)
  (prog (comma-a *lower) 
    (setq comma-a (intern (bldmsg "%w,%c" name the-a))) 
    (put comma-a '*bcc code) 
    (put comma-a 'annul 1)
    (define-machine-code comma-a 4 6 'bcc-gen) 
    (put name '*bcc code)
    (put name 'annul 0)
    (define-machine-code name 4 6 'bcc-gen )))

(de make-bcc (name label)
  (make-type-0 (get name 'annul)
	       (get name '*bcc)
	       (get name '*op)
	       (make-22bit (wquotient (make-offset label) 4))))

(define-bcc-code 'BA   8)   (define-bcc-code 'BN   0)
(define-bcc-code 'BNE  9)   (define-bcc-code 'BE   1)
(define-bcc-code 'BG   10)  (define-bcc-code 'BLE  2)
(define-bcc-code 'BGE  11)  (define-bcc-code 'BL   3)
(define-bcc-code 'BGU  12)  (define-bcc-code 'BLEU 4)
(define-bcc-code 'BCC  13)  (define-bcc-code 'BCS  5)
(define-bcc-code 'BPOS 14)  (define-bcc-code 'BNEG 6)
(define-bcc-code 'BVC  15)  (define-bcc-code 'BVS  7)

(define-fbcc-code 'FBA    8)  (define-fbcc-code 'FBN    0)
(define-fbcc-code 'FBU    7)  (define-fbcc-code 'FBG    6)
(define-fbcc-code 'FBUG   5)  (define-fbcc-code 'FBL    4)
(define-fbcc-code 'FBUL   3)  (define-fbcc-code 'FBLG   2)
(define-fbcc-code 'FBNE   1)  (define-fbcc-code 'FBE    9)
(define-fbcc-code 'FBUE   10) (define-fbcc-code 'FBGE   11)
(define-fbcc-code 'FBUGE  12) (define-fbcc-code 'FBLE   13)
(define-fbcc-code 'FBULE  14) (define-fbcc-code 'FBO    15)

(de define-trap-code (name code)
  (prog (comma-a)
    (put name '*ticc code)
    (define-machine-code name 4 6 'ticc-gen )))
 
(define-trap-code 'TA    8)  (define-trap-code 'TN    0)
(define-trap-code 'TVC   7)  (define-trap-code 'TNEG  6)
(define-trap-code 'TCS   5)  (define-trap-code 'TLEU  4)
(define-trap-code 'TL    3)  (define-trap-code 'TLE   2)
(define-trap-code 'TE    1)  (define-trap-code 'TNE   9)
(define-trap-code 'TG    10) (define-trap-code 'TGE   11)
(define-trap-code 'TGU   12) (define-trap-code 'TCC   13)
(define-trap-code 'TPOS  14) (define-trap-code 'TVC   15)


(de bcc-gen (name label)

  (cond ((equal *asmpass* 1) (store-word 0))
	           %% operand is an offset value
	((numberp label)    (make-bcc name label ))
	           %% operand is a label.
	((assoc label label-alist*)  (make-bcc name label)) 
	((idp label)               %% operand is an ID                   
	  (if *writingfaslfile
	     (store-relocatable-bcc name label)
	     (bcc-gen name 
	        (wdifference 
	          (wplus2 symfnc
	               (wtimes2 (id2int label) addressingunitsperfunctioncell))
	          (wplus2 codebase* currentoffset*)))
	))
	((atom label)                   %% Other atomic labels are illegal.
	      (print-error-status)
	      (error 0 (list "Wrong bcc operand:" label)))
	((eq (car label) 'ENTRY)        %% bcc (entry label)) format 
	    (bcc-gen name (cadr label)))
	((eq (car label) 'INTERNALENTRY)
	   (if (forward-internal-reference-p (cadr label))
	      (progn
	        (push (list name (cadr label) currentoffset!*)
	               forwardinternalreferences!*)
	        (depositinstruction  0))
	      (make-bcc name label)))
	(t (print-error-status)
	   (error 0 (list "Wrong bcc operand:" label)))))

(de ticc-gen (name label)
 
  (cond ((numberp label)    (make-ticc name label ))
        (t (print-error-status)
           (error 0 (list "Wrong ticc operand:" label)))))

(de make-ticc (name number)
  (let ((code (wplus2 (wshift (get name '*ticc) 25)
                      (wplus2 (wshift 2#111010 19)
                              (wplus2 (wshift 1 13) number))))
        (op-type 2))
       (store-word (wplus2 (wshift op-type 30)
                           code))))

%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      12. Call instruction
%%
(define-machine-code 'sethi-dummy 4 0 'sethidum-gen)

(de sethidum-gen (name label)
    (store-relocatable-call label))

(define-machine-code 'CALL 4 0 'call-gen)

(de call-gen (name label)
  (cond ((equal *asmpass* 1) (store-word 0))
	((numberp label)        %% operand is an offset value
	 (make-type-1 (wquotient label 4)))
	((assoc label label-alist*)     %% operand is a label.
	 (make-type-1 (make-30bit (wquotient (make-offset label) 4)))) 
	((idp label)                    %% operand is an ID
	 (if *writingfaslfile
	     (store-relocatable-call label)
	     (call-gen name
	               (wdifference (wplus2 symfnc
	                                   (wtimes2 (id2int label)
	                                    addressingunitsperfunctioncell))
	                           (wplus2 codebase* currentoffset*)))))
	((atom label)                   %% Other atomic labels are illeal.
	 (print-error-status)
	 (error 0 (list "Wrong call operand:" label)))
	((eq (car label) 'ENTRY)        %% (call (entry label)) format
	 (call-gen name (cadr label)))
	((eq (car label) 'INTERNALENTRY)
	 (cond ((forward-internal-reference-p (cadr label))
	        (push (list 'CALL (cadr label) currentoffset!*)
	              forwardinternalreferences!*)
	        (depositinstruction  0))
	       (t
	        (make-type-1 (make-30bit (wquotient (make-offset label) 4))))))
	((eq (car label) 'FOREIGNENTRY)
           (if *writingfaslfile (store-relocatable-call (cadr label))
                                (call-gen name (cadr label))))
	(t
	 (print-error-status)
	 (error 0 (list "Wrong call operand:" label)))))

%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      13. Jmpl instruction
%%
(define-machine-code 'JMPL 4 56 'jmpl-gen )

(de jmpl-gen (name address rd)
  (cond ((cdr address)              %(jmpl (rs1 rs2) rd)
	   (type-2 name (car address) (cadr address) rd))
	(t                          %(jmpl (rs1) rd)
	   (type-2 name (car address) 'g0 rd))))

%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%      14. Floating arithmetics.
%%
(fluid '(fcmp-magic* fmath-magic*))

(setq fcmp-magic* 53 fmath-magic* 52)

(de define-fcnv-instruction (name opf)
  (put name '*opf* opf)
  (define-machine-code name 4 fmath-magic* 'make-fcnv ))

(de define-fcmp-instruction (name opf)
  (put name '*opf* opf)
  (define-machine-code name 4 fcmp-magic* 'make-fcmp))

(de define-fmath-instruction (name opf)
  (put name '*opf* opf)
  (define-machine-code name 4 fmath-magic* 'make-fmath))

(de make-fcnv (name rs2 rd)
  (make-type-2 (get-f-register-code rd)
	       (get name '*op)
	       0 
	       0
	       (wplus2 (wshift (get name '*opf*) 5)
	               (get-f-register-code rs2))))

(de make-fcmp (name rs1 rs2)
  (make-type-2 0
	       (get name '*op)
	       (get-f-register-code rs1)
	       0
	       (wplus2 (wshift (get name '*opf*) 5)
	              (get-f-register-code rs2))))

(de make-fmath (name rs1 rs2 rd)
  (make-type-2 (get-f-register-code rd)
	       (get name '*op)
	       (get-f-register-code rs1)
	       0
	       (wplus2 (wshift (get name '*opf*) 5)
	              (get-f-register-code rs2))))

(de get-f-register-code (r)
  (or (get r '*f-register-code)
      (progn (print-error-status)
	     (error 0 (list "Unknown floating register name:" r)))))

(define-fcnv-instruction 'FiTOs 16#c4 )
(define-fcnv-instruction 'FiTOd 16#c8 )
(define-fcnv-instruction 'FiTOx 16#cc )
(define-fcnv-instruction 'FsTOir 16#c1 )
(define-fcnv-instruction 'FdTOir 16#c2 )
(define-fcnv-instruction 'FxTOir 16#c3 )
(define-fcnv-instruction 'FsTOi 16#d1 )
(define-fcnv-instruction 'FdTOi 16#d2 )
(define-fcnv-instruction 'FxTOi 16#d3 )
(define-fcnv-instruction 'FsTOd 16#c9 )
(define-fcnv-instruction 'FsTOx 16#cd )
(define-fcnv-instruction 'FdTOs 16#c6 )
(define-fcnv-instruction 'FdTOx 16#ce )
(define-fcnv-instruction 'FxTOs 16#c7 )
(define-fcnv-instruction 'FxTOd 16#cb )
(define-fcnv-instruction 'FMOVs 16#01 )
(define-fcnv-instruction 'FNEGs 16#05 )
(define-fcnv-instruction 'FABSs 16#09 )
(define-fcnv-instruction 'FSQRTs 16#29 )
(define-fcnv-instruction 'FSQRTd 16#2a )
(define-fcnv-instruction 'FSQRTx 16#2b )

(define-fmath-instruction 'FADDs 16#41 )
(define-fmath-instruction 'FADDd 16#42 )
(define-fmath-instruction 'FADDx 16#43 )
(define-fmath-instruction 'FSUBs 16#45 )
(define-fmath-instruction 'FSUBd 16#46 )
(define-fmath-instruction 'FSUBx 16#47 )
(define-fmath-instruction 'FMULs 16#49 )
(define-fmath-instruction 'FMULd 16#4a )
(define-fmath-instruction 'FMULx 16#4b )
(define-fmath-instruction 'FDIVs 16#4d )
(define-fmath-instruction 'FDIVd 16#4e )
(define-fmath-instruction 'FDIVx 16#4f )

(define-fcmp-instruction 'FCMPs 16#51 )
(define-fcmp-instruction 'FCMPd 16#52 )
(define-fcmp-instruction 'FCMPx 16#53 )
(define-fcmp-instruction 'FCMPEs 16#55 )
(define-fcmp-instruction 'FCMPEd 16#56 )
(define-fcmp-instruction 'FCMPEx 16#57 )

%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%
%%      25. POP count
(define-machine-code 'POPC   4 2#101110 'type-2)
%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%
%%      25. Read State Register Instructions.
(define-machine-code 'RDY    4 2#101000 'type-2)
(define-machine-code 'RDPSR  4 2#101001 'type-2)
(define-machine-code 'RDWIM  4 2#101010 'type-2)
(define-machine-code 'RDTBR  4 2#101011 'type-2)
%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%
%%      26. Write State Register Instructions.
(define-machine-code 'WRY    4 2#110000 'type-2)
(define-machine-code 'WRPSR  4 2#110001 'type-2)
(define-machine-code 'WRWIM  4 2#110010 'type-2)
(define-machine-code 'WRTBR  4 2#110011 'type-2)

%%||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%%
%%      Pseudo instructions
%%
(de def-pseudo-instr (name length expantion)
  (put name '*code-length length)
  (put name '*pseudo-expantion expantion))

(def-pseudo-instr 'nop 4  '(lambda(name) '((sethi 0 g0))))
(def-pseudo-instr 'oddalign 0 '(lambda (name) (if (eq currentoffset*
						(wand currentoffset* (wnot 7)))
						 NIL
						'((sethi 0 g0)))))
(def-pseudo-instr 'cmp 4  `(lambda(name rs1 rs2) `((subcc ,rs1 ,rs2 g0))))
(def-pseudo-instr 'tst 4  `(lambda(name rs1 rs2) `((orcc ,rs1 ,rs2 g0))))
(def-pseudo-instr 'ret 4  `(lambda(name) `((jmpl (i7 8) g0))))
(def-pseudo-instr 'retl 4 `(lambda(name) `((jmpl (o7 8) g0))))
(def-pseudo-instr 'set 4  `(lambda(name val rd) `((or  g0 ,val ,rd))))
(def-pseudo-instr 'not 4  `(lambda(name rd) `((xnor  ,rd g0 ,rd))))
(def-pseudo-instr 'neg 4  `(lambda(name rs rd) `((sub  g0 ,rs ,rd))))
(def-pseudo-instr 'inc 4  `(lambda(name rd)    `((add ,rd 1 ,rd))))
(def-pseudo-instr 'inccc 4 `(lambda(name rd)   `((addcc ,rd 1 ,rd))))
(def-pseudo-instr 'dec 4   `(lambda(name rd)   `((sub ,rd 1 ,rd))))
(def-pseudo-instr 'deccc 4 `(lambda(name rd)   `((subcc ,rd 1 ,rd))))
(def-pseudo-instr 'btst 4  `(lambda(name rs1 rs2) `((andcc ,rs1 ,rs2 g0))))
(def-pseudo-instr 'bset 4  `(lambda(name rs rd) `((or ,rd ,rs ,rd))))
(def-pseudo-instr 'bclr 4  `(lambda(name rs rd) `((andn ,rd ,rs ,rd))))
(def-pseudo-instr 'btog 4 `(lambda(name rs rd)  `((xor ,rd ,rs ,rd))))
(def-pseudo-instr 'clr 4
	  `(lambda(name adr)(cond ((sparc-register-p adr) `((or g0 g0 ,adr)))
				   (t    `((st g0 ,adr))))))
(def-pseudo-instr 'clrb 4  `(lambda(name adr)   `((stb g0 ,adr))))
(def-pseudo-instr 'clrb 4  `(lambda(name adr)   `((stb g0 ,adr))))
(def-pseudo-instr 'mov 4   `(lambda(name rs rd) `((or g0 ,rs ,rd))))
%%
%%      move-relocatable-object pseudo instructions
%%
(def-pseudo-instr 'move-relocatable-object 8
  `(lambda(name src reg)
	  (cond (*writingfaslfile
	      `((word ,src)             %% when writing to faslfile, generate
	        (relocatable-movex ,reg)))      %% special code for relocation.
	   (t `((sethi (hi ,src) ,reg)  %% otherwise normal code.
	        (add ,reg (lo ,src) ,reg))))))%%

(def-pseudo-instr 'move-relocatable-object-and-load 8
  `(lambda(name src reg)
	  (cond (*writingfaslfile
	      `((word ,src)              %% when writing to faslfile, generate
	        (relocatable-movex-l ,reg)))   %% special code for relocation.
	   (t `((sethi (hi ,src) ,reg)  %% otherwise normal code.
	        (ld (displacement ,reg (lo ,src)) ,reg))))))%%

(def-pseudo-instr 'move-relocatable-object-and-store 8
  `(lambda(name src reg cont)
	  (cond (*writingfaslfile
	      `((word ,src)               %% when writing to faslfile, generate
	        (relocatable-movex-s ,reg ,cont))) %% code for relocation.
	   (t `((sethi (hi ,src) ,reg)  %% otherwise normal code.
	        (st ,cont (displacement ,reg (lo ,src))))))))%%

(define-machine-code 'RELOCATABLE-MOVEX 4 nil 'relocatable-movex-gen)

(de relocatable-movex-gen (name reg)
  (let ((regcode (get-register-code reg)) (mov-instr 16#80002))
       (store-relocatable-movex         % add instr, i=1 
	    (wplus2 (wshift mov-instr 12)
	            (wplus2  (wshift regcode 25) (wshift regcode 14))))))

(define-machine-code 'RELOCATABLE-MOVEX-s 4 nil 'relocatable-movex-s-gen)

(de relocatable-movex-s-gen (name reg cont)
  (let ((regcode (get-register-code reg))
        (regcode1 (get-register-code cont))
        (mov-instr 16#c0202))    % st with i=1
       (store-relocatable-movex 
	    (wplus2 (wshift mov-instr 12)
	            (wplus2  (wshift regcode1 25) (wshift regcode 14))))))

(define-machine-code 'RELOCATABLE-MOVEX-l 4 nil 'relocatable-movex-l-gen)

(de relocatable-movex-l-gen (name reg)
  (let ((regcode (get-register-code reg)) (mov-instr 16#C0002))
       (store-relocatable-movex      % load instr with i=1
	    (wplus2 (wshift mov-instr 12)
	            (wplus2  (wshift regcode 25) (wshift regcode 14))))))

 %
 % Basic Predicates
 %

(de sparc-immediate-p(r)
  (cond ((fixp r) t)
	((stringp r) t)
	((idp r)  (cond ((get r '*register-code) nil)
	                ((get r '*f-register-code) nil)
	                ((get r '*g-register-code) nil)
	                 (t     t)))  %% label
	(t nil)))

(de sparc-relocatable-immediate-p (r)  (eqcar r 'IMMEDIATE))

(de sparc-short-immediate-p (r)
	  (and (fixp r) (greaterp r -4097) (lessp r 4096)))

(de sparc-huge-displacement-p (r)
	   (and (eqcar r 'displacement)
                (numberp (caddr r))
		(or (greaterp (caddr r) 4095) (lessp (caddr r) -4096))))

(de sparc-deferred-p (r)    (eqcar r 'deferred))

(de sparc-indirect-p (r)
  (and (pairp r)
       (or (sparc-register-p (car r))
	   (eq (car r) 'deferred)
	   (eq (car r) 'displacement))))

(de sparc-double-indirect-p (r)
   (and (eqcar r 'deferred) (sparc-indirect-p (cadr r))))

(de sparc-fluid-p (r)
	  (and (pairp r) (member (car r)  '(!$fluid !$global fluid global))))

(de sparc-absolute-p (r)
	  (or (eqcar r 'quote) (stringp r)))

(de sparc-foreignentry-p (r)
	  (eqcar r 'foreignentry))
