%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           sun4-lap.sl
% Title:          SPARC PSL assembler                                          
% Author:         Tsuyoshi Yamamoto
% Created:        19 May 1988
% Status:         Experimental
% Mode:           Lisp
% Package:        Compiler
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Revisions:
%
% 07-feb-91  (Winfried Neun)
%  added odd double word alignment of floats
% 02-jan-90 (Winfried Neun)
%  Included immediatequote stuff
% 07-Aug-88 (Julian Padget)
%  Added back in code from sun-lap.sl to deal with IDLOC's in DEPOSITWORD-
%  EXPRESSION and defined the function DEPOSITWORDIDNUMBER.  Took note of 
%  TY's warning about arithmetic.  Also fixed misspelling in laleloffsets*.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       Note:
%
%       1. All arithmetics in this file must be **WORD ARITHMETIC**.
%          When you modify some part of this file, generic arithmetic functions
%          may cause nasty trouble if you load nbig library.
%
%       2. In PSL/SPARC, WSHIFT does LOGICAL left/right shift.
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       Part I
%
%
%%%%%%%%%(COMPILETIME (remprop 'mkitem 'pass-1-macro)) % for cross compilation
(compiletime (load fasl-decls common))


(global '(*writingasmfile *asmpass* *outputmode* *object*))

(fluid '(labeloffsets!* currentoffset* codesize* codebase* entries*
                        forwardinternalreferences* lapreturnvalue*
                        *writingfaslfile initoffset* *pgwd *pwrds
                        label-alist*))

(compiletime
  (progn (ds labelp (x) (atom x))
         nil))

(loadtime
  (progn (setq !*pwrds t)
         nil))

(fluid '(&links-present& &allocated-infmask&))

(de lap (u)
  (prog (labeloffsets!* lapreturnvalue!* entries!*)
        (unless !*writingfaslfile
          (setq currentoffset!* 0)
          (setq codebase!* (gtbps 0)))

        (setf &links-present& (or (look-for-links u)
                                  (assoc '*Id-apply u)
                                  (assoc '*call u)))

        (setq &allocated-infmask& NIL) % (frame 13) is f8000000 ?
        (setq u (pass1lap (&fillframeholes u)))
        (setq u (lapoptreformjumps u))
    
        (when !*pgwd
          (foreach x in u do (if (atom x)
                     (prin2 x)
                     (printf "          %p%n" x))))

        %
        %       sparc-pass-1 calculate codesize and fixup label address
        %
        (sparc-pass-1 u)

        (unless *writingfaslfile
                (unless (eq nextbps (wand nextbps (wnot 7))) (gtbps 1))
          	(setq codebase* (gtbps (wquotient (wplus2 codesize!* 3) 4))))
        %
        %       sparc-pass-2 generates machine code
        %
        (sparc-pass-2 u)                %% exec sparc assembler

        (defineentries)
        (when (and (not *writingfaslfile) !*pwrds)
          (errorprintf "*** %p: base %o, length %d bytes" 
                       (foreach x in entries!* collect (first (car x))) 
                       codebase!* codesize!*))
        (return (if lapreturnvalue!* (mkitem 20 lapreturnvalue!*) NIL))))

(de look-for-links (u)
       (cond ((null u) nil)
             ((atom (car u)) (look-for-links (cdr u)))
             ((eq (caar u) '*linke)
                  (cond ((flagp (caddar u) 'foreignfunction) t)
                         (t (look-for-links (cdr u)))))
             ((eq (caar u) '*link)
                  (cond
                    ((eq  (cadar u) 'catchsetup) t) % is opencode
                         % but problems will arise in throw if no
                         % allocation takes place
                    ((memq (cadar u) '(fast-codeapply fast-idapply)) t)
                    ((get (cadar u) 'opencode) (look-for-links (cdr u)))
                    (t t)))
             (t (look-for-links (cdr u)))
)      )
 
(de saveentry (x)


  (cond ((equal (second x) '!*!*!*code!*!*pointer!*!*!*) 
         (setq lapreturnvalue!* (if *writingfaslfile
                 currentoffset!*
                 (wplus2 codebase!* currentoffset!*))))
        ((not *writingfaslfile)
         (setq entries!* (cons (cons (rest x) currentoffset!*) entries!*))
         (unless lapreturnvalue!*
           (setq lapreturnvalue!* (wplus2 codebase!* currentoffset!*))))

        ((equal (second x) '!*!*fasl!*!*initcode!*!*) 
         (setq initoffset!* currentoffset!*))

       ((flagp (second x) 'internalfunction) 
         (put (second x) 'internalentryoffset currentoffset!*))

        (t %(and *writingfaslfile
                %(put (second x) 'internalentryoffset currentoffset!*))
           (findidnumber (second x)) 
           (dfprintfasl (list 'putentry (mkquote (second x)) 
                         (mkquote (third x)) currentoffset!*)))))
%
%       Adjust currentoffset!* to be word aligned.
%               T.Y.
%
(de force-dword-aligned ()
  (cond ((null (zerop (wand currentoffset!* 7)))
         (if (equal *asmpass* 2)
             (depositbyte 0)
             (setq currentoffset!* (iadd1 currentoffset!*)))
         (force-dword-aligned))))

(de force-word-aligned ()
  (cond ((null (zerop (wand currentoffset!* 3)))
         (if (equal *asmpass* 2)
             (depositbyte 0)
             (setq currentoffset!* (iadd1 currentoffset!*)))
         (force-word-aligned))))

(de defineentries ()
  (foreach x in entries!* do 
           (putd (first (car x)) (second (car x)) 
                 (mkcode (wplus2 codebase!* (cdr x))))))
%
%       (Comment for SPARC architecture.)
%
%       Deposit string must be word alligned at its end.
%                                               May 19, 1988 T.Yamamto
%(de depositstring (x)
%  (prog (y)
%        (setq x (second x))
%        (setq y (size x))      %% In PSL, why the function size returns
%                               %% string length minus 1?
%                               %% I think it is bad custom of PSL.
%        (for (from i 0 y 1)
%            (do (depositbyte (indx x i))))
%        (depositbyte 0)                %% At least 1 null byte to indicate end of
%                               %% string.
%                               %% long word allignment is required for SPARC.
%       (setq y (remainder (wplus2 2 y) 4))
%       (cond ((zerop y)
%              nil)
%             (t
%              (for (from i 0 (sub1 y))
%                   (do (depositbyte 0)))))))
%
%(put 'string 'instructiondepositfunction 'depositstring)

(de labeloffset (x)
  (cond ((or (eq (tag x) 160) (codep x)) % TEMPPPPPP
         (if *writingfaslfile
             (inf x)
             (wdifference (inf x) codebase*)))
 
        ((setq x (assoc x label-alist*))
         (cdr x))
        (t (print-error-status)
           (stderror (bldmsg "Internal entry forward reference: %r" x)))))


(de depositfloat (x)
  % this will not work in cross-assembly                                   
  (progn (setq x (fltinf (second x)))
         (depositword (floathighorder x))
         (depositword (floatloworder x))))

%(put 'float 'instructiondepositfunction 'depositfloat)

(de depositbyte (x)
  (setf (byte (wplus2 codebase!* currentoffset!*) 0) x)
  (updatebittable 1 0)
  (setq currentoffset!* (iadd1 currentoffset!*)))

(de deposithalfword (x)
  (setf (halfword (wplus2 codebase!* currentoffset!*) 0) x)
  (updatebittable 2 0)
  (setq currentoffset!* (wplus2 currentoffset!* 2)))

(de depositxword (x) % BOOT
 (if (negintp x)
  (progn
   (putmem (wplus2 codebase!* currentoffset!*) -1)
   (putmem (wplus2 4(wplus2 codebase!* currentoffset!*)) x)
   )
  (putmem (wplus2 codebase!* currentoffset!*) (wshift (tag x) 59))
  (putmem (wplus2 4(wplus2 codebase!* currentoffset!*)) (inf x))
  )
  (updatebittable 8 0)
  (setq currentoffset!* (wplus2 currentoffset!* 8)))

(de depositword (x)
  (putmem (wplus2 codebase!* currentoffset!*) x)
  (updatebittable 4 0)
  (setq currentoffset!* (wplus2 currentoffset!* 4)))

(de depositinstruction (x)
  (depositword  x))

%(depositbyte (wshift x -24))
  %(depositbyte (wand 16#ff (wshift x -16)))
  %(depositbyte (wand 16#ff (wshift x -8)))
  %(depositbyte (wand 16#ff x)))

(de deposit-relocated-word (offset)
  % Given an OFFSET from CODEBASE*, deposit a word containing the
  % absolute address of that offset.
  (putmem (wplus2 codebase* currentoffset*)
          (wplus2 offset (if *writingfaslfile 0 codebase*)))
  (updatebittable 4 reloc-word)
  (setq currentoffset* (wplus2 currentoffset* 4)))
  
(de depositwordexpression (x)
  % Only limited expressions now handled
  (let (y)
  (cond ((fixp x) (depositxword (int2sys x)))
        ((stringp x)
         (deposit-relocated-word (labeloffset x)))
        ((memq  (first x) '(!$fluid !$global fluid global))
         (depositvaluecelllocation (second x)))
        ((equal (first x) 'mkitem)
         (deposititem (second x) (third x)))
        ((equal (first x) 'entry)
         (depositfunctioncelllocation (second x)))
        ((equal (first x) 'foreignentry)
         (depositfunctioncelllocation (second x)))
        ((equal (first x) 'internalentry)
         (cond ((forward-internal-reference-p (cadr x))
                (push (list 'WORD (cadr x) currentoffset!*)
                      forwardinternalreferences!*)
                (depositword 0))
               (t
                (setq y (get (second x) 'internalentryoffset))
                (if y
                    (deposit-relocated-word y)
                    (stderror (bldmsg "Internal entry forward reference: %r"
                                      x))))))
        ((equal (first x) 'immediate)
         (cond ((and (pairp (second x)) 
                     (equal (first (second x)) 'mkitem))
                     (depositwordexpression (second x)))
               ((setq y (assoc (second x) label-alist*))
                (deposit-relocated-word (cdr y)))
               ((and (pairp (second x))
                     (equal (first (second x)) 'entry))
                (depositfunctioncelllocation (second (second x))))
               ((and (pairp (second x)))
                     (equal (first (second x)) 'idloc)
                (depositwordidnumber (second (second x))))
               (t
                (print-error-status)
                (stderror (bldmsg "Undefined label: %w" (second x))))))
 
        ((setq y (wconstevaluable x))
         (depositxword (int2sys y)))
 
        (t (print-error-status)
           (stderror (bldmsg "Expression too complicated %r" x))))))
 
(de depositwordidnumber (x) 
  (cond
    ((or (not *writingfaslfile) (wleq (idinf x) 128)) 
     (depositword (idinf X)))
    ('else 
      (putword (wplus2 codebase* currentoffset*) 0 
               (makerelocword reloc-id-number (findidnumber x))) 
      (setq currentoffset* (wplus2 currentoffset* 4)) 
      (updatebittable 4 reloc-word))))


(de deposithalfwordexpression (x)
  (prog (y)
        (return (cond ((fixp x) (deposithalfword x))
                      ((and (equal (first x) 'difference) 
                        (labelp (second x)) (labelp (third x)))
                       (deposithalfword 
                        (wdifference (labeloffset (second x)) 
                         (labeloffset (third x)))))
                      ((equal (first x) 'idloc) 
                       (deposithalfwordidnumber (second x)))
                      ((setq y (wconstevaluable x)) (deposithalfword y))
                      (t (print-error-status)
                         (stderror 
                          (bldmsg "Halfword expression too complicated %r" 
                           x)))))))

(de deposititem (tagpart infpart)
  (if (not *writingfaslfile)
    (depositxword (mkitem tagpart 
                         (cond ((labelp infpart) 
                                (wplus2 codebase!* (labeloffset infpart)))
                               ((equal (first infpart) 'idloc) 
                                (idinf (second infpart)))
                               (t 
                                (print-error-status)
                                (stderror 
                                 (bldmsg "Unknown inf in MkItem %r" 
                                  infpart))))))
    (progn (cond ((labelp infpart)  %BOOT
                  (putmem (wplus2 4 (wplus2 codebase!* currentoffset!*))
			(labeloffset infpart))
                  (putmem (wplus2 codebase!* currentoffset!*)
				% RELOC_CODE_OFFSET = 0
                                                                            
                         (wshift tagpart  27)))
                 ((equal (first infpart) 'idloc) 
                  (putmem (wplus2 4 (wplus2 codebase!* currentoffset!*))
		       (makerelocinf reloc-id-number
				(findidnumber (second infpart))))
                  (putmem (wplus2 codebase!* currentoffset!*)
                        (wshift tagpart  27)))
                 (t (print-error-status)
                    (stderror (bldmsg "Unknown inf in MkItem %r" infpart))))
           (setq currentoffset!* (wplus2 currentoffset!* 8))
           (updatebittable 8 reloc-inf))))

(de deposithalfwordidnumber (x)
  (if (or (not *writingfaslfile) (ileq (idinf x) 128))
    (deposithalfword (idinf x))
    (progn (setf (halfword (wplus2 codebase!* currentoffset!*) 0) 
                 (makerelochalfword reloc-id-number
                  (findidnumber x)))
           (setq currentoffset!* (wplus2 currentoffset!* 2))
           (updatebittable 2 reloc-right-half))))

(de systemfaslfixup ()
  (prog (x type label place saved-currentoffset)
        (setq x (iremainder currentoffset!* 4))
        (while (igreaterp x 0)
          (depositbyte 0)
          (setq x (isub1 x)))
        (setq saved-currentoffset currentoffset!*)
        (while (not (null forwardinternalreferences!*))
          (setq type (caar forwardinternalreferences!*))
          (setq label (cadar forwardinternalreferences!*))
          (setq place (caddar forwardinternalreferences!*))
          (if (null (get label 'internalentryoffset))
            (errorprintf "*** %r not defined in this module, call is incorrect"
                         (cadar forwardinternalreferences!*))
            (cond ((eq type 'CALL)
                   (setq currentoffset!* place)
                   (call-gen 'CALL (list 'INTERNALENTRY label)))
                  ((get type '*bcc) % a bcc instruction ZI
                   (setq currentoffset!* place)
                   (bcc-gen type (list 'INTERNALENTRY label)))
                  ((eq type 'WORD)
                   (setq currentoffset!* place)
                   (store-word (get label 'internalentryoffset)))
                  (t
                   (print-error-status)
                   (error 0 (list "Strange forward internal reference: "
                                  (car forwardinternalreferences!*))))))

          (setq forwardinternalreferences!* 
                (cdr forwardinternalreferences!*)))
        (mapobl 'remove-ieo-property)
        (setq currentoffset!* saved-currentoffset)
        ))

(de remove-ieo-property (y)
  (remprop y 'internalentryoffset))

(de makerelocword (reloctag relocinf)
  (wplus2 (wshift reloctag 30) (field relocinf 2 30)))

(de makerelocinf (reloctag relocinf)
  (wplus2 (wshift reloctag 22) (field relocinf 10 22)))

(de makerelochalfword (reloctag relocinf)
  (wplus2 (wshift reloctag 14) (field relocinf 18 14)))
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%%      These 2 relocatable cell are newly introduced for SPARC architecture.
%%
%%      First,  (movex relocatable-value, reg)
%%      This case, 2 cells are relocated at the time.
%%
(define-constant reloc-movex 1)
(define-constant reloc-call 2)

(de deposit-relocatable-movex (code)
                                 % code is (add r,0,r) where r is the dst.
  (setf (byte (wplus2 codebase!* currentoffset!*) 0)
        (wshift code -24))
  (setf (byte (wplus2 codebase!* currentoffset!*) 1)
        (wand 16#ff (wshift code -16)))
  (setf (byte (wplus2 codebase!* currentoffset!*) 2)
        (wand 16#ff (wshift code -8)))
  (setf (byte (wplus2 codebase!* currentoffset!*) 3)
        (wand 16#ff code))
  (updatebittable 1 reloc-word)
  (updatebittable 3 reloc-movex)  %% reloc-movex is defined in sys-consts.sl
  (setq currentoffset!* (wplus2 currentoffset!* 4)))
%%
%%      Second, (call symbol)
%%
(de deposit-relocatable-call (x)
  (setf (getmem (wplus2 codebase* currentoffset*))
        (makerelocword reloc-function-cell (findidnumber x)))
  (updatebittable 1 reloc-word)
  (updatebittable 3 reloc-call)
  (setf currentoffset* (wplus2 currentoffset* 4)))
%%
%%      Third , (branchoncondition symbol)
%%
(de deposit-relocatable-bcc (name x)
  (setf (getmem (wplus2 codebase* currentoffset*))
        (makerelocword reloc-function-cell 
                (wor (wshift (get name '*bcc) 24) 
                     (wor (wshift (get name '*op) 18)
                          (wor (wshift (get name 'annul) 17)
                               (findidnumber x)))))) % 16 bits for idnumber
  (updatebittable 1 reloc-word)  % faslin has to differentiate  
  (updatebittable 3 reloc-call)
  (setf currentoffset* (wplus2 currentoffset* 4)))
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%%      Part II.
%%
%%
(setq *outputmode* t)   % normal bps output.
%%
%%      *outputmode controls hoe the result of assemble going.
%%
%%      T:      result is going to BSP space directly
%%      NIL:    result is accumulated into *object*
%%
%%
%%
%%      Pass-1 calculates codesize and fixup relative address of labels
%%

(fluid '(*current-asmlist *current-form))
%
%       print-error-status is usefull for debuging.
%
(de print-error-status ()
  (prog (oldwrs)
        (setq oldwrs (wrs nil))
        (terpri)
        (princ "Error signaled at asm-form: ")
        (prin1 *current-form)
        (terpri)
        (print "Current-asm-list: ")
        (prin1 *current-asmlist)
        (terpri)
        (wrs oldwrs)))


(de sparc-pass-1 (asm-list)
  (prog (saved-currentoffset)
        (setq *current-asmlist asm-list)        % Just for debug. T.Y.
        (setq *asmpass* 1)
        (setq *object* nil)
        (setq saved-currentoffset currentoffset!*)
        (setq label-alist* nil) %% reset label alist.
        (for (in u asm-list)
             (do (setq *current-form u)         % Just for debug.
                 (asm-to-code u)))
        (setq codesize* (wdifference currentoffset* saved-currentoffset))
        (setq currentoffset* saved-currentoffset)))
%%
%%      Pass-2 converts assember into machine code
%%
(de sparc-pass-2 (asm-list)
  (setq *current-asmlist asm-list)      % Just for debug. T.Y.
  (setq *asmpass* 2)
  (setq *object* nil)
  (for (in u asm-list)
       (do (setq *current-form u)       % Just for debug. T.Y.
           (asm-to-code u)))
  (setq *object* (reverse *object*)))


(de asm-to-code (as)
 (prog (result)
  (cond ((atom as)
        (if (equal as "double-word-alignment") (force-dword-aligned)
	 (force-word-aligned)   % LABELs must word aligned!!
         (cond ((equal *asmpass* 1)
                (cond ((assoc as label-alist*)
                       (print-error-status)
                       (error 0 (list "Label defined already:" as)))
                      (t
                       (setq label-alist* (cons (cons as currentoffset!*)
                                                label-alist*))))))))
        ((eq (car as) 'oddalign)
                (force-word-aligned)
		(if (eq currentoffset* (wand currentoffset* (wnot 7)))
			(asm-to-code '(nop))))
        ((get (car as) 'asmexpandop)            %% PSL pseudo-ops
         (for (in u (apply (get (car as) 'asmexpandop) (list as)))
              (do (asm-to-code u))))
        ((get (car as) '*pseudo-expantion)      %% SPARC pseudo-ops
         (for (in u (apply (get (car as) '*pseudo-expantion) as))
               (do (asm-to-code u))))
        ((get (car as) '*code-gen)              %% SPARC native instructions
         (apply (get (car as) '*code-gen)
                (expand-operand as)))
        (t
         (print-error-status)
         (error 0 (list "Unknown sparc instrction:" as))))))

(de code-print (x)
  (for (from i 0 7)
       (do (printf "%x" (wand (wshift x (wtimes2 4 (wdifference i 7))) 15)))))

(de pseudo-expantion (as)
  (cond ((atom as) as)          %% label
        ((get (car as) '*pseudo-expantion)
         (pseudo-expantion (apply (get (car as) '*pseudo-expantion)
                             as)))
        (t        as)))

%%
%%      store-xxxx is used for saving code into BPS space. It saves the
%%      code to *object* when *outputmode* is nil.
%%
(de store-word (word)
  (cond ((equal *asmpass* 2)    %% real code is output only when pass is 2.
         (if *outputmode*
             (depositinstruction word)
             (setq *object*
                   (append (word-to-byte word) *object*))))
        (t
         (setq currentoffset!* (wplus2 currentoffset!* 4)))))

(de store-sethi-word (word)
  (cond ((equal *asmpass* 2)    %% real code is output only when pass is 2.
         (if *outputmode*
	    (progn 
             (putmem (wplus2 codebase!* currentoffset!*) word)
	     (setq bittableoffset* (wplus2 bittableoffset* -4))
% hack
             (updatebittable 4 reloc-right-half)
             (setq currentoffset!* (wplus2 currentoffset!* 0)))
	     (setq *object*
                   (append (word-to-byte word) *object*))))
        (t
         (setq currentoffset!* (wplus2 currentoffset!* 0)))))

(de store-halfword (hword)
  (cond ((equal *asmpass* 2)
         (if *outputmode*
             (deposithalfword hword)
             (setq *object*
                   (append (halfword-to-byte hword) *object*))))
        (t (setq currentoffset!* (wplus2 currentoffset!* 2)))))

(de store-byte (byte)
  (cond ((equal *asmpass* 2)
         (if *outputmode*
             (depositbyte byte)
             (setq *object*
                   (cons byte *object*))))
        (t  (setq currentoffset!* (add1 currentoffset!*)))))

(de store-xword-expression (word)

  (force-dword-aligned)          % W.N.

  (cond ((equal *asmpass* 2) (depositwordexpression word))
        (t (setq currentoffset!* (wplus2 currentoffset!* 8)))))

(de store-word-expression (word)

  (force-word-aligned)          % T.Y.

  (cond ((equal *asmpass* 2)
         (if *outputmode*
             (depositwordexpression word)
             (setq *object* (cons (list 'word word) *object*))))
        (t
         (setq currentoffset!* (wplus2 currentoffset!* 4)))))

(de store-double (word)

  (force-word-aligned)          % T.Y.

  (cond ((equal *asmpass* 2)
         (if *outputmode*
             (depositfloat `(FLOAT ,word))
             (setq *object* (cons (list 'double word) *object*))))
        (t
         (setq currentoffset!* (wplus2 currentoffset!* 8)))))

(de store-halfword-expression (hword)
  (cond ((equal *asmpass* 2)
         (if *outputmode*
             (deposithalfwordexpression hword)
             (setq *object* (cons (list 'halfword  hword) *object*))))
        (t
         (setq currentoffset!* (wplus2 currentoffset!* 2)))))

(de store-relocatable-movex (code)
  (cond ((equal *asmpass* 2)
         (if *outputmode*
             (deposit-relocatable-movex code)
             (setq *object* (cons (list 'movex  code) *object*))))
        (t
         (setq currentoffset!* (wplus2 currentoffset!* 4)))))


(de store-relocatable-bcc (name label)
  (cond ((equal *asmpass* 2)
           (if *outputmode*
             (deposit-relocatable-bcc name label)
             (setq *object* (cons (list name label) *object*))))
        (t  (setq currentoffset!* (wplus2 currentoffset!* 4)))))

(de store-relocatable-call (label)
  (cond ((equal *asmpass* 2)
         (if *outputmode*
              (deposit-relocatable-call label)
              (setq *object* (cons (list 'call  label) *object*))))
        (t  (setq currentoffset!* (wplus2 currentoffset!* 4)))))

(de word-to-byte (w)
  (list (wand 255 w)
        (wand 255 (wshift w -8))
        (wand 255 (wshift w -16))
        (wand 255 (wshift w -24))))

(de halfword-to-byte (w)
  (list (wand 255 w)
        (wand 255 (wshift w -8))))
