/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         bpsheap.c
% Description:  Code to dynamically set up bps and heap structures
% Author:       RAM, HP/FSD
% Created:      9-Mar-84
% Modified:
% Mode:         Text
% Package:
%
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 11-Aug-88 (Julian Padget)
%  Added initialization of BPSLOWERBOUND in setupbps().
% 07-Apr-87 (Harold Carr & Leigh Stoller)
%  Put in error checking to ensure that the memory pointers will fit in
%   info field of the lisp item.
% 21-Dec-86 (Leigh Stoller)
%  Added allocatemorebps function, called from try-other-bps-spaces in
%   allocators.sl.
% 18-Dec-86 (Leigh Stoller)
%  Changed to newer model. Bps is now defined in bps.c so that unexec can
%  alter the text/data boundry. Took out code that allowed command line
%  modification of bpssize. (Now set in the Makefile). Added setupbps()
%  that initialzes NEXTBPS and lastbps.
% 20-Sep-86 (Leigh Stoller)
%  Removed assembler alias statements because they are not portable. Instead,
%  a sed script will be used to convert the _variables of C to VARIABLES of
%  PSL.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>


/* Use 1 if using compacting collector ($pxnk/compact-gc.sl).
   Use 2 if using copying collector ($pnk/copying-gc.sl).
   Be sure to update $pxnk/load-psl.sl to include correct collector. */

#define MINSIZE       10000000  /* Default total in number of bytes. */
#define MALLOCSIZE     500000   /* Default size for OS support functions. */
#define EXTRABPSSIZE   100000   /* Minimum amount to increase bps by. */
#define MINIMUMHEAPADD 20000    /* Minimum amount to increase heap by */


#ifndef BPSSIZE
#define BPSSIZE          800000    /* Default bps size in number of bytes */
#endif

char *  imagefile ;
long long     max_image_size;
long long     oldbreakvalue;

extern long long  alreadysetupbpsandheap;
extern long long  HASHTABLE;
extern long long  VIVAREA;
extern char bps[];
extern long long  LASTBPS;
extern long long  NEXTBPS;
extern long long  BPSLOWERBOUND;

extern long long  HEAPLOWERBOUND;
extern long long  HEAPUPPERBOUND;
extern long long  HEAPLAST;
extern long long  HEAPTRAPBOUND;

extern long long  OLDHEAPUPPERBOUND;
extern long long  OLDHEAPLOWERBOUND;
extern long long  OLDHEAPLAST;
extern long long  OLDHEAPTRAPBOUND;

extern long long  SYMVAL;

/* Write this ourselves to keep from including half the math library */
static long long power(x, n)
     int x, n;
{
  long long i, p;

  p = 1;
  for (i = 1; i <= n; ++i)
    p = p * x;
  return(p);
}

setupbpsandheap(argc,argv)
     int argc;
     char *argv[];
{ long long ohl,ohtb,ohlb,ohub,hl,htb,hlb,hub;
  int memset = 0; int feder = 200000;
  FILE * imago;
  long long  headerword [9];
  int    i, bpssize; long long heapsize, total;  long long mallocsize;
  long long    current_size_in_bytes, heapsize_in_bytes;
  double bpspercent, heappercent;
  char   *argp, *scanptr, *scanformat;
  int ii1,ii2,ii3,ii4,ii5,ii6,ii7,ii8,ii9,ii10,ii11;
  char * hugo[20];

  total        = MINSIZE;
  mallocsize    = MALLOCSIZE;

  for (i=1; i<argc-1; i++)
    {
      argp = argv[i];
      if (*argp++ == '-')
        {
          scanformat = "";
          switch (*argp++) {
            case 't': scanptr = (char *)&total;
                      memset = 1;
                      switch (*argp) {
                case 'x': scanformat = "%lx";
                      break;
                        case 'd': scanformat = "%ld";
                                  break;
              }
                      break;
            case 'm': scanptr = (char *)&mallocsize;
                      switch (*argp) {
                        case 'x': scanformat = "%x";
                                  break;
                        case 'd': scanformat = "%d";
                                  break;
                      }
                      break;
           case 'f': imagefile = argv[i+1]; break;
           case 'g': scanptr = (char *)&feder;
                     scanformat = "%d"; break;
          }
          if (*scanformat != 0)
            sscanf(argv[i+1],scanformat,scanptr);
        }
    }   /* end of for loop -- arg vector searched */

  /* insure valid values */
  if (total == 0) total = MINSIZE;

  if (mallocsize <= 0) mallocsize = MALLOCSIZE;



  /* Reserve some space for C's usr of io buffers, etc. By mallocing then
     freeing, the memory is sbrk'ed onto the image, but available for future
     calls to malloc, which will not need to call sbrk again. */

  for (i=0;i<20;i++)
        { hugo[i] = malloc(power(2,i) + 7);}

  external_user_homedir_string(); /* This is done by read-init-file */
  external_anyuser_homedir_string("hugo"); 

  bpssize = BPSSIZE;

  heapsize_in_bytes = total - bpssize;

  /* On systems in which the image does not start at address 0, this won't
     really allocate the full maximum, but close enough. */
  current_size_in_bytes = (((int) sbrk(0))<<5)>>5;
  max_image_size = power(2, 59); /* 1 more than allowable size */

  if ((heapsize_in_bytes + current_size_in_bytes) >= max_image_size) {
    heapsize_in_bytes = max_image_size - current_size_in_bytes;
    total = heapsize_in_bytes + bpssize;
    printf("Size requested will result in pointer values larger than\n");
    printf(" PSL items can handle. Will allocate maximum size instead.\n\n");
  }

  heapsize =(heapsize_in_bytes / 16) * 8;  /* insure full words */
  heappercent = ((float) (total - bpssize) / total) * 100.0;
  bpspercent  = ((float) bpssize / total) * 100.0;

  if (imagefile == NULL)
  { printf("Setting heap limit as follows:\n");
    printf("Total heap & bps space = %ld (%LX), bps = %.2f, heap = %.2f\n",
          total, total, bpspercent, heappercent);
  }

  setupbps();
  getheap(heapsize,feder);


  for (i=0;i<20;i++) { if (hugo[i] != (char *) NULL) free(hugo[i]);}

  if (imagefile == NULL)
  printf("bpssize = %d (%X), heapsize = %ld (%X)\nTotal image size = %ld (%X)\n",
          bpssize, bpssize,
          heapsize, heapsize,
          (int) sbrk(0), (int) sbrk(0));


   if (imagefile != NULL) {
        ohlb= OLDHEAPLOWERBOUND; ohub = OLDHEAPUPPERBOUND;
        ohl = OLDHEAPLAST;    ohtb = OLDHEAPTRAPBOUND;
        hlb = HEAPLOWERBOUND; hub = HEAPUPPERBOUND;
        hl  = HEAPLAST;       htb = HEAPTRAPBOUND;
    /* save the new values around restore of the old ones */

       printf("Loading image file: %s \n",imagefile);
       imago = fopen (imagefile,"r");
       if (imago == NULL) { perror ("error"); exit (-1); }
       ffread (headerword,8,4,imago);

       if (strncmp(headerword,datetag(),31))
                { printf(" Cannot start the image with this bpsl \n");
                  exit (-19); } 

       ffread (headerword,8,8,imago);
       ffread (&SYMVAL,1,headerword[0],imago);
       /* printf (" heaplowerbound = %x (new) %x (file)\n" HEAPLOWERBOUND,
                        headerword[6]); */
        if(headerword[7] > HEAPLOWERBOUND + feder/2) {
                printf (" Cant relocate the image"); exit (-3); }
        if(headerword[7] < HEAPLOWERBOUND - feder/2) {
                printf (" Cant relocate the image"); exit (-3); }
       feder = HEAPLOWERBOUND - headerword[7];
       ffread ((char *) headerword[7],1,headerword[1],imago);
       ffread (&VIVAREA,1,headerword[2],imago);
       ffread (&HASHTABLE,1,headerword[3],imago);
       ffread ((char *) BPSLOWERBOUND,1,headerword[4],imago);
       ffread ((char *) headerword[5],1,headerword[6],imago);
       fclose (imago);
           OLDHEAPLOWERBOUND = ohl -feder; 
           OLDHEAPUPPERBOUND = ohub -feder;
           OLDHEAPLAST  =  ohl -feder;   
           OLDHEAPTRAPBOUND = ohtb -feder;
           HEAPUPPERBOUND = hub -feder;
           HEAPTRAPBOUND = htb -feder; 
       return (4711);
     }
return (0);

}

int ffread(ptr,unit,nunits,fptr)
char * ptr;
int unit , nunits;
FILE* fptr;
  { if (fread(ptr,unit,nunits,fptr) == nunits) return (nunits);
    fprintf(stderr,"*** Error, file too short\n"); }


/* The current procedure is to convert the starting address of the char
   array defined in bps.c to an address and store it in NEXTBPS. A check
   is made to make sure that NEXTBPS falls on an even word boundry.
 */

setupbps()
{ BPSLOWERBOUND = ((long long)bps + 3) & ~3;
  NEXTBPS  = ((long long)bps + 3) & ~3;        /* Up to a multiple of 4. */
  LASTBPS  = ((long long)bps + BPSSIZE) & ~3;    /* Down to a multiple of 4. */
}


/* Allocate alternate bps space. Note: The use of sbrk(), and the fact that
   NEXTBPS is now greater than HEAPLAST means that unexec should be not be
   tried after this routine is called. The image would be huge.
 */
allocatemorebps()
{
  int current_size_in_bytes;
  int old_NEXTBPS = NEXTBPS;

  current_size_in_bytes = sbrk(0);

  if ((current_size_in_bytes + EXTRABPSSIZE) >= max_image_size)
    return(0);

  if (((int)sbrk(0)) % 2)      /* force to even word boundary*/
     NEXTBPS = (int)sbrk(1);

  NEXTBPS = (int)sbrk(EXTRABPSSIZE);   /* allocate extra BPS */
  if (NEXTBPS == -1) {
    NEXTBPS = old_NEXTBPS;
    return(0);
  }
  LASTBPS = NEXTBPS + EXTRABPSSIZE;
  return(EXTRABPSSIZE);   /* This will be a paramter later */
}


/* tag( getheap )
 */

getheap(heapsize,feder)
     long long heapsize; int feder;
{
  HEAPLOWERBOUND        = (long long )sbrk(2 * heapsize + feder );
                  /* allocate first heap */;
  
  if (HEAPLOWERBOUND == -1) {
    perror("GETHEAP");
    exit(-1);
  }
     HEAPLOWERBOUND += feder / 2;
     HEAPLOWERBOUND  = HEAPLOWERBOUND;
     HEAPUPPERBOUND  = HEAPLOWERBOUND + heapsize;
     HEAPLAST        = HEAPLOWERBOUND;
     HEAPTRAPBOUND   = HEAPUPPERBOUND -120;

     OLDHEAPLOWERBOUND = HEAPUPPERBOUND;
     OLDHEAPUPPERBOUND = OLDHEAPLOWERBOUND + heapsize;
     OLDHEAPLAST       = OLDHEAPLOWERBOUND;
     OLDHEAPTRAPBOUND  = OLDHEAPUPPERBOUND -120;

     oldbreakvalue = (long long)sbrk(0);
}


/* Tag( alterheapsize )
 */

alterheapsize(increment)

int increment;

{
/*
  alters the size of the heap by the specified increment.  Returns
  the increment if successful, otherwise returns 0.  May fai i
  the sbrk is unsuccessful or if the user tries to cut the heap back
  to nothing or the current break value does not match the old value.
  The latter case occurs when a malloc or sbrk has allocated space for
  some other software, in which case we cannot allocate any more space
  contiguously.

  Modifies both the heap and gcarray size.
  NOTE: a garbage collection should probably be performed before this
    routine is called.
  NOTE: only implemented for the one heap version on the 68000.
*/

  int heapsize;
  int current_size_in_bytes;

  /* assumes the current heap is the 'lower' one */
  int newbreakvalue;

  if ((long long) sbrk(0) != oldbreakvalue)  /* Non contiguous memory */
      {  printf(" unable to allocate %x %x\n",sbrk(0),oldbreakvalue);
        return(0); }

  current_size_in_bytes = ( (long long) sbrk(0) <<5) >>5;

  if ((current_size_in_bytes + 2* increment) >= max_image_size)
    return(-1);

  if ((int)sbrk(2 * increment) == -1)       /* the sbrk failed. */
     return(-2);

  newbreakvalue = (long long) sbrk(0);

  HEAPUPPERBOUND        = HEAPUPPERBOUND + increment ;
  HEAPTRAPBOUND         = HEAPUPPERBOUND - 120;
  OLDHEAPLOWERBOUND     = OLDHEAPLOWERBOUND + increment;
  OLDHEAPUPPERBOUND     = OLDHEAPUPPERBOUND + 2* increment ;
  OLDHEAPLAST           = OLDHEAPLOWERBOUND;
  OLDHEAPTRAPBOUND      = OLDHEAPUPPERBOUND -120;


  oldbreakvalue = newbreakvalue;
  return(increment);
}

unexec ()
  { return (BPSSIZE); }
