************************************************************************
*
* File : GETB00 - CAL coded support routines for CRAY PSL 3.4 (UNICOS)
*
* Author : Dana Dawson , Cray Research Inc
*          W. Neun ZIB Berlin
*
* Date :  14-May-86
*
* Copyright (C) Konrad Zuse Zentrum , Berlin 1987
************************************************************************
         MACRO
LABEL    JTIME     ADDR
LABEL    =         *
 
         CALLV     SECOND
 
REGS     IFA       REG,ADDR
          IFC       \ADDR\,NE,\S1\,1
           ADDR      S1
REGS     ELSE
          VARADD    A1,ADDR,USE=A5
          0,A1      S1
REGS     ENDIF
 
JTIME    ENDM
         IDENT     GETB00
         TITLE     ' GETB00 - return B00 address of calling routine'
*
*  GETB00 returns in S1 the return address (B00) of the calling routine.
*         This allows a subroutine to find out where it will return to.
*         The method is to follow the TRACEBACK trail back two levels to
*         get the previous value of B00.  Note that this is very depende
*         on the 'new' CFT calling sequence.
*
*         CALL from CFT:
*
*                  IB00 = GETB00()
*
GETB00   ENTER     NP=0,NB=0,NT=0    * No parameters, B, or T regs used.
*
         A1        B02               * Get base of my B reg save area.
         A1        3,A1              * Get base of previous B reg save a
         S1        1,A1              * S1 = ((B02) + 1) = previous B00.
*
         EXIT                        * Done.
*
*
         ENTRY VCOPY
*
* Copy a Vector for Portable Standard Lisp  Version CRAY 3.4
*
*       (vcopy addr-source addr-dest length)
*
* CAUTION: NO Check for OVERLAPPING is done ^^
*
VCOPY    A6   2
         A7   S3
         A0   A7-A6  % Special Case for a Pair
         A6   S1
         JAN  NOTTWO
         S1   0,A6
         S3   1,A6
         A7   S2
         0,A7 S1
         1,A7 S3
         S1   2       Return something defined
         J    B00
*
*
*                    A6=SOURCE ADDR.
NOTTWO   A7   S2     A7=DEST ADDR.
*
* as described in CAL Reference Manual
*
         A0   A6-A7        check if copy backward
         A5   S3           A5=LENGTH
         JAM  BACKW
         S3   -S3
         S0   S3
         JSP  ERROR
         S4   <6            MASK OF 6 1'BITS
         S3   #S3&S4        S3=(LENGTH-1) MOD 64
         A1   S3
         A1   A1+1
         J    LOOP1
LOOP     A0   A7          MAIN LOOP
         V2   V1
         ,A0,1 V2        ................
         A7   A7+A1     These instructions
         A1   D'64      should allow to sustain
LOOP1    VL   A1
         A0   A6        TWO Streams (ON X-MP)
         V1   ,A0,1      .................
         A6   A6+A1
         A5   A5-A1
         A0   A5
         JAN  LOOP       END  MAIN-LOOP
         A0   A7
         ,A0,1 V1
         J    B00        RETURN
 
*          A6=SOURCE ADDR.
*          A7=DEST ADDR.
*
* as described in CAL Reference Manual, but backward copy
*
*                  A5=LENGTH
BACKW    S3   -S3
         A7   A7+A5        source and dest to the end
         S0   S3
         A6   A6+A5        dto.
         JSP  ERROR
         S4   <6            MASK OF 6 1'BITS
         S3   #S3&S4        S3=(LENGTH-1) MOD 64
         A1   S3
         A1   A1+1
         A6   A6-A1         source and dest before last vector
         A7   A7-A1
         J    LOOP3
LOOP2    A0   A7          MAIN LOOP
         V2   V1
         ,A0,1 V2        ................
         A1   D'64      should allow to sustain
         A7   A7-A1     These instructions
         A6   A6-A1
LOOP3    VL   A1
         A0   A6        TWO Streams (ON X-MP)
         V1   ,A0,1      .................
         A5   A5-A1
         A0   A5
         JAN  LOOP2      END  MAIN-LOOP
         A0   A7
         ,A0,1 V1
         J    B00        RETURN
*
* ERROR!
*
ERROR    S1   T00        RETURN NIL^
         J    B00
*
         ENTRY VCLEAR
*
* Clear a Vector for Portable Standard Lisp  Version CRAY 3.4
*
*       (vclear dest length)
*
VCLEAR   A6   S1
*                  A6=dest   ADDR.
         A5   S2           A5=LENGTH
         S2   -S2
         S0   S2
         JSP  ERROR1
         S4   <6            MASK OF 6 1'BITS
         S2   #S2&S4        S3=(LENGTH-1) MOD 64
         A1   S2
         A1   A1+1
         J    LOOP1A
LOOPA    A1   D'64
LOOP1A   VL   A1
         A0   A6
         V1   0
         ,A0,1 V1
         A6   A6+A1
         A5   A5-A1
         A0   A5
         JAN  LOOPA
         J    B00        RETURN
*
* ERROR^
*
ERROR1   S1   T00        RETURN NIL^
         J    B00
*
NEWLEN  ENTER  NP=0,NB=0,NT=0     heap-manager contact
        SREGSAV,  S1
        SREGSAV+1, S2
        SREGSAV+2, S3
        CALLV  HCHECK%
        S0    S1
        JSZ   WEITER
        S3    17
        EXIT  NB=0,NT=0
WEITER  S1     SREGSAV,
        S2     SREGSAV+1,
        S3     SREGSAV+2,
        CALLV  NEWLEN%
        EXIT  NB=0,NT=0
*
SHRINK  ENTER  NP=0,NB=0,NT=0
        CALLV  SHRINK%
        EXIT  NB=0,NT=0
*
SREGSAV   BSS  3
        END
*
*         IDENT  CALRPV        supply an emergency stack  for reprieve process
*         ENTRY  CALRPV        needed for UNICOS 2.1 only
*         EXT    LISPRPV
*CALRPV   A3     0
*         A4     NOTSTCK
*         A6     NOTSTCK+260
*         B54    A6
*         R      LISPRPV
*NOTSTCK  BSS 300
*         END
 
*
         IDENT    NORPV
         EXT      LALLFUNC
NORPV    ENTER
*                     Turn off reprieve processing for abort signal
*
*  BUILD A CONTROLBLOC FOR EX
*
         a6        arglist
         S1        7                error exit
         S2        O'100000         SCTL_KIL
         S1        S1!S2
         0,A6      S1
         S2        P.LALLFUNC
         1,A6      S2
         S1        0
         2,A6      S1
         3,A6      S1
         4,A6      S1
         5,A6      S1
         SYSCALL   FC=$SIGNAL,ARG=A6
         JSZ       RAUS1
         S1        -1
RAUS1    EXIT
arglist  BSSZ      6
         END
*
*       GETLNG: get total length of living program
*
           IDENT GETLNG
           ENTRY GETLNG
GETLNG     A6   ARGLIST
           A5   B00
           3,A6 A5
           S1   0
           1,A6 S1
           S1   1
           0,A6 S1
           EXT  sbrk
           R    sbrk
           A5   ARGLIST+3,
           A3   S1
           S2   0,A3
           S3   1,A3
           B00  A5
           J    B00
ARGLIST    BSS  6
           END
*
*       COSSAVE: save b registers across SAVESYSTEM
*
          IDENT COSSAVE
          ENTRY COSSAVE
          EXT   URWERT
COSSAVE   A4     B00
          BSAVE,0   A4
          R     URWERT
          R     RESTART
          R     COSSAVX
          S2    1
          S0    S1\S2
          JSZ   IMMRET
          R     RSTR
          R     NULLREAL
          EXT   $rtstart
          EXT SCHORSCH:W
          S1    RT
          S2    $rtstart,0
          $rtstart,0   S1
          S1   SCHORSCH,0
          S1   0
          EXT  Mtimesa
          Mtimesa,0  S1
          Mtimesa+1,  S1
          EXT  rsettime
          R    rsettime
IMMRET    A4    BSAVE,0
          B00   A4
          J     B00
          EXT   COSSAVX
          EXT   RESTART
          EXT   NULLREAL
          EXT   RSTR
BSAVE     bssz  1
          END
 
 
*
*       UNICOS1 get batchmode or not via isatty(0) && isatty(1)
*
           IDENT UNICOS1
           ENTRY UNICOS1
UNICOS1    S1   0
           1,a6 S1
           a5   B00
           B00S, a5
           R    isatty
           FVAL, S1
           S1   1
           1,a6 S1
           R    isatty
           S2   FVAL,
           S1   S1&S2
           1,a6 S1
           a5   B00S,
           B00  a5
           EXT  isatty
           J    B00
B00S       BSS  1
FVAL       BSS  1
           END
 
* This is the UNICOS PSL 3.4 reprieve handler
 
         IDENT ESSIG
         FORCE STACK=ON
         SECTION CODE
 
ARGLIST  ALLOC 6
 
         ENTRY CALLFUNC
         ENTRY ssig
         ENTRY sigdata
 
ssig     =     *
ESSIG    ENTER
*
         S1        1,A6
*        S1        0,A1
         VARADD    A6,ARGLIST,STKPTR=A7
*
*  BUILD A CONTROLBLOC FOR EX
*
         0,A6      S1
         S2        P.LALLFUNC
         1,A6      S2
         S1        $sigsta
         2,A6      S1
         S1        0
         3,A6      S1
         4,A6      S1
         5,A6      S1
         SYSCALL   FC=$SIGNAL,ARG=A6
         JSZ       RAUS
         S1        -1
RAUS     EXIT
 
sigon    ENTER
         S1       $sigsta,0
         A4       0
         $sigsta,0 A4
         EXIT
 
sigoff   ENTER
         S1       $sigsta,0
         A4       1
         $sigsta,0 A4
         EXIT
 
*
*  This COMMON is shared with FORTRAN dump routine COSRPV
*  Here the system will put in the exchange package before
*  CALLfunc is invoked.
*
         COMMON    REPREV
$sigpend  BSSZ      1
$sigsta   BSSZ      1
EPACK    BSSZ      37
SIGNUM   BSSZ      1
SIGMODE  BSSZ      1
 
         COMMON    *
*
* This is the data block for SCTL_RET
*
         COMMON    MURKS
EPACK1   BSSZ      d'16
svb      BSSZ      d'64
svt      BSSZ      d'64
vsvm     BSSZ      1
rvsv     BSSZ      d'512
sigdata  BSSZ      1
sigdsv   BSSZ      1
sigmsk   BSSZ      1
         COMMON   *
 
*
* Lallfunc is PSL 3.4's own signal handler. It does  his job independent
* from an environment either in LISP or C world. This handler runs in
* UNICOS 3.0 and with a few modifications in UNICOS 2.1 . It will probably
* not run in UNICOS 4.0. Thanks to Charlie Smith , CRI for help in writing
* this handler.
*
         ENTRY     LALLFUNC
         EDIT      OFF
         ENTRY     __sigvec
         ENTRY     __handler
         EDIT      *
         ENTRY     $sigsta
         ENTRY     $sigpend
         SECTION   *
         SECTION   DATA
         SETSTKAT
 
%%NB     SET       3              number of b's
%%NT     SET       0              number of t's
%%ALLOPS SET       d'10           number of temporary variables
%%MXCALL SET       2              max size of argument list (header plus 1 arg)
%%ARGLST SET       %%STKTAG+%%ALLOPS+1  location of arglist on the stack
%%STKLEN SET       %%NB+1+%%NT+%%ALLOPS+%%MXCALL  stack frame length
 
*        Traceback Name Block.
*
         con       'LALLFUNC'
TRBK     vwd       32/0           mandatory zero
         vwd       16/5           number of following words
         vwd       16/8           name's length
         con       p.LALLFUNC     address of entry point
         vwd       1/0            base level flag
         vwd       7/1            language type
         vwd       12/%%MXCALL    max size of argument list
         vwd       12/0           unused
         vwd       32/%%ALLOPS    size of temp variable storage
         vwd       32/0           size of static constant storage
         vwd       32/0           size of static variable storage
         vwd       50/0           unused
         vwd       7/%%NT         number of t's saved
         vwd       7/%%NB+1       number of b's saved
         con       0              language dependent information
 
*        Entry point.
 
         SECTION   *
 
         COMMON IOSTUF
BROKE    BSSZ   1
         BSS    134
         COMMON *
 
ARGLST1  CON       0
         CON       EPACK1
         BSSZ      5

         EDIT OFF
__sigvec BSSZ   64
         EDIT  *
 
CALLFUNC =         *                  We need to have CALLFUNC also
LALLFUNC =         *
         EDIT      OFF
__handler =  *
         EDIT  *
         SIGNUM,0  A4
         SIGMODE,0 A5
 
* This handler performs two different ways of reprieve processing
* "call" LISPRPV via SCTL_RET on a modified exchange package normally,
* resume operation when an not iterruptable routine like gc or IO is
* involved (BRK = 17) and SIGNUM = 2 (attention)
 
         ARGLST1+2,0  A4          put signal number in argblock for SCTL_RET
         S1        BROKE,
         S2        A4
         S3        2
         S0        S2-S3
         JSN       POP
         S0        S1
         JSN       RESUME
POP      =         *
         A7        EPACK       This is code for UNICOS 3.0
         s1        LISPRPV
         s1        s1<32
         s2        maske,
         s3        EPACK,0
         s3        s1!s3&s2     merge
         EPACK1,0   s3
         J         RESUME1
 
maske    con       o'3777777740000000000
 
*2.1     J         CALRPV       for UNICOS 2.1 ********************
*2.1     EXT       CALRPV       those two lines instead + open CALRPV
         EXT       LISPRPV
 
RESUME   S6        EPACK,0       copy first word unchanged
         EPACK1,0  S6
 
SCTL_RET =         o'01000
 
RESUME1  S1         SCTL_RET
         ARGLST1,0  S1
         S1         -1
         BROKE,0    S1
 
         J         cont1          flushing instruction buffers
         bss       d'32
cont1    J         cont2
         bss       d'32
cont2    j         cont3
         bss       d'32
cont3    j         cont4
         bss       d'32
cont4    =         *
 
         a5        d'64
         vl        a5            put v registers to save area for SCTL_RET
         a0        rvsv
         a1        rvsv
         ,a0,1     v0
         a1        a1+a5
         a0        a1
         ,a0,1     v1
         a1        a1+a5
         a0        a1
         ,a0,1     v2
         a1        a1+a5
         a0        a1
         ,a0,1     v3
         a1        a1+a5
         a0        a1
         ,a0,1     v4
         a1        a1+a5
         a0        a1
         ,a0,1     v5
         a1        a1+a5
         a0        a1
         ,a0,1     v6
         a1        a1+a5
         a0        a1
         ,a0,1     v7
 
         s0        vm            put vm to save area
         vsvm,0    s0
         EPACK+16,0 s0           for output by COSPRV
 
         a0        EPACK+1        copy exchange package from epack to epack1
         a5        d'15           without first word. This word is supplied
         vl        a5             by code above
         v0        ,a0,1
         a0        EPACK1+1
         ,a0,1     v0
 
         a1        EPACK+d'8        Restore S registers
         s2        2,a1
         s3        3,a1
         s4        4,a1
         s5        5,a1
         s6        6,a1
         s7        7,a1
 
         a7        EPACK            Restore A registers
         a0        0,a7
         a1        1,a7
         a2        2,a7
         a3        3,a7
         a4        4,a7
         a5        5,a7
         a6        6,a7
         a7        7,a7
 
         s0      $SIGNAL          finished: syscall SCTL_RET
         s1      ARGLST1
         ex
         END
 
         IDENT RRRR
        MXCALLEN  7
* This code is proposed for cutting off the environment table from
* a restarted savesystem (absoluted binary).
* this code was invented by Dennis Moen , CRI.
* Unfortunately it could not present its full strenght.
*
         ENTRY RESTART
         ENTRY RSTR
         EXT   endcp
         EXT   CPENV
 
URWERT     ENTER
          S1    endcp,0
          URI,0 S1
          EXIT
URI       BSS   1
RESTART =     *
         A2    64
         A0    BUFF
         ,A0 B00,A2
         A0    TUFF
         ,A0 T00,A2
         S1      0
         J      B00
RSTR     =      *
         A2     B00
         BUFF,0 A2
         R      CPENV           copy environment
         A1     64
         A0     BUFF
         B00,A1 ,A0
         A1     64
         A0     TUFF
         T00,A1 ,A0
 
         s2     endcp,0
         oldend,0 s2
         s1     0
         CALL   sbrk,(S1),STKPTR=A7
 
         s2     oldend,0
         s1     endcp,0
         s3     s1-s2
         s3     s3<3
         s1     -s3
*        CALL   sbrk,(S1)
         CALL   sbrk,(S1),STKPTR=A7
 
         a0     BUFF,0
         b00    a0
         S1     1
         J      B00
 
BUFF     BSSZ   100
TUFF     BSSZ   100
oldend   bssz   1
 
         END
 
          IDENT     COPYENV
 
*  D' apres Language startup routine for Cray/Unix
*
* Entry:
*         A1  - points to arguments:  argc, argv, envp.
*         S7  - points to word after the user's field length
*
          ENTRY     CPENV
          EXT       environ
          EXT       endcp
 
         SECTION   CODE
CPENV     =         *
          A6        A1
          S2        endcp,0
          S0        S7-S2
          JSP       work
          J         B00            do nothing savesystem just happened
 
*   move environment words to base of MYENV
work      S5        MYENV
          S6        A6
          S4        S7-S6           move size = endcp - a6
          A3        S4
          A5        S5              A5=destination ($HPENV)
          A4        A6              A4=source of words to move (A6)
          S2        0
mvloop    =         *
              S1       0,A4
              A0       A3-1
              A3       A3-1
              0,A4     S2           zero old locations
              A4       A4+1
              0,A5     S1           move value
              A5       A5+1
              JAN      mvloop
 
*   adjust pointers:
          S6        S6-S5           move distance = a6 - MYENV
 
*        environ = environ - move distance
          S7        MYENV+2,0
          S7        S7-S6
          environ,0 S7
 
*        adjust character pointer array that environ points to
          A1        S7
          S7        0,A1            get character pointer
          A7        32
 
adjenv    =         *
             S0        S7
             JSZ       envend
             A4        S7              isolate word pointer
             S3        A4
             S7        S7>A7           isolate byte pointer
             S3        S3-S6           adjust word pointer
             S7        S7<A7
             S7        S7!S3           combine work and byte pointer
             0,A1      S7
             A1        A1+1
             S7        0,A1            get character pointer
             J         adjenv
envend    =         *
             J         B00          READY!!
 
         SECTION   *
         SECTION   DATA
MYENV     BSS      1000
          END
 
         IDENT INTALLC
*
* the initial allocation of memory. This is different in COS and UNICOS
* S1 ( the size ) is supplied by the caller
*
         ENTRY INTALLC
*
SAVEA    BSS   3
B00S     BSS   1
INTALLC  =     *
         A6    B00
         B00S, A6
         SAVEA+1,0 S1     save the argument
 
         S1    30000      this is George Behnke's idea
*                         Allocating a moderately large
*                         block of memory below the one for
*                         heap and bps ... so that if that
*                         first block is returned a free area
*                         exists below the PSL data, for IO,libs...
         S2    10
         CALLV ALLOC%
         SAVEA, S1       save result for later usage
         S1    SAVEA+1,  get argument
         S2    10        Abort on Error
         CALLV ALLOC%    allocating the 'real' block
         SAVEA+1,  S1    save it
         SAVEA+2,  S2
 
         S1     SAVEA,
         S2     10
         CALLV  DEALLC%      returning fisrt block to system
         S1     SAVEA+1,
         S2     SAVEA+2,     and restore it
 
         A6    B00S,
         B00   A6
         J     B00
         END
 
         ident     $cpused
 
*        Define utimes structure. Utms is continually updated with
*        user execution time information.
 
         SECTION   DATA
         ENTRY     Mtimesa
Mtimesa  bss       0
utms     bssz      2
         SECTION   *
         SECTION   CODE
         MXCALLEN  3
 
cpused   ENTER
         a1        utms
         s0        0,a1
         jsn       again
         CALL      cutimes,(a1),STKPTR=a7
again    =         *
         a1        utms
         s1        rt             real time clock
         s2        1,a1           utms_time   - user CPU time at last update
         s3        0,a1           utms_update - time of last update
         s4        s1-s3          rt - utm_rt
         s1        s2+s4          utm_exec + (rt - utm_rt)
         s0        s4
         jsm       again          interrupted, try again
         EXIT
 
rsettime ENTER
         a1        utms
         CALL      cutimes,(a1),STKPTR=a7
         EXIT
 
         END
 
