%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:LOAD.SL
% Title:        New version of LOAD function, with search path
% Author:       Eric Benson
% Created:      2 April 1982
% Modified:     15-May-84 14:29:31 (Brian Beach)
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:  PL:IF-SYSTEM.B
% Runtime:
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1982, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 2-Jun-86 W. Neun , ZIB Berlin
%  changed errorprintf into conterror in load1 because of infinite loops
% 10-May-84 14:29:22 (Brian Beach)
%   7 Mar 1984 1632-PST (Nancy Kendzierski, Cris Perdue)
%    Rewrote LOAD1.  Now understands loads-in-progress, gives better warning
%    messages, and correctly completes loading and importing of files for a
%    module before believing loading is complete.
%
% 01-Dec-83 14:50:16 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
% Edit by Cris Perdue, 19 Apr 1983 1438-PST
% Changed structured "constants" loadxxx* to be expressions so the
%  values are allocated in the heap, thus modifiable by users!(?)
%  <PSL.KERNEL>LOAD.RED.17, 23-Mar-83 11:44:39, Edit by KESSLER
%  Change Apollo Load directory
% Edit by Cris Perdue, 21 Mar 1983 1440-PST
% Put "" back in loaddirectories*.  Fun, huh?
% Edit by Cris Perdue,  7 Mar 1983 1527-PST
% Removed ".sl" from loadextensions* and "" from loaddirectories*.
% Edit by MLG, 6 March 1983.
%  Corrected bug in fix to Imports -- "else" was matched with incorrect "then".
% Edit by Cris Perdue, 17 Feb 1983 1201-PST
% Corrected use of *verboseload in top of load1
%  MLG, 15 Feb 1983
%   Added !*VERBOSELOAD and !*PRINTLOADNAMES
%  M. Griss, 9 Feb 1983
%   Changed LoadDirectories!* for the VAX to refer to "$pl/"
%  <PSL.NEW>-SOURCE-CHANGES.LOG.15, 15-Dec-82 15:45:55, Edit by PERDUE
%  LOAD will now handle ".sl" extension
%  <PSL.KERNEL>LOAD.RED.7,  1-Dec-82 16:07:38, Edit by BENSON
%  Added if_system(HP9836, ...)
% EDIT by GRISS 28 Oct 1982: Added EvLoad to Imports
%  <PSL.KERNEL>LOAD.RED.4,  4-Oct-82 09:46:54, Edit by BENSON
%  Moved addition of U to Options!* to avoid double load
%  <PSL.KERNEL>LOAD.RED.3, 30-Sep-82 11:57:03, Edit by BENSON
%  Removed "FOO already loaded" message
%  <PSL.KERNEL>LOAD.RED.2, 22-Sep-82 15:38:48, Edit by BENSON
%  Added ReLoad, changed VAX search path
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(loaddirectories* % list of strings to append to front
         loadextensions* % a-list of (str . fn) to append to end            % and apply
 
         pendingloads* % created by Imports, aux loads
       loads-in-progress*  % list of modules being loaded, but not done
         *lower % print IDs in lowercase, for building                      % filename for Unix
 
         *redefmsg % controls printing of redefined                         % function message
 
         *usermode % Controls query of user for redefining                  % system functions
 
         *insideload % Controls "already loaded" message
         *verboseload % Print REDEFs and LOAD file names
         *printloadnames % Print Names of files loading
         options*))
 
 % (compiletime (load if-system))
 
% list of modules already loaded
%%(if_system apollo (setq loaddirectories* (list "" "~p/l/")))
 
% (if_system tops20 (setq loaddirectories* (list "" "pl:")))
 
% (if_system unix (setq loaddirectories* (list "" "$pll/" "$pl/")))
 
% (if_system hp9836 (setq loaddirectories* (list "" "pl:")))
 
(setq loaddirectories* (list "" "$pl/")))
 
(setq loadextensions* (list (cons ".b" 'faslin) (cons ".lap" 'lapin)))
 
(setq *verboseload nil)
 
(setq *printloadnames nil)
 
(dm load (u)
  (list 'evload (mkquote (cdr u))))
 
(de evload (u)
  (foreach x in u do (load1 x)))
 
(dm reload (u)
  (list 'evreload (mkquote (cdr u))))
 
(de evreload (u)
  (foreach x in u do (setq options* (delete x options*)))
  (evload u))
 
(de tstfl (ld fil le)                          % test a file
  (let (file)
   (if (and (pairp ld) (eqcar ld 'LIB))
       (and  (libp (cadr ld))
             (libmembp (setq file (bldmsg "%w%w" fil le)))
             file)
       (when (filep (setq file (bldmsg "%w%w%w" ld fil le)))
             (progn (libp nil)        % remove lib contact
                     file ))
) ))
 
(de load1 (u)
  % Don't load if already in process of or completely loaded.
  (if (memq u loads-in-progress*)
    (conterror 4711
     "*** Warning: Load of %w previously requested, but incomplete" u
         (load1 u))
    (if (memq u options*)
      (if *verboseload (errorprintf "*** %w already loaded" u))
 
      % Otherwise, try to load the module.
      (let ((pendingloads* nil)
          (*insideload t)
          (*redefmsg *verboseload)
          (*usermode nil)
          (loads-in-progress* (cons u loads-in-progress*))
          ld le f found module)
      (setq ld loaddirectories*)
 
      % Try to find the actual file to load.
      (let ((*lower t))
        (while (and ld (not found))
          (setq le loadextensions*)
          (while (and le (not found))
            (when (setq f (tstfl (first ld) u (car (first le))))
                  (setq found (cdr (first le))))
            % Found is function to apply
            (setq le (rest le)))
          (setq ld (rest ld))))
 
      (if (not found)
        (stderror (bldmsg "%r load module not found" u))
        (progn (when (or *verboseload *printloadnames)
               (errorprintf "*** loading %w%n" f))
             (apply found (list f))
             ))
 
      % Perform import requests.
      (while pendingloads*
        (setq module (car pendingloads*))
        (setq pendingloads* (cdr pendingloads*))
        (load1 module))
 
      % Done loading.
      (setq options* (cons u options*))
      (if *verboseload (errorprintf "*** %w loaded" u))
      ))))
 
(de imports (l)
  (if *insideload
    (progn (foreach x in l do (unless (or (memq x options*)
                                       (memq x pendingloads*))
                      (setq pendingloads* (append pendingloads* (list x))))))
    (evload l)))
 
 
%
% load with stacking of *package* to ensure that things are going
% into psl , even when PCLS is running
%
(de load-to-psl (x) % to be redefined when PCLS is present
   (load1 x))
