%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:            SYSTEMIO
% Description:     CRAY-COS specific IO routines for PSL
% Author:          Modified by W.Neun, ZIB Berlin from HP/UX Version by
%                  Russ Fish, Robert R. Kessler, Martin L. Griss
% Created:         23-JAN -86
% Mode:            Lisp
% Package:
% Status:          Experimental (Do Not Distribute)
%
% Copyright (c) 1987, Konrad Zuse-Zentrum Berlin for Cray version
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Each individual system must have the following routines defined.
% SysClearIo, SysOpenRead, SysOpenWrite, SysReadRec, SysWriteRec, SysClose,
% SysMaxBuffer
%
%      The following definitions are used in the routines:
%       FileDescriptor - A machine dependent word that references a file once
%                        opened, here an CFT-Channelnumber
%       FileName - A Lisp string of the file name.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
   (on fast-integers)
 
   (compiletime (load io-decls))
 
   (fluid '(cos-stderr* cos-tty*))
 
   (fluid '(coseof cosnull cosstderr cosstdout cosstdin costty))
 
   (de cosstring (s)
  % Strip tags for passing a string.
     (if (stringp s)
       (strbase (strinf s))
       (stderror (bldmsg "Not a string: %p%n" s))))
 
%------------------------------------------------------------------------------
% Any system dependent clearing that has to be done.
%------------------------------------------------------------------------------
 
   (de syscleario ()
  % Establish channels #'s for initial system.
     (COS-clrio)
 
 %   (setf cosstdin     0)
 %   (setf cosstdout    (cos-cos_stdout))
 %   (setf cosstderr    (cos-cos_stderr))
     (setf cosnull      -1)
     (setf coseof       26)
 %   (setf costty       (cos-cos_tty))
 
     (setf stdin*       0)
     (setf stdout*      1)
     (setf errout*      stdout*)
     (setf promptout*   stdout*)
     (setf cos-stderr*  5)
     (setf cos-tty*     6)
 
  % Hand Build channeltables.     cos-TTY* is the controlling tty even when
  % stdin is redirected.
 
  (clearonechannel stdin*      59  'input  'terminalinputhandler)
  (clearonechannel stdout*     59  'output 'independentwritechar)
  (clearonechannel cos-stderr* 59  'output 'independentwritechar)
  (clearonechannel cos-tty*    59  'input  'terminalinputhandler)
 
 
     nil 
     )
 
%------------------------------------------------------------------------------
% Open FileName for input and return a file descriptor used in later
% references to the file.
%------------------------------------------------------------------------------
 
(de sysopenread (channel filename)
 
  % Open a cos StdIO channel.      fopen returns null if there is an error.
 
     (let ((filed (cos-ioc 17))
      (filenam (if (idp filename) (symnam (idinf filename)) filename))
           (errmsg ))
         (prog ()
             (if (weq (cos-ckerr) 0)
                  (cos-openr filed (strbase (strinf filenam)))
                  (conterror 90 " Cant get a channel for I/O"
                               (sysopenread channel filename))
             )
             (if (weq (cos-ckerr) 0)
                  (return filed)
                  (progn (setq errmsg " Open-Error")
                    (conterror 91 errmsg (sysopenread channel filename))
                      (return nil)
             ) )
)    )   )
%------------------------------------------------------------------------------
% Open FileName for output and return a file descriptor used in later
% references to the file.
%------------------------------------------------------------------------------
 
   (de sysopenwrite (channel filename)
 
  % Open a cos StdIO channel.
 
   (let ((filed (cos-ioc 17))
      (filenam (if (idp filename) (symnam (idinf filename)) filename))
      (errmsg ))
         (prog ()
             (if (weq (cos-ckerr) 0)
                  (cos-openw filed (strbase (strinf filenam)))
                  (conterror 90 " Cant get a channel for I/O"
                               (sysopenwrite channel filename))
             )
             (if (weq (cos-ckerr) 0)
                  (return filed)
                  (progn (setq errmsg " Open-Error")
                   (conterror 91 errmsg (sysopenwrite channel filename))
                      (return nil)
             ) )
)    )   )
 
%------------------------------------------------------------------------------
% Read from the FileDescriptor, putting a line record into the StringBuffer
% terminated by an EOL or EOF character.     Return the length of the string read
% as the indexing bound (one less than the number of characters read.)
%------------------------------------------------------------------------------
 
(de sysreadrec (filedescriptor stringbuffer)
 
   (prog (lng)
  % Check for reading past the end of file.                             ile is reached.
    (return
     (if (wneq (setq lng (cos-readr (sysmaxbuffer filedescriptor)
                         (cosstring stringbuffer) filedescriptor))
             cosnull)
       lng
       (progn
        (setf (strbyt (strinf stringbuffer) 0) $eof$)
        (cos-ckerr filedescriptor)% TTYs can type again after EOF.
        (when (and (not (internp "BEGIN1")) (wneq filedescriptor 59))
                                  %dont ruin stdin and Reduce
         (setq stringbuffer "")
         (conterror 94 "Read past EOF" (sysreadrec filedescriptor
                         stringbuffer)))
 )  )))) 
 
 
 
%------------------------------------------------------------------------------
% Write StringLength+1 characters from StringToWrite from the first position.
%------------------------------------------------------------------------------
 
(de syswriterec (filedescriptor stringtowrite stringlength)
     (cos-writer (cosstring stringtowrite) filedescriptor
                  (wplus2 stringlength 1)
     )
     (if (weq (cos-ckerr 17) 0 ) stringlength
        (conterror 93 " Write Error" (syswriterec filedescriptor
                                      stringtowrite stringlength)
)    )  )
%------------------------------------------------------------------------------
%      Close FileDescriptor, allowing it to be reused.
%------------------------------------------------------------------------------
 
(de sysclose (filedescriptor)
     (cos-close filedescriptor)
)
 
%------------------------------------------------------------------------------
%      Return the maximum buffer size.
%------------------------------------------------------------------------------
 
(de sysmaxbuffer (filedesc)  131) % most frontends like to cut a line
 
%------------------------------------------------------------------------------
 
(de contopenerror (filename accessmode)
     (continuableerror 99
               (bldmsg "`%s' cannot be open for %w" filename accessmode)
               (list 'open (mkstr filename) (mkquote accessmode))))
 
   (off fast-integers)
 
% BinaryWrite and BinaryWriteBlock and BinaryopenWrite
% for faslout. Input-part is defined in the kernel.
% and is replaced by some more friendly functions here
 
(de binaryopenwrite (filename)
     (sysopenwrite 0 filename))
 
(fluid '(onewordbuffer))
 
(de binarywrite (filedesc wordtowrite)
    (progn (setf (wgetv onewordbuffer 0) wordtowrite)
         (binarywriteblock filedesc (loc (wgetv onewordbuffer 0)) 1)
)   )
 
(de binarywriteblock (filedesc blockbase blocksize)
     (prog (len)
            (setq len (cos-bwrit blocksize blockbase filedesc))
            (if (weq (cos-ckerr 17) 0) (return len)
              (conterror  92 "Write Error" (BinaryWriteBlock
                            filedesc blockbase blocksize))
)     )     )
 
(de binaryopenread (filename)
     (sysopenread 0 filename))
 
(de binaryread (channel)    % Fortran's Channel-Number
                            % read 1 Word of 64 bits
      (if (weq (cos-ckerr (cos-bread 1 (loc (wgetv onewordbuffer 0))
                                             channel)) 0)
           (wgetv onewordbuffer 0)
           (conterror  92 "Read Error" (BinaryRead channel))
)   )
