%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:TRAP.SL
% Description:  Signal handling for Sun Unix 4.2.
% Author:       Vicki O'Day, HP Labs/CRC
% Created:      27-Feb-84
% Modified:     2-Jan-85 13:13:16 (Vicki O'Day)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 17-Feb-89 (Chris Burdorf)
%  Use sun3_sigset for os/4.
% 09-Oct-87 (Leigh Stoller)
%  Do not call build-trap-message anymore. It is too flakey to be relied
%   on to tell the truth.
% 31-Apr-87 (Leigh Stoller)
%  After consult with Julian Padget, altered sigunwind to decrease the
%   stack frame by 320 bytes instead of 98. The interrupted address is
%   found 312 bytes into this frame.
% 18-Aug-86 (Leigh Stoller)
%  Reworked so that signals can be added more easily. Also added calls to
%  external functions which set up the trap, and reinit the trap after a
%  signal is recieved.
% 2-Jan-85 10:21:47 (Vicki O'Day)
%  Instead of just calling stderror with a message about the kind of trap,
%  we now find the name of the routine that caused the trap and report
%  that as well.  CODE-ADDRESS-TO-SYMBOL is used to figure out the name.
% 27-Dec-84 10:55:56 (Vicki O'Day)
%  Added sigunwind, to pop the signal handler's stack frame before calling
%  stderror.  (Many thanks to Jim Davis for deciphering the contents of this
%  frame.)
% 17-Dec-84 10:24:41 (Vicki O'Day)
%  Added lots more signals to catch, to try to prevent those occasional
%  dumps out of Nmode.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(signumber*))

(compiletime
 (progn
   (setf *sigcalls* nil)
   %
   % Create a list of initialization code needed to set up the traps. The
   % return value is the entry points for each signal handler.
   % This is later flaged as a
   %  cmacro so the returned code can be inserted directly into the lap coded
   %  procedure 'initializeinterrupts each time it is called. The list
   %  *sigentries is then inserted to define the entry points.
   (de *sigsetup (signumber function handler errorstring)
       (setf *sigcalls* (append
             % Sigset takes a function pointer.
             `((*move (wconst ,signumber) (reg 1))
               (*move ,handler (reg 2))
               (*link sun3_sigset expr 2))
             *sigcalls*))
       % Return the the function definition for the signal handler.
       `((*entry ,function expr 0)
     
        % OS2: handler is called with EIP in reg EBX
        % don't need SIGRELSE
     ,handler
     (*move (reg st) (fluid errorstack))
     (*move (quote ,errorstring) (reg 1))
     (*move (quote ,signumber)(fluid signumber*))
     (*jcall sigunwind))
       )
 
   % Return the entry point list. Defined as a cmacro.
   (de *sigcall ()
       *sigcalls*)
 
   (defcmacro *sigcall)
   (defcmacro *sigsetup)
 
   )
) % End of compiletime.
 
 
(lap '(
       % NO!!! (*sigsetup 1  aborthandler  aborthandlerinstruction  "Abort")
       
       (*sigsetup 5  SegHandler  SegHandlerInstruction
                                    "Segmentation Violation")

       (*sigsetup 3  SigHandler  SigHandlerInstruction
                                    "External Interrupt Signal")

       
       (*sigsetup 16#9b  IZDHandler  IZDHandlerInstruction 
                         "Integer Zero Divide")
       
       (*sigsetup 16#95  FZDHandler  FZDHandlerInstruction  
                         "Float Zero Divide")
       
       (*sigsetup 16#97  FISHandler  FISHandlerInstruction  
                         "Float Illegal Instruction")
       
       (*sigsetup 16#1c  IllHandler  IllHandlerInstruction  
                         "Illegal Instruction")

       (*sigsetup 16#9d  PRIHandler  PRIHandlerInstruction  
                         "Privileged Instruction")


       (*sigsetup 16#9c  IOVHandler  IOVHandlerInstruction  
                         "Integer Overflow")

       (*sigsetup 16#9e  ALGHandler  ALGHandlerInstruction  
                         "Datatype Misalignment")

       
       (*sigsetup 16#6   PAGHandler  PAGHandlerInstruction  
                         "In Page Error")
       
       (*entry initializeinterrupts-1 expr 0)
       (*sigcall)
       (*exit 0)))
 
(de initializeinterrupts ()
       (ieee_handler (strbase (strinf "set"))
                  (strbase (strinf "common"))
                  (symfnc (id2int 'fpehandler)))
       (initializeinterrupts-1)
      )

 
(lap
 '((!*entry sigunwind expr 0)
     % At this point, an arg is already set up in register 1 with a message
     % describing the kind of trap.  First, though, we want to find the
     % name of the routine that caused the trap by using the "interrupted
     % address" in the signal handler stack frame.  Then we want to
     % pop this frame off the stack.  On the SUN, this adds up to 320 bytes.
     % Who know why.
 
     % (*move (memory (reg st) 312) (reg 2)) % Addr. at which signal occurred
     % (fclex)                               % Clear floating-point exceptions
     % (*wplus2 (reg st) 320)                % Pop signal-handling frame
     
     (*link build-trap-message expr 2)     % This leaves its result in reg
                                           % 1, so the new message is
     (*move (reg 1)(reg 2))
     (*move (fluid signumber*)(reg 1))
     (*wplus2 (reg 1)(wconst 10000))
     (*jcall error)
     ))
 
(de build-trap-message (trap-type trap-addr)
    (let (extra-info)
      (if (funboundp 'code-address-to-symbol)
        (setf extra-info
          (bldmsg "%w%n%w%n%w"
              " : the name of the routine that trapped can't be"
              " reported unless the function CODE-ADDRESS-TO-SYMBOL"
              " has been defined, by loading ADDR2ID."))
    % else, get the name of the offending function
    (setf extra-info (bldmsg "%w%w"
                 " in "
                 (code-address-to-symbol (inf trap-addr))))
      )
      (bldmsg "%w%w" trap-type extra-info)))
 
(fluid '(bruch_bruch))

(de terminal-interrupt()
    (setq bruch_bruch 0)
    (error 10002 "terminal interrupt"))

% End of file.
 
