%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         FASTBIG
% Description:  Vector Based BIGNUM package with vector instructions
%               this a totally CRAY specific version
% Author:       H. Melenk, W. Neun
%               derived from NBIG by M. L. Griss & B Morrison
% Created:      15 Dec 1986
% Modified:
% Mode:         Lisp
% Package:      Utilities
% Status:       Experimental
% Compiletime:  FAST-VECTOR.B INUM.B vector-instructions.b
%
% Copyright (c) 1987, Konrad-Zuse-Zentrum Berlin, all rights reserved
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% Copyright (C) 1982, A. C. Norman, B. Morrison, M. Griss
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% A bignum will be a VECTOR of Bigits: (digits in base BigBase):
%  [NIL b1 ... bn   ] or [T b1 ... bn     ].  BigZero is thus [NIL   ]
% All numbers are positive, with T as 0 element to indicate negatives
%
% On CRAY-1, Important Ranges are:
%              --------------------------------
% POSBETA      |    0          |    n         |
%              --------------------------------
%                  43                21       bits
%              --------------------------------
% POSINT       | 0 |       0  |               |
%              --------------------------------
%                5         22       37              bits
%              --------------------------------
% NEGINT       | -1|       -1 |               |
%              --------------------------------
% Thus BETA:  2^21-1
%      INT    2^37-1
%      FIX    2^42-1
%
% BetaRange is chosen so that multiplication is possible (46 bit arith.)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Revisions:
%            W. Neun , 16 January 1989
%            fixed bugs in lshiftaux; (decompressing if needed,treatment of
%            fixnums
%            W. Neun , 14 July 1988
%            fixed a bug in B64registerfromfloat which caused
%            (gcdn -7566047373982433281 144115188075855872)
%            to loop infintely
%            W. Neun,1 . Dec 1987
%            fixed a bug in bharddivide which corrupted all the memory
%            W. Neun,1 . Dec 1987
%            added bigcompress and bigdecompress
%            W. Neun,13. Jan 1988
 
(compiletime (load fast-vector if-system inum vector-instructions nbig0))
(compiletime (setq *syslisp t))
(compiletime (put 'crayFloatone 'opencode '((sconst10 (reg 1)))))
 
(compiletime
  (flag
   '( big2sysaux setbits setupglobals blxor blor bland blnot trimbignum
      trimbignum1 bnumaux bnum nonbignumerror b64bigfromV7save b64makeo ne
      bremainder bdifference bdivide blshift btimes2  b64bigfromV7 blessp-v
      bdigittimes2 b64bigfromV6save bgcdnaux bsimpledivide b64bigfromV6
      bharddivide bgreaterp blessp bunsignedgreaterp irecip bsmalladd
      bsmalltimes2 bchannelprin2 checkifreallybigpair
      bsmalldiff checkifreallybig b64plus2aux b64registeradd bmax bmin
      b64registernormalize b64differenceaux b64registersub b64times2aux
      b99differenceaux b99plus2aux b99times2aux bminusp
      b64specialregistertimes b64smalltimes2aux bminus
      btwopower bplus2 bonep bzerop babs checkifreallybigornil
      b99lshiftaux b64registerleftshift b99rshiftaux b64registerrightshift
      b64move b64harddivide b64remainder b64gcdn bunsigned-geq-v
      b64registerunsigned-geq-v bunsigned-greaterp-v sys2big-i
      b64floatfromregister b64registerfromfloat
      bdivide1000000ip
      bprin printdigit addcarry subcarry unsignedFloatfromBignum
      b64makeone)               'Internalfunction))
 
(fluid '(bigbetahi*       % Largest BetaNum in BIG format
         bigbetalow*      % Smallest BetaNum in BIG format
         betahi*          % Largest BetaNum as Inum
         betalow*         % Smallest BetaNum as Inum
         syshi*           % Largest SYSINT in FixN format
         syslow*          % Smallest SYSINT in FixN format
         bigsyshi*        % Largest SYSINT in BIG format
         bigsyslow*       % Smallest SYSINT in BIG format
         floatsyshi*      % Largest SYSINT in Float format
         floatsyslow*     % Smallest SYSINT in Float format
         bbase*           % BETA, base of system
         floatbbase*      % As a float
         bigfloathi*      % Largest   Float in BIG format
         bigfloatlow*     % Smallest Float in BIG format
         staticbig*       % Warray for conversion of SYS to BIG
         bone*            % A one
         bzero*           % A zero
         bbits*           % Number of Bits in BBASE*
         b64save1 b64save2 b64save3  % for Vregister
         b64save4 b64save5
         staticfloat
         *bigcompress-threshold* % when to compress a bignum
         arithargs0 arithargs1 % for Coerce1/2 (in Tregister)
         big-preallocated*   % an empty space for a bignum
         digit2letter* carry* outputbase*))
 
% --------------------------------------------------------------------------
(compiletime
(progn
 
(put '*bigcompress-threshold* 'registername "B75")
(put '*bigcompress-threshold* 'Breg 8#75)
 
(if_system Cray-2
 (put '*WFloat  'opencode
     '((int2fl (reg 1) (reg 1))))
 (put '*WFloat  'opencode
     '((sconst06 (reg 2))
       (sub (reg 1) (reg 2) (reg 1))
       (fsub (reg 1) (reg 2) (reg 1))
)     )) 
 
(ds StaticIntFloat (X)
      (progn
         (wputv staticFloat 1 (*wFloat x))
         (mkfltn staticFloat)))
 
(dm aplus (u)                        % like plus macro
  (robustexpand (cdr u) 'aplus2 0))
 
(ds notbetarangep (x)
  % Test if SYSINT in reduced INUM range
  (or (wgreaterp x betahi*)
      (wlessp x betalow*)))
 
(ds openbsize (b) (veclen b)) % need no (inf b) !!
 
(ds openbminusp (b) (igetv b 0)) % need no (inf b) !!
 
(ds opengtbignum(lb sign)
  %
  % opencode for BIGNUM allocation
  % CAUTION does not put a Hwords Tag and length
  % because this may be altered later
  %
   (prog (**b**)
 
     (when (wlessp lb 0) (return -5)) % nice dump ]
     (when (not (eq (setq **b** big-preallocated*) 0)) (go **m**))
 
    **ll**
       (setq **b** heaplast)
       (setq heaplast (aplus heaplast lb 2))
       (when (wgeq (wplus2 heaplast 100) heaptrapbound)
                             (setq heaplast **b**)
                             (reclaim) (go **ll**))
    **m**
       (setq **b** (mkbign **b**))
       (putmem (aplus2 **b** 1) sign)
       (return **b**)))
 
(ds openbzerop (v1)
  (and (bigp v1) (izerop (openbsize v1)) (not (openbminusp v1))))
 
 % Easy distinction between representation types of Integers:
 % Bignums start at 2 bigits exactly, all numbers below that are
 % Fixnums or Sysints.
 
(ds checkifreallybig1 (u)
       ((lambda (**uu** **size** **maxsize**)
         (cond ((wgreaterp (setq **size** (openbsize **uu**)) **maxsize**)
                    (bigcompress **uu**)) % then compress it
               ((wgreaterp **size** 2) **uu**)
               (t (sys2int (big2sysaux **uu**)))))
        u 0 *bigcompress-threshold*))
 
(if_system Cray-2
 (put 'unboxedint2float 'opencode
     '((int2fl (reg 1) (reg 1))))
 (put 'unboxedint2float 'opencode '(
   (sconst06 (reg 3))
   (sub (reg 1) (reg 3) (reg 1))
   (fsub (reg 1)(reg 3) (reg 1)) )))
 
 (define-constant bmask** (sub1 bbase*))
 (define-constant bbits** bbits*)
 (define-constant bbase** bbase*)
 
))
 
% Returns type tag of Coerced X type and sets ArithArg0 to be Coerced X
% Beware of ADD1/SUB1 cases, maybe can optimize later
% Added test for compressed bignums
 
(de Coerce1 (X F)
  (prog (T1)
     (setq T1 (TAG X))
     (case T1 ((NEGINT-tag) (progn
                              (setq T1 posint-tag)
                              (when (NOTBETARANGEP X)
                                  (setq T1 bignum-tag)
                                  (setq X (SYS2BIG X)))))
              ((posint-tag) (when (NOTBETARANGEP X)
                                  (setq T1 bignum-tag)
                                  (setq X (SYS2BIG X))))
              ((bignum-tag) (when (wgreaterp (inf (wgetv x 1)) 128)
                               % NOT NICE, but a fast test for T NIL vs
                               % Bigcompress bigdecompress
                               (setq x (bigdecompress x))))
              ((fixnum-tag)
                        (progn(setq T1 posint-tag)
                              (setq X (FIXVAL X))
                              (when (NOTBETARANGEP X)
                                  (setq T1 bignum-tag)
                                  (setq X (SYS2BIG X)))))
              ((floatnum-tag) NIL)
              (nil (return (nonnumber1error X F))))
     (setq ArithArgs0 X)
     (return T1)))
 
% Returns type tag of strongest type and sets ArithArg0 to be Coerced X
% and ArithArg1 to Coerced Y.
% Added test for compressed bignums
 
(de Coerce2 (X Y F)
  (prog (T1 T2 P C PPP )
 
     (setq T1 (TAG X))
     (case T1 ((NEGINT-tag) (progn
                              (setq T1 posint-tag)
                              (when (NOTBETARANGEP X)
                                  (setq T1 bignum-tag)
                                  (setq X (SYS2BIG X)))))
              ((posint-tag) (when (NOTBETARANGEP X)
                                  (setq T1 bignum-tag)
                                  (setq X (SYS2BIG X))))
              ((bignum-tag) (when (wgreaterp (inf (wgetv x 1)) 128)
                               % NOT NICE, but a fast test for T NIL vs
                               % Bigcompress bigdecompress
                               (setq x (bigdecompress x))))
              ((fixnum-tag)
                        (progn(setq T1 posint-tag)
                              (setq X (FIXVAL X))
                              (when (NOTBETARANGEP X)
                                  (setq T1 bignum-tag)
                                  (setq X (SYS2BIG X)))))
              ((floatnum-tag) NIL)
              (nil (return (nonnumber2error X Y F))))
 
     (setq T2 (TAG Y))
     (case T2 ((NEGINT-tag) (progn
                              (setq T2 posint-tag)
                              (when (NOTBETARANGEP Y)
                                  (setq T2 bignum-tag)
                                  (setq Y (SYS2BIG Y)))))
              ((posint-tag) (when (NOTBETARANGEP Y)
                                  (setq T2 bignum-tag)
                                  (setq Y (SYS2BIG Y))))
              ((bignum-tag) (when (wgreaterp (inf (wgetv Y 1)) 128)
                               % NOT NICE, but a fast test for T NIL vs
                               % Bigcompress bigdecompress
                               (setq Y (bigdecompress Y))))
              ((fixnum-tag)
                        (progn(setq T2 posint-tag)
                              (setq Y (FIXVAL Y))
                              (when (NOTBETARANGEP Y)
                                  (setq T2 bignum-tag)
                                  (setq Y (SYS2BIG Y)))))
              ((floatnum-tag) NIL)
              (nil (return (nonnumber2error X Y F))))
     (SETF ArithArgs0 X)
     (SETF ArithArgs1 Y)
     (cond ((Weq T1 T2) (return T1))
           ((Wlessp T1 T2)
               (progn (setq PPP 0) (setq C T2) (setq T2 T1) (setq T1 C)))
           (T (setq PPP 1)))
     (case T1
              ((floatnum-tag)
               (case T2
                    ((posint-tag)
                     (if (weq PPP 0)
                      (setq ArithArgs0 (StaticIntfloat ArithArgs0))
                      (setq ArithArgs1 (StaticIntfloat ArithArgs1))))
                    ((bignum-tag)
                     (if (weq ppp 0)
                      (setq ArithArgs0 (FloatfromBignum ArithArgs0))
                      (setq ArithArgs1 (FloatfromBignum ArithArgs1))))))
              ((bignum-tag)
                      (if (weq ppp 0)
                          (setq ArithArgs0 (SYS2BIG ArithArgs0))
                          (setq ArithArgs1 (SYS2BIG ArithArgs1)))))
     (return T1)))
 
(de setbits (x)
  %
  % This function sets the globals for big bignum package.
  % "x" is assumed to be 44 (so that one can use Cray's Fastints
  (prog (y)
 
    (setq *bigcompress-threshold* 10)   % For a first test
    (setq bbits* (wrshift (isub1 x) 1)) % Total number of bits per word used.
    (setq bbase* (twopower bbits*))     % "Beta", where n=A0 + A1*beta
                                        %                    + A2*(beta^2).
    (setq floatbbase* (intfloat bbase*))
    (setq betahi* (isub1 bbase*))
    (setq betalow* (iminus bbase*))
    (setq bone* (bnum 1))
    (setq bzero* (bnum 0))
    (setq bigbetahi* (bnum betahi*))           % Highest value of Ai
    (setq bigbetalow* (bminus bigbetahi*))     % Lowest value of Ai
                                               % here assume 2's complement
    (setq y (twopower (idifference x 2)))   % eg, 36 bits, 2^35-1=2^34+2^34-1
    (setq syshi* (plus y (difference y 1)))
    (setq y (minus y))
    (setq syslow* (plus y y))
    (setq bigsyshi* (bdifference (btwopower (isub1 x)) bone*))
                 % Largest representable Syslisp integer.
                 % Note that SYSPOS has leading 0, ie only x-1 active bits
    (setq bigsyslow* (bminus (bplus2 bone* bigsyshi*)))))
                 % Smallest representable Syslisp integer.
 
(de nonbignumerror (v l)
  (stderror (bldmsg " Expect a BIGNUM in %r, given %p%n" l v)))
 
(de bsize (v)
  % Upper Limit of bignum
  (if (bigp v) (veclen (vecinf v)) 0))
 
(de gtpos (n)
  % Allocate :NIL a1 ... an   :
  (prog ()
        (if (not (eq big-preallocated* 0))
           (progn
              (wputv big-preallocated* 0 (mkitem hwords-tag n))
              (setq n big-preallocated*))
           (setq n (mkwrds (gtwrds n))))
        (iputv n 0 NIL)
        (return (mkbign (vecinf n)))))
 
(de gtneg (n)
  % Allocate :T a1 ... an     :
  (prog nil
        (if (not (eq big-preallocated* 0))
           (progn
              (wputv big-preallocated* 0 (mkitem hwords-tag n))
              (setq n big-preallocated*))
           (setq n (mkwrds (gtwrds n))))
        (iputv n 0 T)
        (return (mkbign (vecinf n)))))
 
(compiletime
 
 (ds gtbig-open (nn sign)
   % Allocate [SIGN a1 ... ann  ]
  (prog (**n**)
        (if (not (eq big-preallocated* 0))
           (progn
              (wputv big-preallocated* 0 (mkitem hwords-tag nn))
              (setq **n** big-preallocated*))
           (setq **n** (mkwrds (gtwrds nn))))
        (iputv **n** 0 sign)
        (return (mkbign (vecinf **n**)))))
)
 
(de trimbignum (v3)
  % truncate trailing 0
  (if (not (bigp v3))
    (nonbignumerror v3 'trimbignum)
    (trimbignum1 v3 (openbsize v3))))
 
 (de trimbignum1 (b l3)
  (prog (v3 pos diff di2 i)
       (setq v3 (inf b))
       (setq pos (wplus2 1 (wplus2 v3 l3)))
       (setq i pos)
    lupe 
       (setq diff (wdifference i v3))
       (if (wgreaterp diff 5) (vsetvl 5) (vsetvl diff))
       (vload (vreg 3) i -1)
       (setq di2 (vtestn (vreg 3)))
       (when (wneq di2 0) (setq di2 (wldzero di2))
                          (setq l3 (wdifference l3 di2))
                          (go found))
       (setq i (wplus2 i -5))
       (setq l3 (wplus2 l3 -5))
       (when (wgreaterp i v3) (go lupe))
       (setq l3 0)
     found
                         % avoid negative zeroes
       (when (izerop l3) (setf (getmem (aplus2 v3 1)) NIL))
       (return
         (let (len len2 vi)
           (setq len (wrdlen v3))
           (when (not (weq len l3))
                 % Already the size
             (setf (getmem v3) (mkitem hwords-tag l3))
             (setf (wrditm v3 (wplus2 l3 1))
                 (mkitem hwords-tag (wplus2 (wdifference len l3) -2))))
           b))))
 
(de big2sys (u)
  % Convert a BIG to SYS, if in range
  (if (wgreaterp (openbsize u) 3)
    (continuableerror 99 "BIGNUM too large to convert to SYS" u)
    (big2sysaux u)))
 
(de big2sysaux (u)
  % Convert a BIGN that is in range to a SYSINT
  (prog (l sn res)
        (setq l (openbsize u))
        (when (openbminusp u) (setq sn t))
        (when (izerop l) (return 0))
        (setq res (igetv u l))
        (setq l (isub1 l))
        (while (neq l 0)
             (setq res (wlshift res bbits**))
             (setq res (wor res (igetv u l)))
             (setq l (isub1 l)))
        (if sn (return (wminus res))
        (return res))))
 
(de twopower (n)
  %fix/i-num 2**n
  (expt 2 n))
 
(compiletime (ds wtwopower (n) (wlshift 1 n)))
 
(de btwopower (n)
  % gives 2**n; n is fix/i-num; result BigNum
    (prog (quot v)
        (cond ((intp n) nil)
              ((fixnp n) (setq n (fixval n)))
              ((bigp n) (setq n (big2sys n)))
              (t (return (nonintegererror n 'btwopower))))
        (setq quot (longdivide-quotient (longdivide n bbits**)))
        (setq n (longdivide-remainder))
        (setq v (gtbig-open (iadd1 quot) nil))
        (ifor (from i 1 quot 1) (do (iputv v i 0)))
        (iputv v (iadd1 quot) (wtwopower n))
        (return (trimbignum1 v (iadd1 quot)))))
 
(de bzerop (v1)
  (and (bigp v1) (izerop (openbsize v1)) (not (openbminusp v1))))
 
(de bonep (v1)
  (and (not (bminusp v1)) (ionep (openbsize v1)) (ionep (igetv v1 1))))
 
(de babs (v1)
  (if (bminusp v1) (Bminus v1) v1))
 
(de bmax (v1 v2)
  (if (bgreaterp v2 v1) v2 v1))
 
(de bmin (v1 v2)
  (if (blessp v2 v1) v2 v1))
 
%%%%%%%%%%%%%%%%%%%%%%%%%% Logical Operations %%%%%%%%%%%%%%%%%%%%%%%%
 
(de biglor (u v)
  (checkifreallybig1 (blor u v)))
 
(de blor (v1 v2)
  % The main body of the OR code is only obeyed when both arguments
  % are positive, and so the result will be positive;
  (if (or (openbminusp v1) (openbminusp v2))
         (blnot (bland (blnot v1) (blnot v2)))
    (prog (l1 l2 l3 v3 steplth)
          (setq l1 (openbsize v1))
          (setq l2 (openbsize v2))
          (when (igreaterp l2 l1)
            (setq l3 l2)
            (setq l2 l1)
            (setq l1 l3)
            (setq v3 v2)
            (setq v2 v1)
            (setq v1 v3))
          (setq v3 (gtbig-open l1 nil))
          (ifor (from i 1 l2 64)
              (do
                (setq steplth (wplus2 (wdifference l2 i) 1))
                (when (wgreaterp steplth 64) (setq steplth 64))
                (VsetVl steplth)
                (Vload (Vreg 1) (aplus v1 1 i) 1)
                (Vload (Vreg 2) (aplus v2 1 i) 1)
                (VorV (Vreg 3) (Vreg 1) (Vreg 2))
                (Vstore (Vreg 3) (aplus v3 1 i) 1)
          )   )
          (ifor (from i (wplus2 l2 1) l1 64)
              (do
                (setq steplth (wplus2 (wdifference l2 i) 1))
                (when (wgreaterp steplth 64) (setq steplth 64))
                (VsetVl steplth)
                (Vload (Vreg 1)  (aplus v1 1 i) 1)
                (Vstore (Vreg 1) (aplus v3 1 i) 1)
          )   )
          (return v3))))
 
(de biglxor (u v)
  (checkifreallybig1 (blxor u v)))
 
(de blxor (v1 v2)
  % negative arguments are coped with using the identity
  % LXor(a,b) = LNot LXor(Lnot a,b) = LNor LXor(a,Lnot b);
  (prog (l1 l2 l3 v3 s steplth)
        (when (openbminusp v1)
          (setq v1 (blnot v1))
          (setq s t))
        (when (openbminusp v2)
          (setq v2 (blnot v2))
          (setq s (not s)))
        (setq l1 (openbsize v1))
        (setq l2 (openbsize v2))
        (when (igreaterp l2 l1)
          (setq l3 l2)
          (setq l2 l1)
          (setq l1 l3)
          (setq v3 v2)
          (setq v2 v1)
          (setq v1 v3))
        (setq v3 (gtbig-open l1 nil))
        (ifor (from i 1 l2 64)
              (do
                (setq steplth (wplus2 (wdifference l2 i) 1))
                (when (wgreaterp steplth 64) (setq steplth 64))
                (VsetVl steplth)
                (Vload (Vreg 1) (aplus v1 1 i) 1)
                (Vload (Vreg 2) (aplus v2 1 i) 1)
                (VXorV (Vreg 3) (Vreg 1) (Vreg 2))
                (Vstore (Vreg 3) (aplus v3 1 i) 1)
        )   )
        (ifor (from i (wplus2 l2 1) l1 64)
            (do
                (setq steplth (wplus2 (wdifference l2 i) 1))
                (when (wgreaterp steplth 64) (setq steplth 64))
                (VsetVl steplth)
                (Vload (Vreg 1)  (aplus v1 1 i) 1)
                (Vstore (Vreg 1) (aplus v3 1 i) 1)
        )   )
        (setq v1 (trimbignum1 v3 l1))
        (when s
          (setq v1 (blnot v1)))
        (return v1)))
 
(de bigland (u v)
  (checkifreallybig1 (bland u v)))
 
(de bland (v1 v2)
  % If both args are -ve the result will be too. Otherwise result will
  % be positive;
  (if (and (openbminusp v1) (openbminusp v2))
    (blnot (blor (blnot v1) (blnot v2)))
    (prog (l1 l2 l3 v3 steplth)
          (setq l1 (openbsize v1))
          (setq l2 (openbsize v2))
          (setq l3 (if (igreaterp l1 l2) l2 l1)) %wminimum
          (setq v3 (gtbig-open l3 nil))
          (cond
           ((openbminusp v1)
              (ifor (from i 1 l3 64)
               (do
                (setq steplth (wplus2 (wdifference l3 i) 1))
                (when (wgreaterp steplth 64) (setq steplth 64))
                (VsetVl steplth)
                (Vload (Vreg 1) (aplus v1 1 i) 1)
                (VlminusS (Vreg 3) (Vreg 1) bmask**)
                (Vload (Vreg 2) (aplus v2 1 i) 1)
                (VandV (Vreg 5) (Vreg 3) (Vreg 2))
                (Vstore (Vreg 5) (aplus v3 1 i) 1)
              )))
           ((openbminusp v2)
              (ifor (from i 1 l3 64)
               (do
                (setq steplth (wplus2 (wdifference l3 i) 1))
                (when (wgreaterp steplth 64) (setq steplth 64))
                (VsetVl steplth)
                (Vload (Vreg 1) (aplus v1 1 i) 1)
                (Vload (Vreg 2) (aplus v2 1 i) 1)
                (VlminusS (Vreg 3) (Vreg 2) bmask**)
                (VandV (Vreg 5) (Vreg 3) (Vreg 1))
                (Vstore (Vreg 5) (aplus v3 1 i) 1)
              )))
            (t
              (ifor (from i 1 l3 64)
               (do
                (setq steplth (wplus2 (wdifference l3 i) 1))
                (when (wgreaterp steplth 64) (setq steplth 64))
                (VsetVl steplth)
                (Vload (Vreg 1) (aplus v1 1 i) 1)
                (Vload (Vreg 2) (aplus v2 1 i) 1)
                (VandV (Vreg 5) (Vreg 2) (Vreg 1))
                (Vstore (Vreg 5) (aplus v3 1 i) 1)
              ))))
          (return (trimbignum1 v3 l3)))))
 
(de biglnot (u)
  (checkifreallybig1 (blnot u)))
 
(de blnot (v1)
   (prog (v2 sign)
     (setq v2 (bsmalladd v1 1))  %impure, but avoid a second copy
     (setq sign (wgetv v2 1))
     (wputv v2 1 (not sign))
     (return v2)))
 
(de biglshift (u v)
  (checkifreallybig1 (blshift u v)))
 
(de lshiftaux (u v)
  (cond
    ((not (intp v)) (nonintegererror v 'lshift))
    ((bigp u)     (when (wgreaterp (inf (wgetv u 1)) 128)
                      (setq u (bigdecompress u)))
                  (checkifreallybig1 (blshift u v)))
    ((and (betap u) (betap v))
        (cond ((wlessp v 0) (sys2int (wshift u v)))
              ((eq u 0) 0)
              ((wlessp v bbits**) (sys2int (wshift u v)))
              (t (checkifreallybig (blshift (sys2big u) v)))))
    (t (checkifreallybig (blshift (sys2big (int2sys u)) v)))))
 
(de blshift (v1 v2)
  (if (wlessp v2 0)
    (b99rshiftaux v1 (wminus v2))
    (b99lshiftaux v1 v2)))
 
% right shift bignum kernel in registers:
%       V0 = V1 > 1
%              V3 to V5 are destroyed
%              V1, V2, V6 , V7 remain unchanged
%       l1, l2: length of information in V1, V2
%               l2 not greater than l1
%               V2 filled up with trailing zeroes
%  function value:
%       length of result.
 
(de B64Registerrightshift (l1 shiftsize carry)
     (prog (l3)
         %                calculate basical result
         (VsetVl (aplus2 l1 1))
         (VdrshiftV (Vreg 3) (Vreg 1) shiftsize)
                        % delete underflow bits
         (VsetVl l1)
         (VandS (Vreg 2) bmask** (Vreg 3))
 
         %                shift the underflows inside bigit and merge them
         (VsetElement (Vreg 3) 0 carry)
         (VrshiftV (Vreg 4) (Vreg 3) (wdifference 64 bbits**))
         (VorV (Vreg 0) (Vreg 2) (Vreg 4))
                        % return the new carry
         (return (VgetElement (Vreg 3) l1))
))
 
(de B99RShiftaux (big size)
   (prog (movesize shiftsize result lth carry d lth-movesize)
         (setq lth (openbsize big))
         (setq movesize (longdivide-quotient (longdivide size bbits**)))
         (setq shiftsize (longdivide-remainder))
         (setq lth-movesize (wdifference lth movesize))
 
         (when (wleq lth-movesize 0) (return (gtbig-open 0 nil)))
         (setq result (opengtbignum lth-movesize nil))
         (wputv result 1 (wgetv big 1))   % copy sign
         (when (eq shiftsize 0) % case: multiple of bigit
               (B64move result 2 big (wplus2 movesize 2) lth-movesize)
               (wputv result 0 (mkitem hwords-tag lth-movesize))
               (return result) )
         (setq carry 0)
         (ifor (from i lth (wplus2 movesize 1) -63)
              (do
               (setq d (wdifference i movesize) )
               (when (wgreaterp d 63) (setq d 63))
               (VsetVl d)
               (Vload (Vreg 1) (aplus big i 1) -1)
               (setq carry (B64registerRightShift d shiftsize carry))
               (Vstore (Vreg 0) (aplus result
                                   (wdifference i movesize) 1) -1)
         )    )
         (when (eq (wgetv result (aplus2 lth-movesize 1)) 0)
                   % do trimbignum
                   (when (eq big-preallocated* 0)
                         (setq heaplast (aplus2 heaplast -1)))
                   (setq lth-movesize (wplus2 lth-movesize -1)))
         (wputv result 0 (mkitem hwords-tag lth-movesize))
 
         (return result)) )
 
% left shift bignum kernel in registers:
%       V0 = V1 < 1
%              V3 to V5 are destroyed
%              V1, V2, V6 , V7 remain unchanged
%       l1, l2: length of information in V1, V2
%               l2 not greater than l1
%               V2 filled up with trailing zeroes
%  function value:
%       length of result.
 
(de B64RegisterLeftshift (l1 shiftsize carry)
     (prog (l3)
         %                calculate basical result
         (VsetVl l1)
         %          shifting left and getting carries
         (VlshiftV (Vreg 3) (Vreg 1) shiftsize)
         (VandS (Vreg 2) bmask** (Vreg 3))
         %         isolating carries and shifting left to next word
         (VrshiftV (Vreg 4) (Vreg 3) bbits**)
         (VdlshiftV (Vreg 3) (Vreg 4) 64)
         (VsetElement (Vreg 3) (aplus2 l1 -1) carry)
                                             % put last carry rightmost
         % put parts together
         (VorV (Vreg 0) (Vreg 2) (Vreg 3))
                        % return the new carry
         (return (wrshift (VgetElement (Vreg 1) 0)
                           (wdifference bbits** shiftsize)))
))
 
(de B99LShiftaux (big size)
   (prog (movesize shiftsize result lth lthnew carry d)
       (setq lth (inf (getmem big)))
       (setq movesize (longdivide-quotient(longdivide size bbits**)))
       (setq shiftsize (longdivide-remainder))
       (setq lthnew (wplus2 lth movesize))
       (setq result (opengtbignum (wplus2 lthnew 1) nil))
       (ifor (from i 1 movesize 64)
            (do
             (setq d (wplus2 (wdifference movesize i) 1))
             (when (wgreaterp d 64) (setq d 64))
             (VsetVl d)
             (Vclear (Vreg 3))
             (Vstore (Vreg 3) (aplus result i 1) 1)
       )    )
       (wputv result 1 (wgetv big 1))   % copy sign
       (when (eq shiftsize 0) % case: multiple of bigit
             (B64move result (wplus2 movesize 2) big 2 lth)
             (wputv result 0 (mkitem hwords-tag lthnew))
             (when (eq big-preallocated* 0)
               (setq heaplast (aplus2 heaplast -1)))
             (return result) ) % trimbignum necessary
       (setq carry 0)
       (ifor (from i 1 lth 64)
            (do
             (setq d (wplus2 (wdifference lth i) 1))
             (when (wgreaterp d 64) (setq d 64))
             (VsetVl d)
             (Vload (Vreg 1) (aplus big i d) -1)
             (setq carry (B64registerLeftShift d shiftsize carry))
             (Vstore (Vreg 0) (aplus result movesize i d) -1)
       )    )
       (if (neq carry 0)
           (progn
            (setq lthnew (wplus2 lthnew 1))
            (iputv result lthnew carry))
            (when (eq big-preallocated* 0)
             (setq heaplast (aplus2 heaplast -1)))) %trimbignum
       (wputv result 0 (mkitem hwords-tag lthnew))
 
       (return result)  ))
 
%%%%%%%%%%%%%%%%%%%% Arithmetic Functions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% U, V, V1, V2 are Bignum arguments.
 
% Returns V1 if V1 is strictly less than 0, NIL otherwise.
 
(de bminusp (v1)
  (if (igetv v1 0) v1 nil))
 
% To provide a conveninent ADD with CARRY.
(de addcarry (a)
  (prog (s)
        (setq s (iplus2 a carry*))
        (if (igeq s bbase**)
          (setq carry* 1 s (idifference s bbase**))
          (setq carry* 0))
        (return s)))
 
(de subcarry (a)
  (prog (s)
        (setq s (idifference a carry*))
        (if (ilessp s 0)
          (setq carry* 1 s (iplus2 bbase** s))
          (setq carry* 0))
        (return s)))
 
(de bigremainder (u v)
  (checkifreallybig1 (bremainder u v)))
 
(de bremainder (v1 v2)
  (prog (l1 l2 q1 q2 r1 r2 )
        (setq l2 (openbsize v2))
        (when (izerop l2)
          (error 99 "Attempt to divide by 0 in BREMAINDER"))
        (setq l1 (openbsize v1))
        (cond ((ilessp l1 l2) (return v1))
              ((eq l1 l2)
                         (setq r2 (igetv v1 l1) q2 (igetv v2 l2))
                         (when (ilessp r2 q2) (return v1))))
    % This also takes care of case
    % when V1=0.
        (when (ionep l2)
         (return (cdr (bsimpledivide v1 l1 (igetv v2 1)
                                                (openbminusp v2)))))
        (when (eq l1 l2) (setq r1 (wgetv v1 l1) q1 (wgetv v2 l2))
                         (when (and (ilessp r1 q1)
                                    (eq r2 q2)) (return v1))
                         (when (eq r1 q1)
                           (return (cdr (bdivide v1 v2)))))
        %(when (and (wlessp l1 64) (wlessp l2 64))
              %(return (b64remainder v1 l1 v2 l2)))
        (return (cdr (bdivide v1 v2 )))))
 
(de biggcdn (v1 v2)
  (when (openbminusp v1) (setq v1 (bminus v1)))
  (when (openbminusp v2) (setq v2 (bminus v2)))
  (prog (l1 l2)
        (setq l2 (openbsize v2))
        (when (izerop l2) (return v1))
        (setq l1 (openbsize v1))
        (cond ((or (ilessp l1 l2)
                   (and (eq l1 l2) (ilessp (igetv v1 l1) (igetv v2 l2))))
                           (return (bgcdnaux v2 v1))))
        (return (bgcdnaux v2 (bremainder v1 v2)))))
 
(de bgcdnaux (v1 v2)
  (prog (l1 l2)
        (setq l2 (openbsize v2))
        (setq l1 (openbsize v1))
        (when (izerop l2) (return (checkifreallybig v1)))
        (when (and (wlessp l1 3) (wlessp l2 3))  % w-arith case
              (return (wgcdn1 v1 v2)))  %wgcdn1 is in arithmetic
       % (when (and (wlessp l1 64) (wlessp l2 64))
       %       (return (b64gcdn v1 l1 v2 l2)))
        (return (bgcdnaux v2 (bremainder v1 v2 )))))
 
(compiletime
(progn
(ds bsimplequotient (v1 l1 c snc) (car (bsimpledivide v1 l1 c snc)))
 
(ds bsimpleremainder (v1 l1 c snc) (cdr (bsimpledivide v1 l1 c snc)))
))
 
% C is a fixnum (inum?); V1 is a bignum and L1 is its length.
% SnC is T if C (which is positive) should be considered negative.
% Returns quotient . remainder; each is a bignum.
 
(de bsimpledivide (v1 l1 c snc)
  (prog (i p r rr sn1 v2 ii)
        (setq r 0)
        (setq i l1)
        (setq sn1 (openbminusp v1))
        (setq ii (igetv v1 l1))  %prefetch
        (if (or (and sn1 snc) (not (or sn1 snc)))
          (setq v2 (gtbig-open l1 nil))
          (setq v2 (gtbig-open l1 t)))
        (while (not (izerop i))
          (setq p (iplus2 (longtimes2 r bbase**) ii))
          % Overflow.
          (setq ii (wgetv v1 i))
          (iputv v2 i (longdivide-quotient (longdivide p c)))
          (setq r (longdivide-remainder))
          (setq i (isub1 i)))
          (setq rr (gtbig-open 1 sn1))
        (iputv rr 1 r)
        (return (cons (trimbignum1 v2 l1) (trimbignum1 rr 1)))))
 
 (de irecip (a n la bb1 bb2 bb3 bb4 bb5)
                             %IRECIP as in Collins,Mignotte and Winkler
    (prog (ex ex+1 i x x1 m a1 a2 a3 a4 bb34)
 
     (setq bb34 nil i 0)
     (setq big-preallocated* bb5)
     (setq ex+1 (btwopower (wplus2 la 1)))
     (setq a1 (openbsize a) a2 (wplus2 (wquotient la bbits**) 1))
     (setq a3 (wor (wlshift (igetv a a1) bbits**) (wgetv a a1)))
     (setq a4 (wlshift (wtwopower (wremainder la bbits**)) bbits**))
 
      % we have to make estimation by hand, not just
      % quotient (expt 2 la) a) because floating may be impossible
      % because of limited range
 
     (setq x (floatquotient (intfloat a4) (intfloat a3)))
     (setq x1 (wtimes2 (wdifference a2 a1) bbits**))
     (setq x (fastmkfltn (wplus2 (wgetv x 1) (wlshift 84 48))))
     (setq big-preallocated* bb1)
     (setq x (bigfromfloat x))
     (setq big-preallocated* bb3)
     (setq x (blshift x (wdifference x1 84)))
 
   lupe
 %   Newton step
 %   (setq x1 (blshift (btimes2 x (bdifference ex+1 (btimes2 a x)))
 %                                                   (wminus la)))
 %   this had to be rewritten to save heap space
 %   sorry!
 %
     (setq big-preallocated* bb1)
     (setq x1 (btimes2 a x))
     (setq big-preallocated* bb2)
     (setq x1 (bdifference ex+1 x1))
     (setq big-preallocated* bb1)
     (setq x1 (btimes2 x x1))
     (setq big-preallocated* (if bb34 bb3 bb4))
     (setq bb34 (not bb34))
     (setq x1 (blshift x1 (wminus la)))
     (setq big-preallocated* 0)
 
     (when (equal x x1) (return x)) %can be refined
     (setq x x1)
     (when (wlessp (setq i (wplus2 i 1)) n) (go lupe))
 
     (return (conterror 99 "Divergenz in BIG-Division" nil))
))
 
 
(compiletime (if_system Cray-2
  %clearing the frame, passing result
  %sequence must be *CALL, *DEALLOC, *END ....
  % NO exitopencode !!!!!
               (put 'clear-frame 'opencode
                      '((*Move 62 (reg a1))
                        (VsetVl (reg a1))
                        (*Vclear (vreg 5))
                        (move (reg a1) (reg locprevframesize))
                        (Vstorel (vreg 5) (reg a1))))
               (put 'clear-frame 'opencode
                      '((*Move 32 (reg a1))
                        (*move (reg symnam) (reg a0))
                            % symnam is 'clean' strings
                        (blklod (reg a1) (reg T30))))))
 
(de bharddivide (a la b lb)
 
 % Division using Newton approximation of reciprocal
 
 % Caution: This function uses illegal memory fields and therefore
 %          no garbage collection may occur while it is running AND
 %          the frame must be cleared afterwards
 
 (prog (sna snb estim bstrich astrich qstern rem
        bb1 bb2 bb3 bb4 bb5)
 
    (setq bb1 0 bb2 0 bb3 0 bb4 0 bb5 0) % was a BUG
 
    (setq sna (openbminusp a) snb (openbminusp b))
    (setf (wgetv a 1) nil) % put in positive for a moment
    (setf (wgetv b 1) nil)
                      %allocate some space for computation
    (setq estim (wtimes2 50 la)) % its a lot !
 
        % this is the dirty part
 
    (when (wlessp (wdifference heaptrapbound heaplast) estim) (reclaim))
 
    (when (wgreaterp (wdifference heaptrapbound heaplast) estim)
           (setq la (wquotient (wdifference heaptrapbound heaplast) 8))
           (setq bb1 (wplus2 heaplast (wtimes2 la 3)))
           (setq bb2 (wplus2 heaplast (wtimes2 4 la)))
           (setq bb3 (wplus2 heaplast (wtimes2 5 la)))
           (setq bb4 (wplus2 heaplast (wtimes2 6 la)))
           (setq bb5 (wplus2 heaplast (wtimes2 7 la))))
 
    (setq la (wtimes2 (openbsize a) 21) lb (wtimes2 (openbsize b) 21))
    (setq estim (wplus2 (wplus2 (wplus2 la (wminus lb)) 3) 63))
 
    (setq bstrich (irecip b estim la bb1 bb2 bb3 bb4 bb5))
         % produces approximation of reciprocal
    (setq big-preallocated* bb1)
    (setq astrich (blshift a (wminus (wdifference lb 24)))) %21 + 3
    (setq big-preallocated* bb2)
    (setq qstern  (btimes2 astrich bstrich))
    (setq big-preallocated* bb3)
    (setq qstern  (blshift qstern
                           (wdifference (wdifference lb la) 22)))
    (setq big-preallocated* bb1)
    (setq estim (sys2big 3))
    (setq big-preallocated* bb2)
    (setq qstern (bplus2 qstern estim))
    (setq big-preallocated* 0)
    (setq qstern (blshift qstern -2)) %produces "new" qstern
    (setq big-preallocated* bb3)
    (setq rem (btimes2 b qstern))
    (setq big-preallocated* 0)
    (setq rem (bdifference a rem))    %produces "new" remainder
 
    (when (openbminusp rem) (setq qstern (bdifference qstern bone*))
                        (setq rem (bplus2 rem b)))
    (when (equal rem b) (setq rem bzero*) (setq qstern (bplus2 qstern
              bone*))) % hilfskruecke !!
    (when (bgreaterp rem b) (setq rem (bdifference rem b))
                            (setq qstern (bplus2 qstern bone*))) %juhu
    (when (neq (openbsize qstern) 0) % setq signs of result
    (setf (wgetv qstern 1) (if (eq sna snb) nil t)))%set signs of result
    (when (neq (openbsize rem) 0) (setf (wgetv rem 1) sna))
 
    (setf (wgetv a 1) sna)  %restore signs
    (setf (wgetv b 1) snb)
    (return (clear-frame (cons qstern rem)))))
 
(de expt (x n)
  (cond
    ((not (intp n)) (nonintegererror n 'expt))
    ((izerop n) 1)
    ((ionep n) x)
    ((wlessp n 0) (quotient 1.0 (expt (float x) (wminus n))))
    (t
      ((lambda (ans)
        (cond
          ((izerop (wand n 1)) (times2 ans ans))
          (t                   (times2 (times2 ans x) ans))))
      (expt x (iquotient n 2))))))
 
%%%%%%%%%%%%%%%%%%%% compare functions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(de biggreaterp (u v)
  (checkifreallybigornil (bgreaterp u v)))
 
(de bgreaterp (u v)
 (prog (lu lv vs us)
  (setq us (openbminusp u))
  (setq vs (openbminusp v))
  (setq lu (openbsize u))
  (setq lv (openbsize v))
  (return
    (if us (cond ((not vs) nil)
                 ((eq lu lv) (bunsignedgreaterp v u lv lu))
                 ((wgreaterp lu lv) nil)
                 (t u))
           (cond (vs u)
                 ((eq lu lv) (bunsignedgreaterp u v lu lv))
                 ((wgreaterp lu lv) u)
                 (t nil))))))
 
(de biglessp (u v)
  (checkifreallybigornil (blessp u v)))
 
(de blessp (u v)
 (prog (lu lv vs us)
  (setq us (openbminusp u))
  (setq vs (openbminusp v))
  (setq lu (openbsize u))
  (setq lv (openbsize v))
  (return
    (if us (cond ((not vs) u)
                 ((eq lu lv) (bunsignedgreaterp u v lu lv))
                 ((wgreaterp lu lv) u)
                 (t nil))
           (cond (vs nil)
                 ((eq lu lv) (bunsignedgreaterp v u lv lu))
                 ((wgreaterp lu lv) nil)
                 (t u))))))
 
%(de bgeq (u v)
%  (cond ((openbminusp u) (if (openbminusp v)
%                         (bunsignedgeq v u) nil))
%        ((openbminusp v) u)
%        (t (bunsignedgeq u v))))
%
%(de bleq (u v)
%  (cond ((openbminusp u) (if (openbminusp v)
%                         (bunsignedgeq u v) u))
%        ((openbminusp v) nil)
%        (t (bunsignedgeq v u))))
 
(de bunsignedgreaterp (u v lu lv)
  % Compare magnitudes of two bignums;
  (prog (i nextu nextv)
        (if (eq lu lv)
              (setq nextu (igetv u lv)
                    nextv (igetv v lv))
            (if (igreaterp lu lv) (return u) (return nil)))
        (while (and (eq nextu nextv) (igreaterp lv 1))
             (setq nextu (wgetv u lv)
                   nextv (wgetv v lv))
             (setq lv (isub1 lv)))
        (if (igreaterp nextu nextv) (return u)
                                    (return nil))))
 
%(de bunsignedgeq (u v)
%  % Compare magnitudes of two unsigned bignums;
%  (prog (lu lv)
%        (setq lu (openbsize u))
%        (setq lv (openbsize v))
%        (unless (eq lu lv)
%          (if (igreaterp lu lv)
%            (return u)
%            (return nil)))
%        (while (and (eq (igetv u lv) (igetv v lv)) (igreaterp lv 1))
%          (setq lv (isub1 lv)))
%        (if (igreaterp (igetv v lv) (igetv u lv))
%          (return nil)
%          (return u))))
 
%%%%%%%%%%%%%%%%%%%%% Input and Output %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(setq digit2letter*
      '[48 49 50 51 52 53 54 55 56 57 65 66 67 68 69 70 71 72 73 74 75 76
        77 78 79 80 81 82 83 84 85 86 87 88 89 90])
 
% Ascii values of digits and characters.
% OutputBase* is assumed to be positive and less than 37.
%
% a new routine for printing bignums
%
% bchannelprin2 is still the entry point of course, but not much else
%
% if compressed bignum decompress it else
% first make a copy of v1 (the number to be printed), output any
% necessary base information then call bprin which actually does the work
 
(de bchannelprin2 (channel v1)
  ((lambda (v2 myobase i ob sn)
 
    (if (wgreaterp (inf (wgetv v1 1)) 128)
       (progn (setq v2 (bigdecompress v1))
              (setq sn (openbminusp v2))
              (wputv v2 1 'nil))
     (while (not (izerop i))
       (progn (iputv v2 i (igetv v1 i)) (setq i (isub1 i))))
     (setq sn (openbminusp v1)))
 
     (cond (sn (channelwritechar channel (char !-))))
     (cond
       ((neq myobase 10)
        (channelwritesysinteger channel myobase 10)
        (channelwritechar channel (char !#))))
     (setq ob (if (eq ob 10) 1000000
               (progn
                (setq ob (wtimes2 ob ob))  % ob should be < 6 bits
                (setq ob (longtimes2 (wtimes2 ob ob) ob)))))
     (bprin channel v2 ob)
     (setq outputbase!* myobase ))
 
   (gtbig-open (openbsize v1) nil)
    outputbase* (openbsize v1) outputbase* 0))
 
% divide v by 10**6, if the quotient is zero the recursion is terminated
% and the most siginifcant digit is printed (and so on down the stack)
% otherwise save the big digit (=remainder) and recurse
(de bprin (channel v ob)
  ((lambda (digit)
     (cond
       ((bzerop v) (channelwritesysinteger channel digit outputbase!*))
       (t (bprin channel v ob) (printdigit channel digit))))
   (bdivide1000000ip v ob)))
 
% print a base 10**6 digit
(de printdigit (channel d)
  (prog (d1 d2 d3 d4 d5 d6 digl ob)
 
    (setq digl digit2letter!* ob outputbase*) %local copy
 
    (setq d  (longdivide-quotient (longdivide d ob)))
    (setq d1 (longdivide-remainder))
    (setq d  (longdivide-quotient (longdivide d ob)))
    (setq d2 (longdivide-remainder))
    (setq d  (longdivide-quotient (longdivide d ob)))
    (setq d3 (longdivide-remainder))
    (setq d  (longdivide-quotient (longdivide d ob)))
    (setq d4 (longdivide-remainder))
    (setq d  (longdivide-quotient (longdivide d ob)))
    (setq d5 (longdivide-remainder))
    (setq d  (longdivide-quotient (longdivide d ob)))
    (setq d6 (longdivide-remainder))
 
    (channelwritechar channel (igetv digl d6))
    (channelwritechar channel (igetv digl d5))
    (channelwritechar channel (igetv digl d4))
    (channelwritechar channel (igetv digl d3))
    (channelwritechar channel (igetv digl d2))
    (return (channelwritechar channel (igetv digl d1)))))
 
% divide the bignum v1 (of length l1) by 100000, except the quotient is
% accumulated in the same place, the remainder, of course, ripples
% down to the bottom.  Because the argument is modified there is no need
% to CONS up a result, but simply return the remainder.
 
(de bdivide1000000ip (v1 ob)
  ((lambda (p r i)
    (prog (bb quot)
     (setq bb bbase**)
     (while (not (izerop i))
       (progn
         (setq p (iplus2 (longtimes2 r bb) (igetv v1 i)))
         (setq quot (longdivide-quotient (longdivide p ob)))
         (setq r (longdivide-remainder))
         (iputv v1 i quot)
         (setq i (isub1 i))))
     (trimbignum1 v1 (openbsize v1))
     (return (cond ((openbminusp v1) (iminus r)) (t r)))))
  nil 0 (openbsize v1)))
 
(de makestringintolispinteger (s radix sn)
  % radix is < Bbase*
  %s=string of digits, radix=base, sn=1 or -1
  (prog (sz res ch)
        (setq sz (size s))
        (setq res (gtbig-open 1 nil))
        (setq ch (strbyt s 0))
        (when (and (igeq ch (char a)) (ileq ch (char z)))
          (setq ch (iplus2 (idifference ch (char a)) 10)))
        (when (and (igeq ch (char 0)) (ileq ch (char 9)))
          (setq ch (idifference ch (char 0))))
        (iputv res 1 ch)
        (ifor (from i 1 sz 1)
              (do (progn (setq ch (strbyt s i))
                         (when (and (igeq ch (char a)) (ileq ch (char z)))
                           (setq ch
                            (idifference ch (idifference (char a) 10))))
                         (when (and (igeq ch (char 0)) (ileq ch (char 9)))
                           (setq ch (idifference ch (char 0))))
                         (setq res (bsmalltimes2 res radix ))
                         (setq res (bsmalladd res ch )))))
        (when (iminusp sn)
          (setq res (bminus res)))
        (return (checkifreallybig1 res))))
 
(de bsmalladd (v c)
  %V big, C fix.
  (cond ((izerop c) (return v))
        ((openbzerop v)  (return (int2big c)))
        ((openbminusp v) (bminus (bsmalldiff (bminus v) c)))
        ((iminusp c) (bsmalldiff v (iminus c)))
        (t (prog (v1 l1)
                 (setq carry* c)
                 (setq l1 (openbsize v))
                 (setq v1 (gtbig-open (iadd1 l1) nil ))
                 (ifor (from i 1 l1 1)
                       (do (iputv v1 i (addcarry (igetv v i)))))
                 (if (ionep carry*)
                   (iputv v1 (iadd1 l1) 1)
                   (return (trimbignum1 v1 l1)))
                 (return v1)))))
 
(de bnum (n)
  % Creates a Bignum of one BETA digit, value N.
  % N is POS or NEG
  (if (bigp n) n (bnumaux n)))
 
(de bnumaux (n)
  % Creates a Bignum of one BIGIT value N.
  % N is POS or NEG
  (prog (b)
        (cond ((izerop n) (return (gtbig-open 0 nil)))
              ((iminusp n) (setq b (gtbig-open 1 t))(setq n (iminus n)))
              (t (setq b (gtbig-open 1 nil))))
        (iputv b 1 n)
        (return b)))
 
(de bsmalldiff (v c)
  %V big, C fix
  (cond ((izerop c) v)
        ((openbzerop v) (int2big (iminus c)))
        ((openbminusp v) (bminus (bsmalladd (bminus v) c)))
        ((iminusp c) (bsmalladd v (iminus c)))
        (t (prog (v1 l1)
                 (setq carry* c)
                 (setq l1 (openbsize v))
                 (setq v1 (gtbig-open l1 nil))
                 (ifor (from i 1 l1 1)
                       (do (iputv v1 i (subcarry (igetv v i)))))
                 (unless (izerop carry*)
                   (stderror (bldmsg " BSmallDiff V<C %p %p%n" v c)))
                 (return (trimbignum1 v1 l1))))))
 
(compiletime (setq *syslisp nil))
(de int2big (n)
  % Creates BigNum of value N.
  % From any N, BETA,INUM,FIXNUM or BIGNUM
  (case (tag n)
        ((negint-tag posint-tag) (sys2big n))
        ((fixnum-tag) (sys2big (fixval (fixinf n))))
        ((bignum-tag) n)
        (nil (nonintegererror n 'int2big))))
(compiletime (setq *syslisp t))
 
 
% Now Install Interfacing
(de setupglobals ()
  (prin2t '"SetupGlobals")
  (setbits  44)  % (setbits bitsperword) CRAY's Fastintegers
  (prin2t '" ... done")
  nil)
 
%(setq staticbig* (gtwarray 5)) already done in arithmetic, caused problems
% Assume dont need more than 5 slots to represent a BigNum
% Version of SYSint
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Output %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(copyd 'oldchannelprin1 'recursivechannelprin1)
 
(copyd 'oldchannelprin2 'recursivechannelprin2)
 
(de recursivechannelprin1 (channel u level)
  (if (bigp u) (bchannelprin2 channel u)
               (oldchannelprin1 channel u level))
  u)
 
(de recursivechannelprin2 (channel u level)
  (if (bigp u) (bchannelprin2 channel u)
               (oldchannelprin2 channel u level))
  u)
 
(de checkifreallybig (uu)
        % If BIGNUM result is in FIXNUM or INUM range Convert Back.
    (checkifreallybig1 uu))
 
(de checkifreallybigpair (vv)     % Used to process DIVIDE
  (rplaca vv (checkifreallybig (car vv)))
  (rplacd vv (checkifreallybig (cdr vv)))
  vv)
 
(de checkifreallybigornil (uu)    % Used for EXTRA-boolean tests
    (if (null uu) NIL (checkifreallybig1 uu)))
 
(de bigadd1 (u)
  (checkifreallybig1 (bsmalladd u 1)))
 
(de bigsub1 (u)
  (checkifreallybig1 (bsmalldiff u 1)))
 
(de bigminus (u)
  (checkifreallybig1 (bminus u)))
 
(de bigminusp (u)
  (checkifreallybigornil (if (openbminusp u) u nil)))
 
(de bigonep (u)
  (checkifreallybigornil (bonep u)))
 
(de bigzerop (u)
  (checkifreallybigornil (bzerop u)))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Input %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (setq *syslisp nil))
(de int2sys (n)
  % Convert a random FIXed number to WORD Integer
  (case (tag n)
        ((posint-tag negint-tag) n)
        ((fixnum-tag) (fixval (fixinf n)))
        ((bignum-tag) (big2sysaux n))
        (nil (nonnumber1error n 'int2sys))))
(compiletime (setq *syslisp t))
 
(de sys2big (val)
   % sn nil for positive, T for negative
   (prog (sn val1 val2 val3)
    (when (wlessp val 0) (setq sn t) (setq val (wminus val)))
    (setq val1 (wand val bmask**))
    (setq val2 (wand (wrshift val bbits**) bmask**))
    (setq val3 (wand (wrshift val (times2 bbits** 2)) bmask**))
 
    (cond((wneq val3 0)       (setq sn (gtbig-open 3 sn))
                              (iputv sn 1 val1)
                              (iputv sn 2 val2)
                              (iputv sn 3 val3)
                              (return sn))
         ((wneq val2 0)       (setq sn (gtbig-open 2 sn))
                              (iputv sn 1 val1)
                              (iputv sn 2 val2)
                              (return sn))
         ((wneq val1 0)       (setq sn (gtbig-open 1 sn))
                              (iputv sn 1 val1)
                              (return sn))
         (t                   (return (gtbig-open 0 sn)))
)))
 
% Coercion/Transfer Functions
(copyd 'oldfloatfix 'floatfix)
 
(de floatfix (u)
  % Careful of sign and range
  (if (and (leq floatsyslow* u) (leq u floatsyshi*))
    (oldfloatfix u)
    (bigfromfloat u)))
 
(de betap (x)
  % test if NUMBER in reduced INUM range
  (if (and (wleq x betahi*) (wgeq x betalow*)) t nil))
 
(copyd 'lshift 'lshiftaux)
 
(copyd 'lsh 'lshift)
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Start of Vector codes for 64 items each
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% add two positive bignums
 
(de b64plus2aux (b1 b2 l1 l2 sign)
 
  (prog (l3 x)
 
       (when (wgreaterp l2 l1)
             (setq l3 l2)
             (setq l2 l1)
             (setq l1 l3)
             (setq x b2)
             (setq b2 b1)
             (setq b1 x) )
        %                    load operands
       (setq x (gtbig-open (wplus2 l1 1) sign))
       (VsetVL l1)
       (Vclear (vreg 2))
       (Vload (Vreg 1) (Aplus2 b1 2) 1)
       (VsetVl l2)
       (Vload (Vreg 2) (aplus2 b2 2) 1)
        %                    do the operation
       (setq l3 (B64RegisterAdd l1 l2))
       (VsetVl l3)
       (Vstore (Vreg 0) (aplus2 x 2) 1)
        %                    return the value
        (return (trimbignum1 x l3)) ))
 
% add the two positive bignum kernels in registers:
%       V0 = V1 + V2
%              V3 to V5 are destroyed
%              V1, V2, V6 , V7 remain unchanged
%       l1, l2: length of information in V1, V2
%               l2 not greater than l1
%               V2 filled up with trailing zeroes
%  function value:
%       length of result.
 
(de B64RegisterAdd (l1 l2)
     (prog (l3)
         %                calculate basical result
         (VsetVl l1)
         (VplusV (Vreg 0) (Vreg 1) (Vreg 2))
 
         %                test overflow bigits
         (VrshiftV (Vreg 3) (Vreg 0) bbits**)
         (setq l3 l1)
 
         %                case no overflow: return
         (when (eq (VtestN (Vreg 3)) 0) (return l3))
 
                        % delete overflow bits
         (VmoveV (Vreg 4) (Vreg 0) 0)  % move
         (VandS (Vreg 0) bmask** (Vreg 4))
  lupe
                        % case first bigit causes overflow: increase lth
         (when (not (eq (VgetElement (Vreg 3) (wplus2 l3 -1)) 0))
               (Vsetelement (Vreg 0) l3 0)
               (setq l3 (wplus2 l3 1)))
 
         %                shift the overflows to next bigit and add them
         (VsetVl l3)
         (VdrshiftV (Vreg 4) (Vreg 3) 64)
         (VplusV (Vreg 5) (Vreg 0) (Vreg 4))
 
         %                do testing with full length now
         (VrshiftV (Vreg 3) (Vreg 5) bbits**)
         (VandS (Vreg 0) bmask** (Vreg 5))
         (when (eq (VtestN (Vreg 3)) 0) (return l3))
         (go lupe)
))
 
 
%  Normalize the bigits of a bignum kernel by reducing overflowbits
%
% same routine as B64registeradd but only one operand in V1
% result in V0
% new length is function value
 
(de B64RegisterNormalize (l1)
     (prog (l3)
         %                calculate basical result
         (VsetVl l1)
         (VmoveV (Vreg 0) (Vreg 1))
 
         %                test overflow bigits
         (VrshiftV (Vreg 3) (Vreg 0) bbits**)
         (setq l3 l1)
 
         %                case no overflow: return
         (when (eq (VtestN (Vreg 3)) 0) (return l3))
 
                        % delete overflow bits
         (VmoveV (Vreg 4) (Vreg 0) 0)  % move
         (VandS (Vreg 0) bmask** (Vreg 4))
  lupe
                        % case first bigit causes overflow: increase lth
         (when (not (eq (VgetElement (Vreg 3) (wplus2 l3 -1)) 0))
               (Vsetelement (Vreg 0) l3 0)
               (setq l3 (wplus2 l3 1)))
 
         %                shift the overflows to next bigit and add them
         (VsetVl l3)
         (VdrshiftV (Vreg 4) (Vreg 3) 64)
         (VplusV (Vreg 5) (Vreg 0) (Vreg 4))
 
         %                do testing with full length now
         (VrshiftV (Vreg 3) (Vreg 5) bbits**)
         (VandS (Vreg 0) bmask** (Vreg 5))
         (when (eq (VtestN (Vreg 3)) 0) (return l3))
         (go lupe)
))
 
 
% subtract two positive bignum kernels with definitive positive result
 
(de b64differenceaux (b1 b2 l1 l2 sign)
           % calculate b1 - b2
           % sign: sign for the result
  (prog (l3 x)
       (when (wgeq l2 l1)
           (when (or (wgreaterp l2 l1) (bunsignedgreaterp b2 b1 l2 l1))
                     (setq x b1) (setq b1 b2) (setq b2 x)
                     (setq x l1) (setq l1 l2) (setq l2 x)
                     (setq sign (not sign))))
       (setq x (gtbig-open l1 sign))
       (VsetVL l1)
       (Vclear (vreg 2))
       (Vload (Vreg 1) (Aplus2 b1 2) 1)
       (VsetVl l2)
       (Vload (Vreg 2) (aplus2 b2 2) 1)
        %                    do the operation
       (setq l3 (B64RegisterSub l1 l2))
       (when (wgreaterp l3 0) (VsetVl l3)
                              (Vstore (Vreg 0) (aplus2 x 2) 1))
        %                    return the value
       (return (trimbignum1 x l3)) ))
 
% subtract two positive bignum kernels in registers:
%       V0 = V1 - V2
%              V3 to V7 auxilary
%       l1, l2: length of information in V1, V2
%               l2 not greater than l1
%               V2 filled up with trailing zeroes
%       the operands are allowed to have overflown bigits
%       the result is a normalized bignum:
%               - no bigits with overflow,
%               - no leading zero
%  function value:
%       length of result.
 
(de B64RegisterSub (l1 l2)
     (prog (l3)
         %                calculate basical result
         (VsetVl l1)
         (VminusV (Vreg 0) (Vreg 1) (Vreg 2))
 
         %                test underflow bigits
         (VrshiftV (Vreg 3) (Vreg 0) 63)        % sign bits now right
         (setq l3 l1)
 
  lupe
         (when (neq (VGetElement (Vreg 3) (aplus2 l1 -1)) 0)
                    (return (wminus l1)))
         %                case no overflow: return
         (when (eq (VtestN (Vreg 3)) 0) (go ready))
 
                        % delete overflow bits: add (bit * bbase)
         (VlshiftV (Vreg 5) (Vreg 3) bbits**)
         (VmoveV (Vreg 4) (Vreg 0) 0)  % move
         (VplusV (Vreg 0) (Vreg 5) (Vreg 4))
 
         %                shift the overflows to next bigit and add them
         (VdrshiftV (Vreg 4) (Vreg 3) 64)
         (VminusV (Vreg 5) (Vreg 0) (Vreg 4))
 
         %                testing underflow once more
         (VrshiftV (Vreg 3) (Vreg 5) 63)
         (VmoveV (Vreg 0) (Vreg 5))
         (go lupe)
 
ready
         %                reduce length for leading zeroes
         (while
             (and (wgreaterp l3 0)
                  (eq 0 (VgetElement (Vreg 0) (setq l1 (Wplus2 l3 -1)))))
             (setq l3 l1) )
         (return l3)
))
 
(de b64times2aux (b1 b2 l1 l2 sign)
           % calculate b1 * b2
           % sign: sign for the result
  (prog (carry b3 l3 x digit l4)
       (setq carry 0)
       (setq l3 (wplus2 l1 l2)) % result length
       (setq x (getmem (aplus2 b2 2)))
       (setq b3 (opengtbignum l3 sign))
       %                      load first operand, convert to float, V7
       (VsetVL l1)
       (Vload (Vreg 2) (Aplus2 b1 2) 1)
       (Vclear (vreg 1))
       (VlShiftV (Vreg 7) (Vreg 2) 24)
        %                     loop over b2 bigits
       (setq l4 (wplus2 l2 -1))
       (ifor (from i 1 l4 1)
             (do
               (VFtimesS (Vreg 6) (Vreg 7) (wlshift x 24))
                        %                 add to intermediate sum
               (Vplusv (Vreg 0) (Vreg 1)(Vreg 6))
               %                 take the last bigit for the result
               (setq digit (wplus2 (VgetElement (Vreg 0) 0) carry))
               (setq x (getmem (aplus2 (wplus2 b2 i) 2)))
               (Putmem (aplus2 (wplus2 b3 i) 1) (wand digit bmask**))
 
               (setq carry (wrshift digit bbits**))
               %                 shift the intermediate sum by one pos.
               (VdlshiftV (Vreg 1) (Vreg 0) 64)
       )      )
       % loop unrolled
               (VFtimesS (Vreg 6) (Vreg 7) (wlshift x 24))
                        %                 add to intermediate sum
               (Vplusv (Vreg 0) (Vreg 1)(Vreg 6))
               %                 take the last bigit for the result
               (setq digit (wplus2 (VgetElement (Vreg 0) 0) carry))
               (Putmem (aplus2 (wplus2 b3 l2) 1) (wand digit bmask**))
 
               (setq carry (wrshift digit bbits**))
               %                 shift the intermediate sum by one pos.
               (VdlshiftV (Vreg 1) (Vreg 0) 64)
       % store the last intermediate sum to the result
       % carries have to be reported
 
       (VsetElement (Vreg 1) 0
           (wplus2 carry (VgetElement (Vreg 1) 0)))
 
      % adding carries (two times unconditional
 
       (VSetVl l1)
       (VrshiftV (Vreg 3) (Vreg 1) bbits**)
       (VandS (vreg 2) bmask** (Vreg 1))
       (VdrshiftV (Vreg 4) (Vreg 3) 64)
       (VplusV (Vreg 1) (Vreg 2) (Vreg 4))
       (VrshiftV (Vreg 3) (Vreg 1) bbits**)
    lll
       (VandS (vreg 2) bmask** (Vreg 1))
       (VdrshiftV (Vreg 4) (Vreg 3) 64)
       (VplusV (Vreg 1) (Vreg 2) (Vreg 4))
       (VrshiftV (Vreg 3) (Vreg 1) bbits**)
       (when (neq (Vtestn (Vreg 3)) 0) (go lll))
 
       %do trimbignum
 
       (when (eq (Vgetelement (Vreg 1) (aplus2 l1 -1)) 0)
            (setq l1 (wplus2 l1 -1))
            (when (eq big-preallocated* 0)
              (setq heaplast (aplus2 heaplast -1)))
            (setq l3 (wplus2 l3 -1)))
 
       (VsetVl l1)
 
       % finally make bignum ready
 
       (Putmem b3 (mkitem hwords-tag l3))
       (when (neq l1 0) (Vstore (Vreg 1) (aplus2 (wplus2 b3 l2) 2) 1))
       (return b3)
 ))
 
% auxilary function: convert the bigits of a bignum as preparation
% for subsequent mulitplications
 
% special multiplication routine: b2 to be converted to a sequence
% of float digits already
 
(de b64specialregistertimes (l1 b2 l2)
           % calculate vreg0 = vreg1(length l1) * b2 b1;
           % result: new length
  (let ((start 0)
        (carry 0)
        b3 l3 l4 x digit)
       (setq l3 (wplus2 l1 l2)) % result length
       %                      load first operand, convert to float, V7
       (VsetVL l1)
       (VlShiftV (Vreg 0) (Vreg 1) 24)
       (setq digit l1)
       (setq l1 l2) (setq l2 digit)
 
       (VsetVl l1)
       (Vload (Vreg 4)(aplus2 b2 2) 1)       % the second operand
       (VlShiftV (Vreg 1) (Vreg 4) 24)
       (Vclear (vreg 3))
       (while (and (wlessp start l2)
                   (eq 0 (setq x (VgetElement (Vreg 0) start))))
          (setq start (wplus2 1 start)))
 
        %                     loop over b2 bigits
       (ifor (from i (wplus2 1 start) l2 1)
             (do
               (VFtimesS (Vreg 5) (Vreg 1) x)
     %                 add to intermediate sum
               (Vplusv (Vreg 4) (Vreg 3)(Vreg 5))
 
               (setq x (VgetElement (Vreg 0) i )) %preload
               %                 take the last bigit for the result
               (setq digit (wplus2 (VgetElement (Vreg 4) 0) carry))
               (VsetElement (Vreg 0) (aplus2 i -1)
                       (wand digit bmask**))
               (setq carry (wrshift digit bbits**))
               %                 shift the intermediate sum by one pos.
               (VdlshiftV (Vreg 3) (Vreg 4) 64)
       )      )
       % store the last intermediate sum to the result
       % carries have to be reported
 
       (setq digit (VgetElement (Vreg 3) 0)) %preload
       (ifor (from i 1 l1 1)
             (do
               (setq digit (wplus2 digit carry))
               (Vsetelement (Vreg 0) (aplus l2 i -1)
                       (wand digit bmask**))
               (setq carry (wrshift digit bbits**))
               (setq digit (VgetElement (Vreg 3) i))
       )     )
       (return l3)
 ))
 
% mult bignum kernel in reg V1 by one integer number
% result in V0
% value is length of V0
 
(de b64smalltimes2aux (b1 x sign)
           % calculate b1 * x
           % sign: sign for the result
  (let ((l1 (inf (getmem (inf b1))))
        b3 l3 l4)
       (setq l3 (wplus2 l1 1)) % result length
       (setq b3 (gtbig-open l3 sign))
       %                      load first operand, convert to float, V7
       (VsetVL l3)
       (Vclear (Vreg 1))
       (VsetVL l1)
               %                 convert to float
       (when (wlessp x 0) (setq x (wminus x)))
       (setq x (wlshift x 24))
 
       (Vload (Vreg 6) (Aplus2 b1 2) 1)
       (VlshiftV (Vreg 7) (Vreg 6) 24)
               %                 multiply
       (VFtimesS (Vreg 2) (Vreg 7) x)
               %                 add to intermediate sum
       (setq l4 (B64RegisterAdd l1 l1))
       % store the last intermediate sum to the result
       (Vsetvl l4)
       (Vstore (Vreg 0) (aplus2 b3 2) 1)
       (return (trimbignum1 b3 l4))
 ))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                       ----
%  interface to PSL standard BIG module
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 
(de bminus (v1)
    (prog (steplth lng sn res)
       (VsetVl 2)
       (Vload (Vreg 1) v1 1)
       (setq lng (inf (VgetElement (Vreg 1) 0)))
       (setq sn (VgetElement (Vreg 1) 1))
       (setq res (gtbig-open lng (not sn)))
       (ifor (from i 1 lng 64)
             (do
                (setq steplth (wplus2 (wdifference lng i) 1))
                (when (wgreaterp steplth 64) (setq steplth 64))
                (VsetVl steplth)
                (Vload (Vreg 1)  (aplus v1 i 1) 1)
                (Vstore (Vreg 1) (aplus res i 1) 1)
      )      )
      (return res)
))
 
(de bigplus2 (u v)
  (checkifreallybig1 (bplus2 u v)))
 
(de bplus2 (v1 v2)
  (prog (sn1 sn2 l1 l2)
           % read control information
        (VsetVl 2)
        (Vload (Vreg 1) v1 1)
        (Vload (Vreg 2) v2 1)
        (setq sn1 (VgetElement (Vreg 1) 1))
        (setq sn2 (VgetElement (Vreg 2) 1))
        (setq l1 (inf (VgetElement (Vreg 1) 0)))
        (setq l2 (inf (VgetElement (Vreg 2) 0)))
        (when (eq l1 0) (return v2))
        (when (eq l2 0) (return v1))
        (when (and sn1 (not sn2))
             (return (bdifference v2 (bminus v1) nil)))
        (when (and sn2 (not sn1))
             (return (bdifference v1 (bminus v2) nil)))
        (when (and (wlessp l1 64) (wlessp l2 64))
              (return (b64plus2aux v1 v2 l1 l2 sn1)))
        (return (b99plus2aux v1 v2 l1 l2 sn1))))
 
(de bigdifference (u v)
  (checkifreallybig1 (bdifference u v)))
 
(de bdifference (v1 v2)
        (prog (sn1 sn2 l1 l2)
              (VsetVl 2)
              (Vload (Vreg 1) v1 1)
              (Vload (Vreg 2) v2 1)
              (setq sn1 (VgetElement (Vreg 1) 1))
              (setq sn2 (VgetElement (Vreg 2) 1))
              (setq l1 (inf (VgetElement (Vreg 1) 0)))
              (setq l2 (inf (VgetElement (Vreg 2) 0)))
              (when (eq l2 0) (return v1))
              (when (eq l1 0) (return (bminus v2)))
%             (when (or (and sn1 (not sn2)) (and sn2 (not sn1)))
              (when (not (eq sn1 sn2))
                   (if (and (wlessp l1 64) (wlessp l2 64))
                      (return (b64plus2aux v1 (bminus v2) l1 l2 sn1))
                      (return (b99plus2aux v1 (bminus v2) l1 l2 sn1))))
              (when (and (wlessp l1 64) (wlessp l2 64))
                       (return (b64differenceaux v1 v2 l1 l2 sn1)))
              (return (b99differenceaux v1 v2 l1 l2 sn1))))
 
(de bigtimes2 (u v)
  (checkifreallybig1 (btimes2 u v)))
 
(de btimes2 (v1 v2)
   (prog (sn1 sn2 l1 l2 sign x)
        (VsetVl 2)
        (Vload (Vreg 1) v1 1)
        (Vload (Vreg 2) v2 1)
        (setq sn1 (VgetElement (Vreg 1) 1))
        (setq sn2 (VgetElement (Vreg 2) 1))
        (setq l1 (inf (VgetElement (Vreg 1) 0)))
        (setq l2 (inf (VgetElement (Vreg 2) 0)))
        (setq sign (if sn1 (not sn2) sn2))
        (when (or (eq l1 0) (eq l2 0)) (return (gtbig-open 0 nil)))
        (when (wgreaterp l2 l1)
             (setq x l2) (setq l2 l1) (setq l1 x)
             (setq x v1) (setq v1 v2) (setq v2 x))
        (cond
            ((wlessp l1 64)
               (return (b64times2aux v1 v2 l1 l2 sign)))
            ((wgreaterp l2 63)
               (return (b99times2aux v1 v2 l1 l2 sign)))
            ((wgreaterp l2 5)
               (return (b64times2aux v2 v1 l2 l1 sign)))
            (t (return (b99times2aux v1 v2 l1 l2 sign)))
)))
 
(de bsmalltimes2 (v1 c)
  % V1 is a BigNum, C a fixnum.
  % Assume C positive, ignore sign(V1)
    (prog (j carry l1 l2 l3 v3)
          (setq l1 (openbsize v1))
          (when (or (eq c 0) (eq l1 0)) (return (gtbig-open 0 nil)))
          (when (wlessp l1 64)
                 (return (b64smalltimes2aux v1 c nil)))
          (setq l2 (iplus2 (longquotient c bbase**) l1))
          (setq l3 (iadd1 l2))
          (setq v3 (gtbig-open l3 nil))
          (setq carry 0)
          (ifor (from h 1 l1 1)
                (do
                  (progn (setq j
                          (iplus2 (longtimes2 (igetv v1 h) c) carry))
                          (iputv v3 h (iremainder j bbase**))
                          (setq carry (longquotient j bbase**)))))
          (ifor (from h (iadd1 l1) l3 1)
                   (do
                     (progn (iputv v3 h
                          (iremainder (setq j carry) bbase**))
                          (setq carry (longquotient j bbase**)))))
          (return (trimbignum1 v3 l3))))
 
(de bigdivide (u v)
  (checkifreallybigpair (bdivide u v)))
 
(de bigquotient (u v)
  (checkifreallybig1 (car (bdivide u v))))
 
(de bdivide (v1 v2)
  (prog (l1 l2 q r v3)
        (setq l2 (openbsize v2))
        (when (izerop l2)
          (error 99 "Attempt to divide by 0 in BDIVIDE"))
        (setq l1 (openbsize v1))
        (cond ((ilessp l1 l2)
 
 % may use bzero*  here, because checkifreallybig will
 % make sysints from them
 
                   (return (cons bzero* v1)))
              ((eq l1 l2) (when (wordsequal v1 v2) (return
                                  (cons bone* bzero*)))
                          (when (ilessp (igetv v1 l1) (igetv v2 l2))
                                 (return (cons bzero* v1)))))
               % This also takes care of case
               % when V1=0.
        (when (ionep l2)
          (return (bsimpledivide v1 l1 (igetv v2 1) (openbminusp v2))))
        (when (eq l1 l2) (when (and(ilessp (wgetv v1 l1) (wgetv v2 l2))
                                (eq (igetv v1 l1) (igetv v2 l2)))
                           (return (cons bzero* v1)))
                         (when (eq (wgetv v1 l1) (wgetv v2 l2))
                           (return (bharddivide v1 l1 v2 l2 ))))
      % (when (and (wlessp l1 64) (wlessp l2 64))
      %  (return (b64harddivide v1 (openbsize v1) v2 (openbsize v2))))
        (return (bharddivide v1 l1 v2 l2))))
 
    % move lth bigits of b beginning with posb to a beginning with posa
 
(de B64move(a posa b posb lth)
    (prog (steplth lthplus1)
       (setq posa (wplus2 posa -1))
       (setq posb (wplus2 posb -1))
       (setq lthplus1 (wplus2 lth 1))
       (ifor (from i 1 lth 64)
             (do
                (setq steplth (wdifference lthplus1 i))
                (when (wgreaterp steplth 64) (setq steplth 64))
                (VsetVl steplth)
                (Vload (Vreg 1) (aplus  b posb i) 1)
                (Vstore (Vreg 1) (aplus a posa i) 1)
      )      )
      (return a)))
 
(compiletime (progn
  (put 'openFtimes2 'opencode
     '((fmul (reg 1) (reg 1) (reg 2))))
 
(put 'Floatquotient  'opencode
     '((*move (reg 2) (reg a6))
       (*move (memory (reg a6) 1) (reg 2))
       (*move (reg 1) (reg a6))
       (*move (memory (reg a6) 1) (reg 1))
       (hrecip (reg 3) (reg 2))
       (fmul  (reg 1) (reg 1) (reg 3))
       (imul  (reg 2) (reg 2) (reg 3))
       (fmul  (reg 1) (reg 1) (reg 2))
)     )
))
 
(de B64harddivide (a la b lb)
    (prog (x flb lres l1 sn lx sna snb baux estim lresold)
     (VSetVl 64)
     (VClear (Vreg 7))
     (setq snb (wgetv b 1))
 
     (VsetVl 2)
     (Vload (Vreg 5) a 1)
     (setq flb (floatquotient 1.0 (unsignedFloatfromBignum b)))
     (setq sna (Vgetelement (Vreg 5) 1))
     (VsetVl la)
     (Vload (Vreg 6) (aplus2 a 2) 1)
     (setq sn (not (eq sna snb)))
     (setq estim (openftimes2 (B64floatfromregister la) flb))
   first-correction
     (setq lres (b64registerfromfloat estim))
     (VSetVl lres)
     (VmoveV (Vreg 7) (Vreg 0))
     (setq baux b)
     (VSetVl lres)
     (VmoveV (Vreg 1) (Vreg 7))
     (setq l1 (B64SpecialRegisterTimes lres baux lb))
     (VsetVl la)
     (Vclear (Vreg 2))
     (VMoveV (Vreg 1) (Vreg 6))
     (VsetVl l1)
     (VmoveV (Vreg 2) (Vreg 0))
     (setq la (b64registersub la l1))
     (when (wlessp la 0)
           (setq la (wminus la))
           (setq estim (wdifference estim 1))
           (go first-correction))
     (VsetVl la)
     (VMoveV (Vreg 6) (Vreg 0))
 
     (when (wleq la lb)
       (when (or (wlessp la lb)
           (blessp-v la lb b)) (return (b64bigfromV7 lres sn la sna))))
 
   loop
     (setq estim (openftimes2 (B64floatfromregister la) flb))
   loop-correction
     (setq lx (b64registerfromfloat estim))
     (VSetVl (setq lresold lres))
     (VClear (Vreg 2))
     (VmoveV (Vreg 1) (Vreg 7))
     (VStore (Vreg 7) b64save1 1)
     (VsetVl lx)
     (VmoveV (Vreg 2) (Vreg 0))
     (setq lres (B64registeradd lres lx))
     (VSetVl lres)
     (VMoveV (Vreg 7) (Vreg 0))
     (VsetVl lx)
     (VMoveV (Vreg 1) (Vreg 2))
     (setq l1 (B64SpecialRegisterTimes lx baux lb))
     (VsetVl la)
     (Vclear (Vreg 2))
     (VMoveV (Vreg 1) (Vreg 6))
     (VsetVl l1)
     (VmoveV (Vreg 2) (Vreg 0))
     (setq la (b64registersub la l1))
     (when (wlessp la 0)
           (setq la (wminus la))
           (setq estim (wdifference estim 1))
           (VSetVl (setq lres lresold))
           (Vload (Vreg 7) b64save1 1)
           (go loop-correction))
     (VsetVl la)
     (VMoveV (Vreg 6) (Vreg 0))
 
     (when (wgreaterp la lb) (go loop))
     (when (or (wlessp la lb)
            (blessp-v la lb b)) (return (B64bigfromV7 lres sn la sna)))
     (go loop)))
 
(de B64remainder (a la b lb)
 %
 % remainder for 64 type bignums
 %
    (prog (x flb lres l1 sn lx sna baux estim)
 
     (VsetVl 2)
     (Vload (Vreg 5) a 1)
     (setq flb (floatquotient 1.0 (unsignedFloatfromBignum b)))
     (setq sna (Vgetelement (Vreg 5) 1))
     (VsetVl la)
     (Vload (Vreg 6) (aplus2 a 2) 1)
     (setq baux b)
 
     (setq estim (openftimes2 (B64floatfromregister la) flb))
   first-correction
     (setq lres (b64registerfromfloat estim))
     (VSetVl lres)
     (VmoveV (Vreg 1) (Vreg 0))
     (setq l1 (B64SpecialRegisterTimes lres baux lb))
     (VsetVl la)
     (Vclear (Vreg 2))
     (VMoveV (Vreg 1) (Vreg 6))
     (VsetVl l1)
     (VmoveV (Vreg 2) (Vreg 0))
     (setq la (b64registersub la l1))
     (when (wlessp la 0)
           (setq la (wminus la))
           (setq estim (wdifference estim 1))
           (go first-correction))
     (VsetVl la)
     (VMoveV (Vreg 6) (Vreg 0))
 
     (when (wleq la lb)
       (when (or (wlessp la lb)
                 (blessp-v la lb b)) (return (b64bigfromV6 la sna))))
 
   loop
     (setq estim (openftimes2 (B64floatfromregister la) flb))
   loop-correction
     (setq lx (b64registerfromfloat estim))
     (VsetVl lx)
     (VMoveV (Vreg 1) (Vreg 0))
     (setq l1 (B64SpecialRegisterTimes lx baux lb))
     (VsetVl la)
     (Vclear (Vreg 2))
     (VMoveV (Vreg 1) (Vreg 6))
     (VsetVl l1)
     (VmoveV (Vreg 2) (Vreg 0))
     (setq la (b64registersub la l1))
     (when (wlessp la 0)
           (setq la (wminus la))
           (setq estim (wdifference estim 1))
           (go loop-correction))
     (VsetVl la)
     (VMoveV (Vreg 6) (Vreg 0))
 
     (when (wgreaterp la lb) (go loop))
     (when (or (wlessp la lb)
               (blessp-v la lb b)) (return (B64bigfromV6 la sna)))
     (go loop)))
 
(de B64gcdn (a la b lb)
 %
 % gcdn for 64 type bignums
 %
    (prog (x flb lres l1 sn lx baux baux1 baux2 estim)
                (setq baux1 (mkbign (inf b64save4)))
                (setq baux2 (mkbign (inf b64save5)))
 
   step1 
     (VsetVl la)
     (Vload (Vreg 6) (aplus2 a 2) 1)
 
     (setq baux b)
 
     (setq flb (floatquotient 1.0 (unsignedFloatfromBignum b)))
     (setq estim (openftimes2 (B64floatfromregister la) flb))
   first-correction
     (setq lres (b64registerfromfloat estim))
     (VSetVl lres)
     (VmoveV (Vreg 1) (Vreg 0))
     (setq l1 (B64SpecialRegisterTimes lres baux lb))
     (VsetVl la)
     (Vclear (Vreg 2))
     (VMoveV (Vreg 1) (Vreg 6))
     (VsetVl l1)
     (VmoveV (Vreg 2) (Vreg 0))
     (setq la (b64registersub la l1))
     (when (wlessp la 0)
           (setq la (wminus la))
           (setq estim (wdifference estim 1))
           (go first-correction))
     (VsetVl la)
     (VMoveV (Vreg 6) (Vreg 0))
 
     (when (wleq la lb)
       (when (or (wlessp la lb)
                 (blessp-v la lb b)) (go nextremainder)))
 
   loop
     (setq estim (openftimes2 (B64floatfromregister la) flb))
   loop-correction
     (setq lx (b64registerfromfloat estim))
     (VsetVl lx)
     (VMoveV (Vreg 1) (Vreg 0))
     (setq l1 (B64SpecialRegisterTimes lx baux lb))
     (VsetVl la)
     (Vclear (Vreg 2))
     (VMoveV (Vreg 1) (Vreg 6))
     (VsetVl l1)
     (VmoveV (Vreg 2) (Vreg 0))
     (setq la (b64registersub la l1))
     (when (wlessp la 0)
           (setq la (wminus la))
           (setq estim (wdifference estim 1))
           (go loop-correction))
     (VsetVl la)
     (VMoveV (Vreg 6) (Vreg 0))
 
     (when (or (wgreaterp la lb)
             (not (or (wlessp la lb) (blessp-v la lb b)))) (go loop))
 
   nextremainder
 
  % passed 1 remainder step
  % check whether we are ready (remainder = 0)
  % or rest of computation may de done in fix-arith (length's <=2)
 
     (when (eq la 0)
           (setq a (gtbig-open lb nil))
           (vector-copy b a (wplus2 lb 2))
%%%%%%     (if (wgreaterp lb 3) (return a)
                                (return (checkifreallybig1 a)))
 
     (when (and (wlessp la 3) (wlessp lb 3))
          (return (wgcdn1 b (b64bigfromV6 la nil))))
 
  % else doit again
 
     (setq x baux2) (setq baux2 baux1) (setq baux1 x)
     (VsetVl la)
     (VStore (Vreg 6) (aplus2 baux1 2) 1)
     (setq x la)
     (WputV baux1 0 (mkitem hwords-tag la))
     (WputV baux1 1 nil)
     (setq a b la lb)
     (setq b baux1 lb x)
     (go step1)
))
 
%(de div2 (a b n) % this is the portable kernel of Newton iteration
%   (prog (ex ex+1 x x1 m) % for bharddivide/Irecip
%
%    (setq ex (expt 2 n))
%    (setq ex+1 (expt 2 (add1 n)))
%    (setq x (fix (quotient (float ex) (float b))))
%    (print (list "startwert" x))
%  (for (from m 1 10 1) (do
%    (setq x1 (quotient (times x (difference ex+1 (times b x))) ex))
%    (setq x x1)
%    (print (list "naeherung:" x1))))
%    (return (quotient (add1 (times a x1)) ex))
%))
 
(compiletime (put 'openfloatplus2 'opencode
                '((fadd (reg 1) (reg 1) (reg 2)))))
 
(de blessp-v (la lb b)
             (b64registerunsigned-geq-v  b la nil))
 
(de b64registerunsigned-geq-v( b la retval)
    (prog  (vala valb)
     (while (and (eq (setq vala (Vgetelement (Vreg 6) (aplus2 la -1)))
                     (setq valb (igetv b la))) (wgreaterp la 1))
         (setq la (wplus2 la -1)))
     (return (if(wgeq vala valb) retval (not retval)))))
 
 
(de bunsigned-greaterp-v(a b la retval)
    (prog  (vala valb)
     (while (and (eq (setq vala (igetv a la))
                     (setq valb (igetv b la)))(wgreaterp la 1))
         (setq la (wplus2 la -1)))
     (return (if(wgreaterp vala valb) retval (not retval)))))
 
(compiletime
(ds floathack(digit pos)
    (wor (wlshift
           (wplus2
              (wplus2 8#40000 (wtimes2 pos bbits**))
              (wdifference 48 bbits**))
           48)
         digit))
)
 
(de FloatfromBignum (b)
     (prog (s x i sn)
       (when (wgreaterp (inf (wgetv b 1)) 128)
		(setq b (bigdecompress b)))
       (VsetVl 2)
       (VLoad (Vreg 1) b 1)
       (VsetVl 3)
       (setq s (inf (Vgetelement (Vreg 1) 0)))
       (Vload (Vreg 2) (aplus b s 1) -1)
       (setq sn (Vgetelement (Vreg 1) 1))
       (setq i (Vgetelement (Vreg 2) 0))
       (setq x (openfloatplus2 0 (floathack i s))) % normalize
       (setq s (wplus2 s -1))
       (when (eq s 0) (go exit))
       (setq i (Vgetelement (Vreg 2) 1))
       (setq x (openfloatplus2 x (floathack i s)))
       (setq s (wplus2 s -1))
       (when (eq s 0) (go exit))
       (setq i (Vgetelement (Vreg 2) 2))
       (setq x (openfloatplus2 x (floathack i s)))
    exit 
       (setq x (fastmkfltn x)) % put it into a box
       (if sn (return (floatdifference 0.0 x)) (return x))
))
 
(de unsignedFloatfromBignum (b)
     (prog (s x i sn)
       (setq s (inf (getmem b)))  % length
       (VsetVl 4)
       (Vload (Vreg 2) (aplus b s 1) -1)
       (setq i (Vgetelement (Vreg 2) 0))
       (setq x (openfloatplus2 0 (floathack i s))) % normalize
       (setq s (wplus2 s -1))
       (when (eq s 0) (go exit))
       (setq i (Vgetelement (Vreg 2) 1))
       (setq x (openfloatplus2 x (floathack i s)))
       (setq s (wplus2 s -1))
       (when (eq s 0) (go exit))
       (setq i (Vgetelement (Vreg 2) 2))
       (setq x (openfloatplus2 x (floathack i s)))
       (setq s (wplus2 s -1))
       (when (eq s 0) (go exit))
       (setq i (Vgetelement (Vreg 2) 3))
       (setq x (openfloatplus2 x (floathack i s)))
    exit 
       (setq x (wplus2 x 1))   % for safety
       (setq x (fastmkfltn x)) % put it into a box
       (return x)
))
 
(de b64floatfromregister (len)
%
% expects input in Vreg 6 !!
%
     (prog (s x i)
       (setq s len)
       (setq i (Vgetelement (Vreg 6) (aplus2 len -1)))
       (setq x (openfloatplus2 0 (floathack i s))) % normalize
       (setq s (wplus2 s -1))
       (when (eq s 0) (go exit))
       (setq i (Vgetelement (Vreg 6) (aplus2 len -2)))
       (setq x (openfloatplus2 x (floathack i s)))
       (setq s (wplus2 s -1))
       (when (eq s 0) (go exit))
       (setq i (Vgetelement (Vreg 6) (aplus2 len -3)))
       (setq x (openfloatplus2 x (floathack i s)))
    exit 
       (return x)
))
 
% Convert BIGNUMs to FLOAT
(de bigfromfloat (x)
     (prog (exp mantisse sn val result nn nn2 rest)
        (setq val (wgetv x 1))
        (setq sn (eq (wrshift val 63) 1))
        (setq mantisse (wrshift (wlshift val 16) 16))
        (setq exp (wdifference (wrshift (wlshift val 1) 49) 8#40000))
           % remove sign
        (when (wleq exp 0) (return 0))
        (when (wleq exp 48)
          (setq val (wrshift mantisse (wdifference 48 exp)))
          (return (sys2big-i exp val sn)))
 
        (setq nn (longdivide-quotient (longdivide exp bbits**)))
        (setq rest (longdivide-remainder))
        (setq nn (wplus2 nn 1))
        (setq result (gtbig-open (if (eq rest 0) (wplus2 nn -1) nn)sn))
        (VSetVl (if (wlessp nn 65) nn 64))
        (Vclear (Vreg 1))
        (VStore (Vreg 1) (aplus2 result 2) 1)
        (setq nn2 (wplus2 nn 1))
        (Ifor (from i 66 nn2 1) (do (wputv result i 0)))
 
        (when (neq rest 0)
          (iputv result nn (wrshift mantisse (wdifference 48 rest))))
        (setq rest (wplus2 rest bbits**))
        (iputv result (wplus2 nn -1)
               (wand (wshift mantisse (wdifference rest 48)) bmask**))
        (setq rest (wplus2 rest bbits**))
        (iputv result (wplus2 nn -2)
               (wand (wshift mantisse (wdifference rest 48)) bmask**))
        (setq rest (wplus2 rest bbits**))
        (when (wgreaterp nn 3)
          (iputv result (wplus2 nn -3)
               (wand (wshift mantisse (wdifference rest 48)) bmask**)))
        (return result)
))
 
(de b64registerfromfloat (x)
     (prog (exp mantisse nn rest rest1)
        (setq exp (wdifference (wrshift x 48) 8#40000))
        (when (wleq exp 0) (return (b64makeone)))
     % must produce a 1 instead of a 0.9999999999 -> 0 !!
 
        (setq nn (longdivide-quotient (longdivide exp bbits**)))
        (setq rest (setq rest1 (longdivide-remainder)))
        (setq nn (wplus2 nn 1))
        (setq rest (wdifference rest 48))
        (setq mantisse (wrshift (wlshift x 16) 16))
        (VSetElement (Vreg 0) (aplus2 nn -1)
                              (wrshift mantisse (wminus rest)))
        (setq rest (wplus2 rest bbits**))
        (VSetElement (Vreg 0) (aplus2 nn -2)
               (wand (wrshift mantisse (wminus rest)) bmask**))
        (setq rest (wplus2 rest bbits**))
        (VSetElement (Vreg 0) (aplus2 nn -3)
               (wand (wshift mantisse rest) bmask**))
        (setq rest (wplus2 rest bbits**))
        (VSetElement (Vreg 0) (aplus2 nn -4)
               (wand (wshift mantisse rest) bmask**))
        (when (wgreaterp nn 4)
           (VSetVl (aplus2 nn -4))
           (Vclear (Vreg 0)))
        (when (eq rest1 0) (return (wplus2 nn -1)))
        (return nn)
))
 
(de b64makeone ()
  %                   a big one into Vreg 0 , return 1 !
    (VsetElement (Vreg 0) 0 1)
    1)
 
(de b64bigfromV7 (len sn len2 sn2)
   (when (eq len 0) (setq sn nil)) % no negative zero
   (when (eq len2 0) (setq sn2 nil)) % no negative zero
   (if (wlessp (wdifference heaptrapbound heaplast) 133)
            %garbage collection needed in this case
       (b64bigfromV7save len sn len2 sn2)
 
    (progn
     (setq sn (gtbig-open len sn))
     (VsetVl len)
     (when (wgreaterp len 0) (Vstore (Vreg 7) (aplus2 sn 2) 1))
     (setq sn2 (gtbig-open len2 sn2))
     (VsetVl len2)
     (when (wgreaterp len2 0) (Vstore (Vreg 6) (aplus2 sn2 2) 1))
     (cons sn sn2)
) ) )
 
(progn
(setq b64save1 (gtwarray 66))
(setq b64save2 (gtwarray 66))
(setq b64save3 (gtwarray 66))
(setq b64save4 (gtwarray 66))
(setq b64save5 (gtwarray 66))
nil)
 
(de b64bigfromV7save (len sn len2 sn2)
     (VsetVl len)
     (Vstore (Vreg 7) b64save1 1)
     (VsetVl len2)
     (Vstore (Vreg 6) b64save2 1)
 
     (setq sn (gtbig-open len sn))
     (setq sn2 (gtbig-open len2 sn2))
     (VsetVl len)
     (Vload (Vreg 7) b64save1 1)
     (VsetVl len2)
     (Vload (Vreg 6) b64save2 1)
 
     (VsetVl len)
     (when (wgreaterp len 0) (Vstore (Vreg 7) (aplus2 sn 2) 1))
     (VsetVl len2)
     (when (wgreaterp len2 0) (Vstore (Vreg 6) (aplus2 sn2 2) 1))
     (cons sn sn2)
)
 
(de b64bigfromV6 (len2 sn2)
   (when (eq len2 0) (setq sn2 nil)) % no negative zero
   (if (wlessp (wdifference heaptrapbound heaplast) 133)
            %garbage collection needed in this case
       (b64bigfromV6save len2 sn2)
 
    (progn
     (setq sn2 (gtbig-open len2 sn2))
     (VsetVl len2)
     (when (wgreaterp len2 0) (Vstore (Vreg 6) (aplus2 sn2 2) 1))
     sn2 
) ) )
 
(de b64bigfromV6save (len2 sn2)
     (VsetVl len2)
     (Vstore (Vreg 6) b64save3 1)
 
     (setq sn2 (gtbig-open len2 sn2))
     (VsetVl len2)
     (Vload (Vreg 6) b64save3 1)
     (when (wgreaterp len2 0) (Vstore (Vreg 6) (aplus2 sn2 2) 1))
     sn2 
)
 
(de sys2big-i (exp val sn)
   % exp exponent <=46
   % val mantissa
   % sn nil for positive, T for negative
    (cond ((wleq exp bbits**) (setq sn (gtbig-open 1 sn))
                              (iputv sn 1 val) sn)
          ((wleq exp (wplus2 bbits** bbits**))
                              (setq sn (gtbig-open 2 sn))
                              (iputv sn 1 (wand val bmask**))
                              (iputv sn 2 (wrshift val bbits**))
                              sn)
          (t                  (setq sn (gtbig-open 3 sn))
                              (iputv sn 1 (wand val bmask**))
                              (setq val (wrshift val bbits**))
                              (iputv sn 2 (wand val bmask**))
                              (iputv sn 3 (wrshift val bbits**))
                              sn)))
 
 
% add two positive bignums, full length routine
 
(de b99plus2aux (b1 b2 l1 l2 sign)
 
  (prog (l3 x y l carry)
        (setq carry 0)
 
         %            preliminary: interchange b1 and b2 so that b1
         %            is the longer one
       (when (wgreaterp l2 l1)
             (setq l3 l2)
             (setq l2 l1)
             (setq l1 l3)
             (setq x b2)
             (setq b2 b1)
             (setq b1 x) )
        %                    preformat the result number
       (setq x (gtbig-open (wplus2 l1 1) sign))
 
                    % first phase: add the bigits in those part, where
                    % both numbers have bigits
       (ifor (from i 1 l2 63)
             (do
                    % actual vector length
                    (setq l (wplus2 (wdifference l2 i) 1))
                    (when (wgreaterp l 63) (setq l 63))
                    (VsetVl l)
                        % fetch the operands
                    (Vload (Vreg 1) (Aplus b1 i 1) 1)
                    (Vload (Vreg 2) (aplus b2 i 1) 1)
        %                    do the operation
                    (when (not (eq carry 0))
                        (VsetElement (Vreg 1) 0
                              (wplus2 carry (VgetElement (Vreg 1) 0)))
                        (setq carry 0))
                    (setq l3 (B64RegisterAdd l l))
                    (VsetVl l)
        %                      store the result
                    (Vstore (Vreg 0) (aplus x i 1) 1)
        %                      new carry?
                    (when (wgreaterp l3 l)
                          (setq carry 1))     % add has carries 0 or 1
        %                    end first phase
              ))
 
        %                    second phase: second operand finished,
        %                    propagate the carries
        (setq l (wplus2 l2 1))
        (while (and (not (eq carry 0)) (wleq l l1))
               (setq l (wplus2 l 1))
               (setq y (wplus2 (getmem (aplus2 b1 l)) carry))
               (putmem (aplus2 x l) (wand y bmask**))
               (setq carry (wrshift y bbits**)) )
        %                    only one carry left?
        (when (and (not (eq carry 0)) (wgreaterp l l1))
              (putmem (aplus x l 1) 1)   % store last carry
              (return x))                          % no trim necessary
 
        %                    third phase: copy rest of first operand
       (ifor (from i l l1 64)
             (do
                    % actual vector length
                    (setq l (wplus2 (wdifference l1 i) 1))
                    (when (wgreaterp l 64) (setq l 64))
                    (VsetVl l)
                        % copy the operands
                    (Vload (Vreg 1) (Aplus b1 i 1) 1)
                    (Vstore (Vreg 1) (Aplus x i 1) 1)
                    ))
 
         %                   final result
        (return (trimbignum1 x l1)) ))
 
%
% subtract two positive bignums, full length routine
%
 
(de b99differenceaux (b1 b2 l1 l2 sign)
 
  (prog (l3 x y l carry)
        (setq carry 0)
 
         %            preliminary: interchange b1 and b2 so that b1
         %            is the longer one
       (when (wgeq l2 l1)
           (when (or (wgreaterp l2 l1) (bunsignedgreaterp b2 b1 l2 l1))
                     (setq x b1) (setq b1 b2) (setq b2 x)
                     (setq x l1) (setq l1 l2) (setq l2 x)
                     (setq sign (not sign))))
        %                    preformat the result number
       (setq x (gtbig-open l1 sign))
 
                    % first phase: sub the bigits in those part, where
                    % both numbers have bigits
       (ifor (from i 1 l2 63)
             (do
                    % actual vector length
                    (setq l (wplus2 (wdifference l2 i) 1))
                    (when (wgreaterp l 63) (setq l 63))
                    (VsetVl l)
                        % fetch the operands
                    (Vload (Vreg 1) (Aplus b1 i 1) 1)
                    (Vload (Vreg 2) (aplus b2 i 1) 1)
        %                    do the operation
        %                    old carry present?
                    (when (not (eq carry 0))
                        (VsetElement (Vreg 1) 0
                         (wdifference (VgetElement (Vreg 1) 0) carry))
                        (setq carry 0))
        %                    provision for the new carry
                    (VsetElement (Vreg 1) l 1)
                    (VsetElement (Vreg 2) l 0)
                    (setq l3 (B64Registersub (wplus2 l 1) l))
                    (VsetVl l)
        %                      store the result
                    (Vstore (Vreg 0) (aplus x i 1) 1)
        %                      has carry been eaten?
                    (when (not (eq (VgetElement (Vreg 0) l) 1))
                          (setq carry 1))
        %                    end first phase
              ))
 
        %                    second phase: second operand finished,
        %                    propagate the carries
        (setq l (wplus2 l2 1))
        (while (and (not (eq carry 0)) (wleq l l1))
               (setq l (wplus2 l 1))
               (setq y (wdifference (getmem (aplus2 b1 l)) carry))
               (if (wlessp y 0)
                        (setq y (wand y bmask**)) % -1 : carry remains
                        (setq carry 0))
               (putmem (aplus2 x l) y))
 
        %                    third phase: copy rest of first operand
       (ifor (from i l l1 64)
             (do
                    % actual vector length
                    (setq l (wplus2 (wdifference l1 i) 1))
                    (when (wgreaterp l 64) (setq l 64))
                    (VsetVl l)
                        % copy the operands
                    (Vload (Vreg 1) (Aplus b1 i 1) 1)
                    (Vstore (Vreg 1) (Aplus x i 1) 1)
                    ))
 
         %                   final result
        (return (trimbignum1 x l1)) ))
 
%%% full size multiply
%   the problem of multiplying two numbers bbbbbbbbbbbb * cccccccccc
%   is subdivided:
%   -   substrings of bbbbbb (length max. 64) are taken into
%       a vector register and multiplied by the c's . The resulting
%       digits are added into the result without carry propagation
%       The substrings of bbbb are processed one after the other
%       from right to left
%   -   when all adds are performed, the result is cleaned up by
%       propagating all carries in a 63-element vector loop
%   The deferred carry propagation is possible because the product
%   of two digits has at most 42 bit; so (2**22-1) of them can
%   be added into one 64-bit-word
 
(de b99times2aux (b1 b2 l1 l2 sign)
           % sign: sign for the result
  (prog (l carry b3 l3 x digit l4)
       (setq carry 0)
        %                    phase one:
        %                    preformat output number
        %                    multiple of 64
       (setq l3 (wplus2 l1 l2)) % result length
       (setq b3 (opengtbignum l3 sign))
       (Vsetvl 64)
       (Vclear (Vreg 0))
       (ifor (from i 1 l3 64)   % clear result for later add into
             (do
                    (setq l (wplus2 (wdifference l3 i) 1))
                    (when (wlessp l 64) (Vsetvl l))
                    (Vstore (Vreg 0) (Aplus b3 i 1) 1) ))
 
       %                      phase two:
       %                      calclate partial sums and add into res.
       %                      load first operand, convert to float, V7
       %                      multiple of 64
       (ifor (from j 1 l1 64)
        (do
          (setq l (wplus2 (wdifference l1 j) 1))
          (when (wgreaterp l 64) (setq l 64))
       %                      load and convert operand
          (VsetVL l)
          (Vload (Vreg 3) (Aplus2 (wplus2 b1 j) 1) 1)
          (Vclear (vreg 1))
          (VlshiftV (Vreg 7) (Vreg 3) 24)
        %                     loop over b2 bigits
          (ifor (from i 1 l2 1)
             (do
               (setq x (getmem (aplus2 (wplus2 b2 i) 1)))
               (VFtimesS (Vreg 2) (Vreg 7) (wlshift x 24))
                        %                 add to intermediate sum
               (Vplusv (Vreg 0) (Vreg 1)(Vreg 2))
               %                 take the last bigit for the result
               %                 and add it without carry propagation
               (setq digit (wplus2 (VgetElement (Vreg 0) 0) carry))
               (Putmem (wplus2 (wplus2 b3 i) j)
                       (wplus2 digit
                               (getmem (wplus2 b3 (wplus2 i j))) ))
 
               %                 shift the intermediate sum by one pos.
               (VdlshiftV (Vreg 1) (Vreg 0) 64)
           )  )
       %   add the last intermediate sum to the result
%          (VsetVl (aplus2 l -1))
           (Vload (Vreg 0) (aplus 1 b3 l2 j) 1)
           (VplusV (Vreg 2) (Vreg 1)(Vreg 0))
           (Vstore(Vreg 2) (aplus 1 b3 l2 j) 1)
 
       ) )     % end of j loop
 
       %    phase three:
       %    propagate the carries in the result
       %    multiple of 63 (one element needed for carry)
       (setq carry 0)
       (ifor (from i 1 l3 63)
        (do
          (setq l (wplus2 (wdifference l3 i) 1))
          (when (wgreaterp l 63) (setq l 63))
          (VsetVL l)
          (Vload (Vreg 1) (Aplus2 (wplus2 b3 i) 1) 1)
          (VsetElement (Vreg 1) 0
               (wplus2 carry (VgetElement (Vreg 1) 0)))
          (setq carry 0)
          (VsetElement (Vreg 1) l 0)
 
          (prog()
      %      adding carries (two times unconditionally)
 
             (VrshiftV (Vreg 3) (Vreg 1) bbits**)
             (VandS (vreg 2) bmask** (Vreg 1))
             (VdrshiftV (Vreg 4) (Vreg 3) 64)
             (VplusV (Vreg 1) (Vreg 2) (Vreg 4))
             (setq carry (wplus2 (Vgetelement (Vreg 3) (aplus2 l -1))
                                                                carry))
             (VrshiftV (Vreg 3) (Vreg 1) bbits**)
          lll
             (VandS (vreg 2) bmask** (Vreg 1))
             (VdrshiftV (Vreg 4) (Vreg 3) 64)
             (VplusV (Vreg 1) (Vreg 2) (Vreg 4))
             (setq carry (wplus2 (Vgetelement (Vreg 3) (aplus2 l -1))
                                                                carry))
             (VrshiftV (Vreg 3) (Vreg 1) bbits**)
             (when (neq (Vtestn (Vreg 3)) 0) (go lll))
          )  % end prog
          (Vstore (Vreg 1) (Aplus2 (wplus2 b3 i) 1) 1)
       )) % end loop i
 
 
       % finally make bignum ready
 
       (Putmem b3 (mkitem hwords-tag l3))
       (return (trimbignum1 b3 l3))
 ))
 
(setupglobals)
 
(de bigcompress (v)   % put 3 items in one
 (prog (uu u uuu nn nn3 nrem)
   (when (not (bigp v)) (return v))
   (when  (or (eq nn3 'bigcompress)
              (eq nn3 'bigdecompress)) (dumpo v 200))
   (setq nn   (openbsize v))
   (setq nn3  (wquotient (wplus2 nn 2) 3))
   (setq uuu  (mkbign (gtwrds (wplus2 nn3 1))))
   (putmem (aplus2 uuu 1) (if (openbminusp v) 'bigcompress
                                            'bigdecompress)) %for tests
   (putmem (aplus2 uuu 2) nn) % save original length
   (setq uu   (aplus2 uuu 3))
   (setq u    (aplus2 v 2))
   (setq nrem (longdivide-remainder (longdivide nn 3)))
   (cond ((eq nrem 0)
              (while (wgreaterp nn 0)
                   (setq nn3  (wquotient (wplus2 nn 2) 3))
                   (VsetVl (if (wgreaterp nn3 64) 64 nn3))
                   (Vload (Vreg 1) u 3)
                   (Vload (Vreg 2) (aplus2 u 1) 3)
                   (Vload (Vreg 3) (aplus2 u 2) 3)
                   (VlshiftV (Vreg 5) (Vreg 2) 21)
                   (VlshiftV (Vreg 6) (Vreg 3) 42)
                   (VorV (Vreg 7) (Vreg 1) (Vreg 5))
                   (VorV (Vreg 4) (Vreg 7) (Vreg 6))
                   (VStore (Vreg 4) uu 1)
                   (setq u  (aplus2 u 192))
                   (setq uu (aplus2 uu 64))
                   (setq nn (wdifference nn 192))))
         ((eq nrem 1)
              (while (wgreaterp nn 0)
                   (setq nn3  (wquotient (wplus2 nn 2) 3))
                   (VsetVl (if (wgreaterp nn3 64) 64 nn3))
                   (Vload (Vreg 1) u 3)
                   (if (wlessp nn 192) (progn
  %make sure that compressed bignums are 'clean' in the
  %last word
                            (vclear (vreg 2))
                            (vclear (vreg 3))
                            (vsetVl (wplus2 nn3 -1))
                            (Vload (Vreg 2) (aplus2 u 1) 3)
                            (Vload (Vreg 3) (aplus2 u 2) 3)
                            (vsetVl nn3))
                          (Vload (Vreg 2) (aplus2 u 1) 3)
                          (Vload (Vreg 3) (aplus2 u 2) 3))
                   (VlshiftV (Vreg 5) (Vreg 2) 21)
                   (VlshiftV (Vreg 6) (Vreg 3) 42)
                   (VorV (Vreg 7) (Vreg 1) (Vreg 5))
                   (VorV (Vreg 4) (Vreg 7) (Vreg 6))
                   (VStore (Vreg 4) uu 1)
                   (setq u  (aplus2 u 192))
                   (setq uu (aplus2 uu 64))
                   (setq nn (wdifference nn 192))))
         ((eq nrem 2)
              (while (wgreaterp nn 0)
                   (setq nn3  (wquotient (wplus2 nn 2) 3))
                   (VsetVl (if (wgreaterp nn3 64) 64 nn3))
                   (Vload (Vreg 1) u 3)
                   (Vload (Vreg 2) (aplus2 u 1) 3)
                   (if (wlessp nn 192) (progn
     %make sure that compressed bignums are 'clean' in the
     %last word
                            (vclear (vreg 3))
                            (vsetVl (wplus2 nn3 -1))
                            (Vload (Vreg 3) (aplus2 u 2) 3)
                            (vsetVl nn3))
                          (Vload (Vreg 3) (aplus2 u 2) 3))
                   (VlshiftV (Vreg 5) (Vreg 2) 21)
                   (VlshiftV (Vreg 6) (Vreg 3) 42)
                   (VorV (Vreg 7) (Vreg 1) (Vreg 5))
                   (VorV (Vreg 4) (Vreg 7) (Vreg 6))
                   (VStore (Vreg 4) uu 1)
                   (setq u  (aplus2 u 192))
                   (setq uu (aplus2 uu 64))
                   (setq nn (wdifference nn 192)))))
 (return uuu)
))
 
(de bigdecompress (v)       %retrieve 3 items from one
 (prog (uu u uuu nn nn3)
   % test for true compressed numbers
   (when (not (bigp v)) (return v))
   (setq nn3 (getmem (aplus2 v 1)))
   (when (not (or (eq nn3 'bigcompress)
                  (eq nn3 'bigdecompress))) (dumpo v 100))
   (setq nn (veclen v))
   (setq uuu (gtbig-open (getmem (aplus2 v 2))
                     (eq (getmem (aplus2 v 1)) 'bigcompress)))
   (setq uu (aplus2 uuu 2))
   (setq u  (aplus2 v 3))
  (while (wgreaterp nn 0)
   (VsetVl (if (wgreaterp nn 64) 64 nn))
   (Vload (Vreg 1) u 1)
   (Vset  (Vreg 2) 2097151)  %  mask 2** 21 -1
   (VrshiftV (Vreg 5) (Vreg 1) 21)
   (VrshiftV (Vreg 6) (Vreg 1) 42)
   (VandV (Vreg 3) (Vreg 1) (Vreg 2))
   (VStore (Vreg 3) uu 3)
   (VandV (Vreg 4) (Vreg 5) (Vreg 2))
   (VStore (Vreg 4) (aplus2 uu 1) 3)
   (VandV (Vreg 7) (Vreg 6) (Vreg 2))
   (VStore (Vreg 7) (aplus2 uu 2) 3)
   (setq u  (aplus2 u 64))
   (setq uu (aplus2 uu 192))
   (setq nn (wdifference nn 64)))
 (return uuu)
))
 
 
% Now install the important globals for this machine
 
(setq BigFloatHi* (btimes2 (bsmalldiff (btwopower 48) 1)
                                                  (btwopower 8#17777)))
(setq BigFloatLow* (bminus BigFloatHi*))
(setq FloatSysHi*  (float SysHi*))
(setq FloatSysLow* (float SysLow*))
 
(compiletime (setq *syslisp nil))
