%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:              SPY
% Description:       Spying by ticks
% Author:            H. Melenk ,W. Neun       ZIB Berlin
% Created:           10-Oct-86
% Mode:              Lisp
% Package:           Utilities (Cray-COS)
% Status:            Experimental (Do not distribute)
%
% Copyright (c) 1987, Konrad Zuse Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SPYing LISP via COS SPY calls             %
%   functions spyon and spyoff              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime
 (setq *syslisp t)
 (progn (de ijksize (x) x) % needed by EX opcode function
        (put 'exchange-jump 'opencode '((Move (Reg S0) (Reg 2))
                                        (EX 000)))))
 
(compiletime (flag '(spyjoin spysort spysort1 getaddresstable)
                                                'internalfunction))
 
(fluid '(&spylow& &spyhigh&  % interval of interest during spy run
         &spyres&     % intermediate: address list of functions
                      % list of pairs ( addr . function)
                      % will be sorted by addresses
         &spybucket&  % bucket size from initialization call
         &spyminimum& % minimum refs during printing
         &spying&     % indicator that spy is active
         &spymintime& % minimal time slice for spying
         &spytotal&   % total refs counter while printing
         &spysysrefs& % address table of those functions, which are
                      % not known at runtime via symfnc
                      % list of pairs ( addr . function)
         ))
 
(setq &spyminimum& 5)
(setq &spymintime& 100)
(setq &spying& nil)
(setq &spysysrefs& nil)
 
(de spyon (fwa lwa timeslice bucketsize) %building table for f$spy
  (prog (vbl)
   (setq vbl (gtvect 6))
   (setf (vecitm vbl 0)
                   (wor (wor fwa (wshift timeslice 24)) (wshift 1 56)))
   (setf (vecitm vbl 1) (wor lwa (wshift bucketsize 48)))
   (setf (vecitm vbl 2) 0)
   (setf (vecitm vbl 3) 0)
   (setq &spying& t)
   (exchange-jump (wplus2 (inf vbl) 1) 8#072)   % f$spy
) )
 
(de spyoff (bufferleng) %building table for f$spy
  (prog (vbl buffer)
   (setq buffer (make-vector (wplus2 bufferleng 5) nil))
   (setq vbl (gtvect 6))
   (setq buffer (inf buffer))
   (setq vbl (inf vbl))
   (setf (vecitm vbl 0) (wshift 2 56))
   (setf (vecitm vbl 1) 0)
   (setf (vecitm vbl 2) (wor (wshift bufferleng 48)(wplus2 buffer 1)))
   (setf (vecitm vbl 3) 0)
   (exchange-jump (wplus2 vbl 1) 8#072)   % f$spy
   (setq &spying& nil)
   (return buffer)
) )
 
(de spywhole (bucket)    %spy whole code area
    (if (memq bucket '(2 4 8 16 32 64 128 256)) t (setq bucket 4))
    (setq &spyhigh& lastbps)
    (setq &spylow&  128)
    (setq &spybucket&  bucket)
    (spyon &spylow& &spyhigh& &spymintime& bucket)
)
 
(de spyprint ()         %print results sorted
    (prog (length result addresstable)
 
       (when (not &spying& ) (return nil)) %trivial but necessary
 
       (setq length (wquotient (wdifference &spyhigh& &spylow&)
                               &spybucket&))
       (setq length (wplus2 length 1))
       (setq result (spyoff length))
       (setq addresstable (getaddresstable))
       (setq addresstable (spyjoin addresstable result ))
       (mapc addresstable (function (lambda (x)
                                       (when (wgreaterp (car x)
                                                        &spyminimum&)
                                           (prin2 (car x))
                                           (tab 10)
                                          (let ((z (quotient
                                                   (times2 (car x)1000)
                                                    &spytotal& )))
                                           (prin2 (wquotient z 10))
                                           (prin2 ".")
                                           (prin2 (wremainder z 10))
                                           (prin2 "%")
                                           (tab 20)
                                           (prin2t (cdr x)))))))))
 
(de getaddresstable ()
     (setq &spyres& (append &spySysRefs& nil))
     (mapobl (function (lambda (x)
                         (cond
                           ((funboundp x) nil)
                           ((flambdalinkp x) nil)
                           (t (setq &spyres&
                                  (cons (cons (symfnc x) x) &spyres&)))
     )        )        ) )
     (setq &spyres& (cons (cons heap 'end-of-code) &spyres&))
     (spysort (prog1 &spyres&
                     (setq &spyres& nil)))
)
 
(de spyjoin (addrtable vector)
     (prog (base counter currentfunction currenthigh currentcount
            final veclength nextbucket)
          (setq veclength (veclen vector))
          (setq counter 0)
          (setq &spytotal& 0)
      loop
          (while (and (wlessp counter veclength)
                      (vecitm vector (wplus2 counter 4))
                      (eq (vecitm vector counter) 0))
                           (setq counter (wplus2 counter 1)))
          (setq base (wplus2 &spylow& (wtimes2 &spybucket& counter)))
          (while (and addrtable (wlessp (caadr addrtable) base))
                  (setq addrtable (cdr addrtable)))
 
          (when (or (wgeq counter veclength) (not addrtable)
                    (not (vecitm vector (wplus2 counter 4))))
                     (return (spysort1 final)))
 
          (setq currentfunction (cdar addrtable))
          (setq currentcount 0)
          (setq currenthigh (caadr addrtable))
          (setq nextbucket (wplus2 base &spybucket&))
 
          (while (and (wleq nextbucket currenthigh)
                      (vecitm vector (wplus2 counter 4))
                      (wlessp counter veclength))
                  (setq currentcount (wplus2 currentcount
                                           (vecitm vector counter)))
                  (setq base (wplus2 base &spybucket&))
                  (setq nextbucket (wplus2 base &spybucket&))
                  (setq counter (wplus2 counter 1))
          )
          (setq currentcount (wplus2 currentcount
                                           (vecitm vector counter)))
          (when (and (wgreaterp currentcount 0)
                     (not (memq currentfunction '(buildrelocationfields
                          externalmarkfrombase
                          externalupdateitem !%reclaim
                          updatesymbols makeidfreelist
                          markfromvectorregisters updateallbases))))
% keep gc off percentage
                 (setq &spytotal& (wplus2 &spytotal& currentcount))
                 (setq final
                      (cons (cons currentcount currentfunction) final)))
          (setq addrtable (cdr addrtable))
          (go loop)
)  )
 
(de spysort (l)        % sort labels to ascending sequence
   (prog (changed actptr x1)
    loop (setq changed nil)
         (setq actptr l)
         (while (cdr actptr)
                (when (wgreaterp (caar actptr) (caadr actptr))
                      (setq changed t)
                      (setq x1 (cadr actptr))
                      (rplaca (cdr actptr)(car actptr))
                      (rplaca actptr x1))
                (setq actptr (cdr actptr)))
         (when changed (go loop))
         (return l)
) )
 
(de spysort1 (l)        % sort labels to descending sequence
   (prog (changed actptr x1)
    loop (setq changed nil)
         (setq actptr l)
         (while (cdr actptr)
                (when (wlessp (caar actptr) (caadr actptr))
                      (setq changed t)
                      (setq x1 (cadr actptr))
                      (rplaca (cdr actptr)(car actptr))
                      (rplaca actptr x1))
                (setq actptr (cdr actptr)))
         (when changed (go loop))
         (return l)
) )
 
(de quit ()
    (when &spying& (spyprint))
    (exit-To-Caller) )
(compiletime
 (setq *syslisp nil))
